//=============================================================================
// Calculates aim for a Sentinel.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelControllerComponent_Aiming extends Component
	within SentinelController;

/** Prediction will never see target as moving faster than this. */
var() float MaxPredictionSpeed;
/** If true, prediction will use trace(s) to check if predicted path is blocked. */
var() bool bPredictionUsesTrace;
/** If true, prediction will take into account the target's acceleration, not just their current velocity. */
var() bool bPredictAcceleration;
/** Redo the projectile travel time calculation based on where the target will be after the travel time has elapsed. */
var() bool bIterateProjectileTime;
/** Relative locations within target's bounding cylinder to try to aim at when failed to aim at the target's location. */
var() array<Vector> AimTestLocations;

/**
 * Determines the best place to shoot at to hit the target.
 */
function bool FindAimToHit(Actor A, out Vector AimSpot, out Rotator AimRotation)
{
	local Vector Origin;
	local Vector TargetLocation;
	local bool bCanHit;

	Origin = Cannon.GetPawnViewLocation();

	//If the enemy is not visible, point at where they were last detected.
	if(!bEnemyIsVisible && !bForceTarget)
	{
		AimSpot = LastDetectedLocation;
	}
	//If using an instant-hit weapon or the target doesn't move, just aim straight at it.
	else if(Cannon.SWeapon.GetProjectileClass() == none || A.IsStationary())
	{
		AimSpot = A.GetTargetLocation();
		TargetLocation = AimSpot;
		bCanHit = TraceCheckAim(A, Origin, TargetLocation, AimSpot);

		if(!bCanHit && A.bHasAlternateTargetLocation)
		{
			AimSpot = A.GetTargetLocation(, true);
			TargetLocation = AimSpot;
			bCanHit = TraceCheckAim(A, Origin, TargetLocation, AimSpot);
		}
	}
	//For moving targets, try to aim where they will be when the shot reaches them.
	else
	{
		PredictTargetLocation(A, Origin, AimSpot);
		TargetLocation = A.GetTargetLocation();
		bCanHit = TraceCheckAim(A, Origin, TargetLocation, AimSpot);

		if(!bCanHit)
		{
			//If the predicted location cannot be shot at, try half way.
			AimSpot = AimSpot - ((AimSpot - TargetLocation) / 2.0);
			bCanHit = TraceCheckAim(A, Origin, TargetLocation, AimSpot);

			if(!bCanHit)
			{
				//Still can't hit? Try not aiming ahead at all.
				AimSpot = TargetLocation;
				bCanHit = TraceCheckAim(A, Origin, TargetLocation, AimSpot);

				if(!bCanHit && A.bHasAlternateTargetLocation)
				{
					AimSpot = A.GetTargetLocation(, true);
					TargetLocation = AimSpot;
					bCanHit = TraceCheckAim(A, Origin, TargetLocation, AimSpot);
				}
			}
		}
	}

	AimRotation = Rotator(AimSpot - Origin);
	Cannon.SWeapon.AdjustAimToHit(A, AimSpot, AimRotation);

	return bCanHit;
}

/**
 * Tests to see if there is a clear shot to the target, and tries adjusting the aim if not.
 */
function bool TraceCheckAim(Actor A, Vector Origin, Vector TargetLocation, out Vector AimSpot)
{
	local bool bSuccess;
	local Vector Offset;
	local Vector TestSpot;
	local float TargetRadius, TargetHeight;
	local float CheckRadius, CheckHeight;
	local Vector V;

	A.GetBoundingCylinder(TargetRadius, TargetHeight);
	CheckRadius = TargetRadius - VSize2D(A.Location - TargetLocation);
	CheckHeight = TargetHeight - Abs(A.Location.Z - TargetLocation.Z);

	foreach AimTestLocations(V)
	{
		Offset.X = CheckRadius * V.X;
		Offset.Y = CheckRadius * V.Y;
		Offset.Z = CheckHeight * V.Z;
		TestSpot = AimSpot + Offset;

		if(A.ContainsPoint(TargetLocation + Offset) && FastTrace(TestSpot, Origin))
		{
			AimSpot = TestSpot;
			bSuccess = true;
			break;
		}
	}

	return bSuccess;
}

/**
 * Predicts where the target will be based on its velocity. Only call this if the Sentinel is using a projectile weapon.
 */
function PredictTargetLocation(Actor A, Vector Origin, out Vector AimSpot)
{
	local float PredictionTime;
	local float MaxTargetSpeed;
	local Pawn PawnTarget;
	local Vector PredictionVelocity;
	local Vector Extent;
	local Vector TraceStart, TraceEnd;
	local Vector HitLocation, HitNormal;

	AimSpot = A.GetTargetLocation();

	//How long it will take for projectile to reach target.
	PredictionTime = Cannon.SWeapon.GetProjectileTimeToLocation(AimSpot, Origin, Cannon.SController);

	if (bIterateProjectileTime)
	{
		//For best accuracy this whole function would be iterated, but that would be slow, so just project the target's current velocity.
		PredictionTime = Cannon.SWeapon.GetProjectileTimeToLocation(AimSpot + (A.Velocity * PredictionTime), Origin, Cannon.SController);
	}

	//Where the target will probably be by then.
	PredictionVelocity = A.Velocity;

	if(bPredictAcceleration)
	{
		PredictionVelocity += A.Acceleration * PredictionTime * 0.5;

		if(A.Physics == PHYS_Falling)
		{
			PredictionVelocity.Z += A.GetGravityZ() * PredictionTime * 0.5;
		}
	}

	PawnTarget = Pawn(A);
	
	if (PawnTarget == none)
	{
		MaxTargetSpeed = MaxPredictionSpeed;
	}
	else
	{
		switch(PawnTarget.Physics)
		{
			case PHYS_Walking:
				MaxTargetSpeed = FMin(MaxPredictionSpeed, PawnTarget.GroundSpeed);
				break;
			case PHYS_Falling:
				MaxTargetSpeed = FMin(MaxPredictionSpeed, PawnTarget.GetTerminalVelocity());
				break;
			case PHYS_Flying:
				MaxTargetSpeed = FMin(MaxPredictionSpeed, PawnTarget.AirSpeed);
				break;
			case PHYS_Swimming:
				MaxTargetSpeed = FMin(MaxPredictionSpeed, PawnTarget.WaterSpeed);
				break;
			case PHYS_Ladder:	
				MaxTargetSpeed = FMin(MaxPredictionSpeed, PawnTarget.LadderSpeed);
				break;
			default:
				MaxTargetSpeed = MaxPredictionSpeed;
				break;
		}
	}

	if(VSize(PredictionVelocity) > MaxTargetSpeed)
	{
		PredictionVelocity = Normal(PredictionVelocity) * MaxTargetSpeed;
	}

	if(bPredictionUsesTrace)
	{
		//Trace from target's current location to predicted location and assume they will stop if there is an obstacle.
		TraceStart = AimSpot;
		TraceEnd = TraceStart + (PredictionVelocity * PredictionTime * Cannon.AimAhead);
		A.GetBoundingCylinder(Extent.X, Extent.Z);
		Extent.Y = Extent.X;

		if(Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true, Extent,, TRACEFLAG_Blocking) == none)
		{
			AimSpot = TraceEnd;
		}
		else
		{
			AimSpot = HitLocation;
		}
	}
	else
	{
		AimSpot += PredictionVelocity * PredictionTime * Cannon.AimAhead;
	}
}

defaultproperties
{
	MaxPredictionSpeed=500.0

	AimTestLocations.Add((X=0.0,Y=0.0,Z=0.0))

	AimTestLocations.Add((X=0.4,Y=0.0,Z=0.0))
	AimTestLocations.Add((X=-0.4,Y=0.0,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=0.4,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=-0.4,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=0.0,Z=0.4))
	AimTestLocations.Add((X=0.0,Y=0.0,Z=-0.4))

	AimTestLocations.Add((X=0.8,Y=0.0,Z=0.0))
	AimTestLocations.Add((X=-0.8,Y=0.0,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=0.8,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=-0.8,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=0.0,Z=0.8))
	AimTestLocations.Add((X=0.0,Y=0.0,Z=-0.8))

	AimTestLocations.Add((X=0.95,Y=0.0,Z=0.0))
	AimTestLocations.Add((X=-0.95,Y=0.0,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=0.95,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=-0.95,Z=0.0))
	AimTestLocations.Add((X=0.0,Y=0.0,Z=0.95))
	AimTestLocations.Add((X=0.0,Y=0.0,Z=-0.95))
	AimTestLocations.Add((X=0.95,Y=0.0,Z=0.95))
	AimTestLocations.Add((X=-0.95,Y=0.0,Z=0.95))
	AimTestLocations.Add((X=0.0,Y=0.95,Z=0.95))
	AimTestLocations.Add((X=0.0,Y=-0.95,Z=0.95))
	AimTestLocations.Add((X=0.95,Y=0.0,Z=-0.95))
	AimTestLocations.Add((X=-0.95,Y=0.0,Z=-0.95))
	AimTestLocations.Add((X=0.0,Y=0.95,Z=-0.95))
	AimTestLocations.Add((X=0.0,Y=-0.95,Z=-0.95))
}