//=============================================================================
// There is no way to draw arbitrary actors on the map, so this objective hooks
// into it and thence allows all Sentinels to draw whatever they like.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelDummyObjective extends UTGameObjective;

simulated function PostBeginPlay()
{
	local UTMapInfo MP;

	super(Objective).PostBeginPlay();

	MP = UTMapInfo(WorldInfo.GetMapInfo());

	//Adding self to the UTMapInfo's Objective array causes it to call RenderExtraDetails as long as it has bFullDetail=true.
	if(MP != none)
	{
		MP.Objectives.AddItem(self);
	}
}

simulated function RenderExtraDetails(UTMapInfo MP, Canvas Canvas, UTPlayerController PlayerOwner, float ColorPercent, bool bSelected)
{
	local Sentinel S;

	foreach WorldInfo.AllPawns(class'Sentinel', S)
	{
		if((PlayerOwner == S.InstigatorController) || WorldInfo.GRI.OnSameTeam(PlayerOwner, S.PlayerReplicationInfo) || PlayerOwner.PlayerReplicationInfo.bOnlySpectator)
		{
			S.RenderMapIcon(MP, Canvas, PlayerOwner);
		}
	}
}

simulated function Destroyed()
{
	local UTMapInfo MP;

	super(Objective).Destroyed();

	MP = UTMapInfo(WorldInfo.GetMapInfo());

	if(MP != none)
	{
		MP.Objectives.RemoveItem(self);
	}
}

/**
 * Finds a preexisting instance of this class if one exists, or spawns a new one if not.
 *
 * @return	an instance of this class, or none if there is no UTMapInfo for this level.
 */
static final function SentinelDummyObjective CreateDummyObjective(WorldInfo WI)
{
	local UTMapInfo MI;
	local SentinelDummyObjective DummyObjective;

	MI = UTMapInfo(WI.GetMapInfo());

	if(MI != none)
	{
		MapExtentHack(MI);

		foreach WI.DynamicActors(class'SentinelDummyObjective', DummyObjective)
		{
			`log("Found DummyObjective:"@DummyObjective);
			break;
		}

		if(DummyObjective == none)
		{
			DummyObjective = WI.Spawn(default.Class);
			`log("Spawned DummyObjective:"@DummyObjective);
		}
	}

	return DummyObjective;
}

/**
 * Calculates MapExtent for the given UTMapInfo, ignoring SentinelDummyObjectives. Prevents maps that don't have a manually defined MapExtent from having their minimaps' scale changed by the presence of the extra objective.
 */
static final function MapExtentHack(UTMapInfo MI)
{
	local int NumNodes;
	local UTGameObjective O;
	local WorldInfo WI;

	if(MI.MapExtent == 0)
	{
		WI = WorldInfo(MI.Outer);

		MI.MapCenter = vect(0,0,0);

		ForEach WI.AllActors(class'UTGameObjective', O)
		{
			if(SentinelDummyObjective(O) == none)
			{
				MI.MapCenter += O.Location;
				NumNodes++;
			}
		}

		if(NumNodes > 0)
		{
			MI.MapCenter = MI.MapCenter / NumNodes;

			ForEach WI.AllActors(class'UTGameObjective', O)
			{
				if(SentinelDummyObjective(O) == none)
				{
					MI.MapExtent = FMax(MI.MapExtent, 2.75 * vsize2D(MI.MapCenter - O.Location));
				}
			}
		}
	}
}

defaultproperties
{
	bFirstObjective=false //So it doesn't try to do anything with other objectives on spawning.
	bIsDisabled=true
	SupportedEvents.Empty

	bNoDelete=false
	bStatic=false
	bCollideWhenPlacing=false
	bMustBeReachable=false

	RemoteRole=ROLE_SimulatedProxy
	bNetTemporary=true
}
