//=============================================================================
// Base class for all projectiles fired by a Sentinel.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile extends UTProjectile
	config(Mutator_Sentinel);

var config bool bProjectileLight;

function PreBeginPlay()
{
	super.PreBeginPlay();

	//Damage by a Sentinel is attributed to its owner, so set InstigatorController appropriately.
	if(Sentinel(Instigator) != none)
	{
		InstigatorController = Sentinel(Instigator).InstigatorController;
	}
}

simulated function SpawnFlightEffects()
{
	//Create projectile light for everyone.
	if(WorldInfo.NetMode != NM_DedicatedServer)
	{
		CreateProjectileLight();
	}

	super.SpawnFlightEffects();
}

simulated function CreateProjectileLight()
{
	if(bProjectileLight && !WorldInfo.bDropDetail && ProjectileLight == none && ProjectileLightClass != none)
	{
		ProjectileLight = new(self) ProjectileLightClass;
		AttachComponent(ProjectileLight);
	}
}

/**
 * Stop ambient sound component.
 */
simulated function TurnOffAmbientSound()
{
	local AudioComponent AComponent;

	foreach ComponentList(class'AudioComponent', AComponent)
	{
		if(AComponent.SoundCue == AmbientSound)
		{
			AComponent.Stop();
			break;
		}
	}
}

simulated function HideProjectile()
{
	super.HideProjectile();

	//Fixes ambient sound remaining when projectile hidden.
	TurnOffAmbientSound();
}

//Overriden to prevent Sentinel damaging itself.
simulated function ProcessTouch(Actor Other, Vector HitLocation, Vector HitNormal)
{
	if(DamageRadius > 0.0)
	{
		Explode(HitLocation, HitNormal);
	}
	else
	{
		if(Instigator != none && (Pawn(Other) == none || !Instigator.IsSameTeam(Pawn(Other))))
		{
			Other.TakeDamage(Damage, InstigatorController, HitLocation, MomentumTransfer * Normal(Velocity), MyDamageType,, self);
			Sentinel(Instigator).NotifyCausedDamage(Other);
		}

		Shutdown();
	}
}

//Overriden to prevent momentum being imparted to friendly vehicles, and damage being dealt to friendly hoverboards.
simulated singular event HitWall(Vector HitNormal, Actor Wall, PrimitiveComponent WallComp)
{
	ImpactedActor = Wall;

	if(!Wall.bStatic && !Wall.bWorldGeometry)
	{
		if(DamageRadius == 0.0)
		{
			if(Instigator != none && (Pawn(Wall) == none || !Instigator.IsSameTeam(Pawn(Wall))))
			{
				Wall.TakeDamage(Damage, InstigatorController, Location, MomentumTransfer * Normal(Velocity), MyDamageType,, self);
				Sentinel(Instigator).NotifyCausedDamage(Wall);
			}
		}
	}

	Explode(Location, HitNormal);
	ImpactedActor = None;
}

//Overriden to prevent Sentinel damaging itself or its owner.
//Also contains fixes for applying splash damage to large actors.
simulated function bool ProjectileHurtRadius(float DamageAmount, float InDamageRadius, class<DamageType> DamageType, float Momentum,
						vector HurtOrigin, vector HitNormal, optional class<DamageType> ImpactedActorDamageType)
{
	local bool bCausedDamage, bInitializedAltOrigin, bFailedAltOrigin;
	local Actor	Victim;
	local Vector AltOrigin;

	// Prevent HurtRadius() from being reentrant.
	if(bHurtEntry)
		return false;

	bHurtEntry = true;
	bCausedDamage = false;

	// if ImpactedActor is set, we actually want to give it full damage, and then let him be ignored by super.HurtRadius()
	if ( (ImpactedActor != none) && (ImpactedActor != self)  )
	{
		if(Pawn(ImpactedActor) == none || !Instigator.IsSameTeam(Pawn(ImpactedActor)))
		{
			ImpactedActor.TakeRadiusDamage( InstigatorController, DamageAmount, InDamageRadius,
							(ImpactedActorDamageType != none) ? ImpactedActorDamageType : DamageType,
							Momentum, HurtOrigin, true, self );
			// need to check again in case TakeRadiusDamage() did something that went through our explosion path a second time
			if (ImpactedActor != none)
			{
				bCausedDamage = ImpactedActor.bProjTarget;
			}
		}
	}

	foreach CollidingActors(class'Actor', Victim, InDamageRadius, HurtOrigin, true)
	{
		if ( !Victim.bWorldGeometry && (Victim != self) && (Victim != ImpactedActor) && (Victim.bProjTarget || (NavigationPoint(Victim) == none)) )
		{
			if(UTOnslaughtPowernode(Victim) != none)
			{
				bCausedDamage = bCausedDamage || PowernodeRadiusDamage(UTOnslaughtPowernode(Victim), DamageAmount, InDamageRadius, DamageType, Momentum, HurtOrigin);
			}
			else if(Pawn(Victim) == none || !Instigator.IsSameTeam(Pawn(Victim)))
			{
				if(!FastTrace(HurtOrigin, Victim.Location,, TRUE))
				{
					// try out from wall, in case trace start was embedded
					if ( !bInitializedAltOrigin )
					{
						// initialize alternate trace start
						bInitializedAltOrigin = true;
						AltOrigin = HurtOrigin + class'UTPawn'.Default.MaxStepHeight * HitNormal;
						if ( !FastTrace(HurtOrigin, AltOrigin,, TRUE) )
						{
							if ( Velocity == vect(0,0,0) )
							{
								bFailedAltOrigin = true;
							}
							else
							{
								AltOrigin = HurtOrigin - class'UTPawn'.Default.MaxStepHeight * normal(Velocity);
								bFailedAltOrigin = !FastTrace(HurtOrigin, AltOrigin,, TRUE);
							}
						}
					}
					if ( bFailedAltOrigin || !FastTrace(AltOrigin, Victim.Location,, TRUE) )
					{
						continue;
					}
				}

				if(Victim.CollisionComponent != none)
				{
					BigActorRadiusDamage(Victim, DamageAmount, InDamageRadius, DamageType, Momentum, HurtOrigin);
				}
				else
				{
					Victim.TakeRadiusDamage(InstigatorController, DamageAmount, InDamageRadius, DamageType, Momentum, HurtOrigin, false, self);
				}

				bCausedDamage = bCausedDamage || Victim.bProjTarget;
			}
		}
	}
	bHurtEntry = false;

	return bCausedDamage;
}

/**
 * Big actors like vehicles sometimes don't register splash damage via TakeRadiusDamage due to lack of sophistication in the distance calculations. This fixes that.
 * TODO: Remove this if TakeRadiusDamage is ever improved.
 */
final function BigActorRadiusDamage(Actor Victim, float DamageAmount, float InDamageRadius, class<DamageType> DamageType, float Momentum, Vector HurtOrigin)
{
	local Vector HitLocation, HitNormal;
	local float Dist;
	local float DamageScale;

	if(TraceComponent(HitLocation, HitNormal, Victim.CollisionComponent, Victim.CollisionComponent.GetPosition(), HurtOrigin))
	{
		Dist = VSize(HitLocation - HurtOrigin);
		Dist = FClamp(Dist, 0.0, InDamageRadius);
		DamageScale = 1 - Dist/InDamageRadius;

		if(DamageScale > 0.0)
		{
			Victim.TakeDamage(DamageAmount * DamageScale, InstigatorController, HitLocation, Normal(HitLocation - HurtOrigin) * Momentum * DamageScale, DamageType,, self);
			Sentinel(Instigator).NotifyCausedDamage(Victim);

			if(Vehicle(Victim) != none)
			{
				Vehicle(Victim).DriverRadiusDamage(DamageAmount, InDamageRadius, InstigatorController, DamageType, Momentum, HurtOrigin, self);
			}
		}
	}
}

/**
 * Special case for applying radius damage to power nodes, since their extra-tall shape makes TakeRadiusDamage fail if they are hit on the top part without bFullDamage.
 * TODO: Remove this if TakeRadiusDamage is ever fixed for power nodes.
 */
final function bool PowernodeRadiusDamage(UTOnslaughtPowernode PN, float DamageAmount, float InDamageRadius, class<DamageType> DamageType, float Momentum, Vector HurtOrigin)
{
	local Vector HitLocation, HitNormal;
	local float Dist;
	local float DamageScale;
	local bool bCausedDamage;

	if(VSize(PN.EnergySphereCollision.GetPosition() - HurtOrigin) < VSize(PN.NodeBase.GetPosition() - HurtOrigin))
	{
		if(!TraceComponent(HitLocation, HitNormal, PN.EnergySphereCollision, PN.EnergySphereCollision.GetPosition(), HurtOrigin))
		{
			HitLocation = PN.Location;
		}
	}
	else if(!TraceComponent(HitLocation, HitNormal, PN.NodeBase, PN.NodeBase.GetPosition(), HurtOrigin))
	{
		HitLocation = PN.Location;
	}

	Dist = VSize(HitLocation - HurtOrigin);
	Dist = FClamp(Dist, 0.0, InDamageRadius);
	DamageScale = 1 - Dist/InDamageRadius;

	if(DamageScale > 0.0)
	{
		PN.TakeDamage(DamageAmount * DamageScale, InstigatorController, HitLocation, Normal(HitLocation - HurtOrigin) * Momentum, DamageType,, self);
		Sentinel(Instigator).NotifyCausedDamage(PN);
		bCausedDamage = true;
	}

	return bCausedDamage;
}

defaultproperties
{
	bProjectileLight=true
}