//=============================================================================
// Simple component that contains functionality for bouncing off a surface.
// http://mrevil.pwProj.blueyonder.co.uk/unreal/
//=============================================================================
class SentinelProjectileComponent_Physics extends Component
	within SentinelProjectile;

/** Randomness added to velocity when bouncing. */
var() float BounceRandomness;
/** Overall reduction in speed when bouncing. */
var() float BounceDamping;
/** Damping perpendicular to surface. 2.0 => no damping, 1.0 => stick parallel to surface, <1.0 => pass through surface (assuming it doesn't bounce again immediately). */
var() float BounceAngle;

/** Reduction in velocity on entering water. */
var() float EnterWaterDamping;
/** AccelRate set to this when entering water. */
var() float WaterAccelRate;
/** MaxSpeed set to this when entering water. */
var() float WaterMaxSpeed;

/** Sound to play on entering water. */
var() SoundCue WaterEntrySound;
/** Effect to spawn on entering water. */
var() ParticleSystem WaterEntryEffect;
/** Sound to play on exiting water. */
var() SoundCue WaterExitSound;
/** Effect to spawn on exiting water. */
var() ParticleSystem WaterExitEffect;

/**
 * Bounce with damping.
 *
 * @param	HitNormal	unit vector perpendicular to surface
 */
function Bounce(Vector HitNormal)
{
	local float CurrentSpeed;

	CurrentSpeed = VSize(Velocity);

	Velocity -= BounceAngle * HitNormal * (Velocity dot HitNormal);

	if(BounceRandomness != 0.0)
	{
		Velocity = Normal(Velocity) * (1.0 - BounceRandomness);
		Velocity += VRand() * BounceRandomness;
		Velocity = Normal(Velocity) * CurrentSpeed;
	}

	Velocity = Normal(Velocity) * CurrentSpeed * (1.0 - BounceDamping);
}

/**
 * Handles changing PhysicsVolume e.g. water splashes etc.
 */
function PhysicsVolumeChange(PhysicsVolume NewVolume)
{
	if(NewVolume.bWaterVolume)
	{
		if(PhysicsVolume == none || !PhysicsVolume.bWaterVolume)
		{
			EnterWater();
		}
	}
	else
	{
		if(PhysicsVolume == none || PhysicsVolume.bWaterVolume)
		{
			ExitWater();
		}
	}
}

/**
 * Plays effects and alters velocity etc. on entering water. No need to call this directly, PhysicsVolumeChange calls it automatically.
 */
function EnterWater()
{
	AccelRate = WaterAccelRate;
	MaxSpeed = WaterMaxSpeed;
	Velocity *= (1.0 - EnterWaterDamping);

	if(EffectIsRelevant(Location, false, MaxEffectDistance))
	{
		if(WaterEntryEffect != none)
			WorldInfo.MyEmitterPool.SpawnEmitter(WaterEntryEffect, Location);

		if(WaterEntrySound != none)
			PlaySound(WaterEntrySound, true);
	}
}

/**
 * Plays effects and alters velocity etc. on leaving water. No need to call this directly, PhysicsVolumeChange calls it automatically.
 */
function ExitWater()
{
	AccelRate = default.AccelRate;
	MaxSpeed = default.MaxSpeed;

	if(EffectIsRelevant(Location, false, MaxEffectDistance))
	{
		if(WaterExitEffect != none)
			WorldInfo.MyEmitterPool.SpawnEmitter(WaterExitEffect, Location);

		if(WaterExitSound != none)
			PlaySound(WaterExitSound, true);
	}
}

defaultproperties
{
	BounceRandomness=0.1
	BounceDamping=0.2
	BounceAngle=1.8
}