//=============================================================================
// I'll take your brain to another dimension.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_Disturber extends SentinelProjectile;

/** FireInfo for zapping. */
var() const SentinelWeaponComponent_FireInfo ZapFireInfo;
/** Beam effect for zapping. */
var() SentinelWeaponComponent_MuzzleFlash ZapFlash;

struct ZapStruct
{
	var Pawn Zapee;
	var Vector ZapEndLocation;
	var byte ZapCount;
};
var repnotify ZapStruct ZapInfo;
var byte LastZapCount;

/** Holds values used whilst imploding (prefixed with "Explode" because they are set in Explode()). */
var Vector ExplodeLocation, ExplodeNormal;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		ZapInfo;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'ZapInfo')
	{
		SpawnBeamEffect();
	}

	super.ReplicatedEvent(VarName);
}

/**
 * Returns the closest valid target pawn within zapping range.
 */
function Pawn FindZapee()
{
	local Pawn P, ClosestPawn;
	local float Distance, ClosestDistance;

	foreach WorldInfo.AllPawns(class'Pawn', P, Location, ZapFireInfo.MaxRange)
	{
		if(!P.IsValidTargetFor(InstigatorController))
			continue;

		if(Instigator.IsSameTeam(P)) //Prevent Sentinels from zapping their owners in non-team games.
			continue;

		Distance = VSize(P.Location - Location);

		if(ClosestPawn != none && ClosestDistance < Distance)
			continue;

		if(!FastTrace(P.Location, Location,, true))
			continue;

		ClosestPawn = P;
		ClosestDistance = Distance;
	}

	return ClosestPawn;
}

/**
 * Zaps the closest pawn.
 */
function bool Zap()
{
	local Pawn Zapee;
	local bool bZapped;

	Zapee = FindZapee();

	if(Zapee != none)
	{
		ZapInfo.Zapee = Zapee;
		ZapInfo.ZapEndLocation = Zapee.Location;
		ZapInfo.ZapCount++;

		Zapee.TakeDamage(ZapFireInfo.Damage, InstigatorController, Zapee.Location, ZapFireInfo.Momentum * Normal(Zapee.Location - Location), ZapFireInfo.DamageType,, self);
		Sentinel(Instigator).NotifyCausedDamage(Zapee);
		MakeNoise(ZapFireInfo.FireLoudness);
		SpawnBeamEffect();

		bZapped = true;
	}

	return bZapped;
}

simulated function SpawnBeamEffect()
{
	if(ZapInfo.ZapCount != LastZapCount)
	{
		LastZapCount = ZapInfo.ZapCount;

		if(WorldInfo.NetMode != NM_DedicatedServer)
		{
			PlaySound(ZapFireInfo.FireSound, true,,, Location);

			ZapFlash.ResetToDefaults();
			AttachComponent(ZapFlash);

			if(ZapInfo.Zapee != none && !ZapInfo.Zapee.bDeleteMe)
			{
				ZapFlash.SetActorParameter('BeamTarget', ZapInfo.Zapee);
			}
			else
			{
				ZapFlash.SetVectorParameter('BeamEnd', ZapInfo.ZapEndLocation);
			}

			ZapFlash.Flash();
		}
	}
}

simulated function Explode(Vector HitLocation, Vector HitNormal)
{
	TurnOffAmbientSound();

	if(ProjEffects != none)
	{
		ProjEffects.DeactivateSystem();
	}

	ExplodeLocation = HitLocation;
	ExplodeNormal = HitNormal;

	SetPhysics(PHYS_None);
	SetCollision(false, false);
	SpawnExplosionEffects(ExplodeLocation, ExplodeNormal);
	GotoState('Imploding');

	//If the projectile explodes immediately on spawning, this happens before either SpawnFlightEffects or SetInitialState.
	bShuttingDown = true; //Stops effects from being spawned.
	InitialState = 'Imploding'; //Ensure implosion happens.
}

state Imploding
{
	function Pawn FindZapee()
	{
		return none;
	}

	function bool Zap(){}
	simulated function SpawnBeamEffect(){}
	simulated function Explode(Vector HitLocation, Vector HitNormal){}

Begin:
	ProjectileHurtRadius(Damage, DamageRadius, MyDamageType, MomentumTransfer, ExplodeLocation, ExplodeNormal);
	Sleep(0.2);
	ProjectileHurtRadius(Damage, DamageRadius * 0.9, MyDamageType, MomentumTransfer, ExplodeLocation, ExplodeNormal);
	Sleep(0.2);
	ProjectileHurtRadius(Damage, DamageRadius * 0.8, MyDamageType, MomentumTransfer, ExplodeLocation, ExplodeNormal);
	Sleep(0.2);
	ProjectileHurtRadius(Damage, DamageRadius * 0.6, MyDamageType, MomentumTransfer, ExplodeLocation, ExplodeNormal);
	Sleep(0.2);
	ProjectileHurtRadius(Damage, DamageRadius * 0.4, MyDamageType, MomentumTransfer, ExplodeLocation, ExplodeNormal);
	Sleep(0.2);
	ProjectileHurtRadius(Damage, DamageRadius * 0.2, MyDamageType, MomentumTransfer, ExplodeLocation, ExplodeNormal);
	Sleep(2.5); //Time for effects to finish.
	Destroy();
}

defaultproperties
{
	ProjFlightTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Disturber_Projectile'
	ProjExplosionTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Disturber_Explosion'
	MaxEffectDistance=9000.0
	ProjectileLightClass=class'ProjectileLight_Disturber'

	Speed=800.0
	MaxSpeed=1000.0
	AccelRate=0.0

	Damage=15
	DamageRadius=300.0
	MomentumTransfer=-30000.0

	MyDamageType=class'UTDmgType_SentinelDisturber'
	LifeSpan=30.0
	bNetTemporary=false
	bCollideComplex=false
	bProjTarget=true

	AmbientSound=SoundCue'Sentinel_Resources.Sounds.Weapons.aliendrone1Cue'
	ExplosionSound=SoundCue'Sentinel_Resources.Sounds.Weapons.ImplodeCue'
	ExplosionLightClass=class'UTExplosionLight_Disturber'

	Begin Object Name=CollisionCylinder
		CollisionRadius=8.0
		CollisionHeight=8.0
		AlwaysLoadOnClient=true
		AlwaysLoadOnServer=true
		BlockNonZeroExtent=true
		BlockZeroExtent=true
		BlockActors=true
		CollideActors=true
	End Object

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		Damage=66.6
		Momentum=0.0
		DamageType=class'UTDmgType_SentinelDisturber_Zap'
		MaxRange=1500.0
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.electricity1Cue'
	End Object
	ZapFireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlash Name=MuzzleFlash0
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Disturber_Beam_Zap'
	End Object
	ZapFlash=MuzzleFlash0
	Components.Add(MuzzleFlash0)
}