//=============================================================================
// A ball of burning stuff.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_Flame extends SentinelProjectile;

var() SentinelProjectileComponent_Physics PhysicsComponent;

/** Time between checking for things to damage. */
var() float DamageCheckInterval;
/** Class of inventory to give pawns to handle setting them on fire. */
var() class<UTInv_Cooker> CookerClass;
/** Controls how quickly victims are set on fire, in conjunction with DamageCheckInterval. */
var() float Heat;
/** Actual damage radius increases from StartDamageRadius when first spawned, to EndDamageRadius at end of LifeSpan. */
var() float StartDamageRadius;
var() float EndDamageRadius;

function Init(Vector Direction)
{
	super.Init(Direction);

	SetTimer(DamageCheckInterval, true);
}

function Timer()
{
	FlameHurtRadius(Damage, DamageRadius, MyDamageType, MomentumTransfer, Location);
}

function bool FlameHurtRadius(float DamageAmount, float InDamageRadius, class<DamageType> DamageType, float Momentum, Vector HurtOrigin)
{
	local bool bCausedDamage;
	local Actor Victim;
	local Pawn P;

	// Prevent HurtRadius() from being reentrant.
	if(bHurtEntry)
		return false;

	DamageRadius = (((default.LifeSpan - LifeSpan) / default.LifeSpan) * (EndDamageRadius - StartDamageRadius)) + StartDamageRadius;

	bHurtEntry = true;
	bCausedDamage = false;

	foreach CollidingActors(class'Actor', Victim, DamageRadius, HurtOrigin, true)
	{
		if(!Victim.bWorldGeometry && (Victim != self) && (Victim.bProjTarget || (NavigationPoint(Victim) == none)))
		{
			P = Pawn(Victim);

			if(UTOnslaughtPowernode(Victim) != none)
			{
				bCausedDamage = bCausedDamage || PowernodeRadiusDamage(UTOnslaughtPowernode(Victim), DamageAmount, InDamageRadius, DamageType, Momentum, HurtOrigin);
			}
			if(P == none || !Instigator.IsSameTeam(P))
			{
				if(FastTrace(HurtOrigin, Victim.Location,, true))
				{
					if(Victim.CollisionComponent != none)
					{
						BigActorRadiusDamage(Victim, DamageAmount, InDamageRadius, DamageType, Momentum, HurtOrigin);
					}
					else
					{
						Victim.TakeRadiusDamage(InstigatorController, DamageAmount, DamageRadius, DamageType, Momentum, HurtOrigin, false, self);
					}

					bCausedDamage = bCausedDamage || Victim.bProjTarget;

					if(UTPawn(P) != none)
					{
						IgnitePawn(P);
					}
				}
			}
		}
	}

	bHurtEntry = false;

	return bCausedDamage;
}

function IgnitePawn(Pawn P)
{
	local Inventory Cooker;

	Cooker = P.FindInventoryType(CookerClass);

	if(Cooker == none)
	{
		Cooker = P.CreateInventory(CookerClass);
	}

	if(Cooker != none)
	{
		UTInv_Cooker(Cooker).Cook(Heat, InstigatorController);
	}
}

simulated function ProcessTouch(Actor Other, Vector HitLocation, Vector HitNormal)
{
	PhysicsComponent.Bounce(HitNormal);
	Acceleration = vect(0.0, 0.0, 0.0);
}

simulated function HitWall(Vector HitNormal, Actor Wall, PrimitiveComponent WallComp)
{
	PhysicsComponent.Bounce(HitNormal);
	Acceleration = vect(0.0, 0.0, 0.0);
}

simulated function PhysicsVolumeChange(PhysicsVolume NewVolume)
{
	super.PhysicsVolumeChange(NewVolume);

	PhysicsComponent.PhysicsVolumeChange(NewVolume);

	if(NewVolume.bWaterVolume)
	{
		Destroy();
	}
}

defaultproperties
{
	DamageCheckInterval=0.1
	CookerClass=class'UTInv_Cooker'
	Heat=1.5

	ProjFlightTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Flame_Projectile'
	MaxEffectDistance=7000.0
	AmbientSound=SoundCue'Sentinel_Resources.Sounds.Weapons.FireChuckerFireSoundCue'
	ProjectileLightClass=class'ProjectileLight_Flame'
	bProjectileLight=false

	Speed=500.0
	MaxSpeed=900.0
	AccelRate=100.0

	Damage=6
	StartDamageRadius=64.0
	DamageRadius=64.0
	EndDamageRadius=160.0
	MomentumTransfer=0.0
	NetCullDistanceSquared=144000000.0

	MyDamageType=class'UTDmgType_SentinelFlame'
	LifeSpan=3.0

	bCanBeDamaged=false
	bSwitchToZeroCollision=false
	bCollideComplex=false
	bRotationFollowsVelocity=true

	Begin Object Name=CollisionCylinder
		CollisionRadius=8
		CollisionHeight=8
		AlwaysLoadOnClient=true
		AlwaysLoadOnServer=true
		BlockNonZeroExtent=false
		BlockZeroExtent=false
		BlockActors=false
		CollideActors=true
	End Object

	Begin Object Class=SentinelProjectileComponent_Physics Name=PhysicsComponent0
		BounceRandomness=0.9
		BounceDamping=0.8
		BounceAngle=1.1
		WaterEntrySound=SoundCue'Sentinel_Resources.Sounds.Weapons.steamfx4Cue'
		WaterEntryEffect=ParticleSystem'Sentinel_Resources.Effects.Misc.P_Steam'
	End Object
	PhysicsComponent=PhysicsComponent0
}