//=============================================================================
// Has a hell of a kick.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_MOG extends SentinelProjectile;

/** Rdius at which force drops to zero. */
var() float ForceRadius;
/** Magnitude of force applied to projectiles. */
var() float ProjectileForceMagnitude;
/** Magnitude of force to apply to non-RB Pawns. */
var() float PawnForceMagnitude;
/** Manitude of force applied to vehicles. */
var() float VehicleForceMagnitude;
/** Magnitude of force applied to non-vehicle RB actors. */
var() float RBForceMagnitude;
/** UTPawns travelling faster than this many times their GroundSpeed are forced to ragdoll mode. */
var() float RagdollThresholdScale;

simulated state Forcing
{
	simulated function Tick(float DeltaTime)
	{
		local Actor A;
		local Pawn P;
		local Projectile Proj;

		if(Instigator == none || Instigator.PlayerReplicationInfo == none)
		{
			return;
		}

		//DynamicActors must be used instead of CollidingActors, or else zero-extent projectiles are not found.
		foreach DynamicActors (class'Actor', A)
		{
			if(VSize(Location - A.Location) > ForceRadius)
				continue;

			P = Pawn(A);

			if(P != none)
			{
				if(P.IsValidEnemyTargetFor(SentinelReplicationInfo(Instigator.PlayerReplicationInfo).OwnerPRI, false))
				{
					if(P.Physics == PHYS_RigidBody)
					{
						AffectRBPawn(P);
					}
					else
					{
						AffectPawn(P, DeltaTime);
					}
				}
			}
			else
			{
				Proj = Projectile(A);

				if(Proj != none)
				{
					if(!WorldInfo.GRI.OnSameTeam(Proj, SentinelReplicationInfo(Instigator.PlayerReplicationInfo).OwnerPRI))
					{
						AffectProjectile(Proj, DeltaTime);
					}
				}
			}
		}
	}
}

simulated function AffectProjectile(Projectile Proj, float DeltaTime)
{
	local Vector NewVelocity;

	if(FastTrace(Location, Proj.Location))
	{
		NewVelocity = Proj.Location - Location;
		NewVelocity = Normal(NewVelocity) * FMax(1.0 - (VSize(NewVelocity) / ForceRadius), 0.0);
		NewVelocity *= DeltaTime * ProjectileForceMagnitude;
		NewVelocity += Proj.Velocity;

		if(VSize(NewVelocity) > Proj.MaxSpeed)
		{
			NewVelocity = Normal(NewVelocity) * Proj.MaxSpeed;
		}

		Proj.Velocity = NewVelocity;
		//Setting rotation and acceleration like this is not ideal, since not all projectiles behave like this.
		Proj.SetRotation(Rotator(NewVelocity));
		Proj.Acceleration = Normal(NewVelocity) * VSize(Proj.Acceleration);
	}
}

simulated function AffectPawn(Pawn P, float DeltaTime)
{
	local Vector HitLocation, HitNormal;
	local Vector AppliedForce;
	local UTPawn UTP;

	if(P == Instigator)
		return;

	if(TraceComponent(HitLocation, HitNormal, P.CollisionComponent, P.CollisionComponent.GetPosition(), Location))
	{
		if(FastTrace(Location, HitLocation))
		{
			P.TakeDamage(0, InstigatorController, HitLocation, vect(0.0, 0.0, 0.0), MyDamageType,, self);
			Sentinel(Instigator).NotifyCausedDamage(P);

			AppliedForce = HitLocation - Location;
			AppliedForce = Normal(AppliedForce) * FMax(1.0 - (VSize(AppliedForce) / ForceRadius), 0.0);
			AppliedForce *= DeltaTime * PawnForceMagnitude;
			AppliedForce /= P.Mass;
			P.AddVelocity(AppliedForce, HitLocation, MyDamageType);

			UTP = UTPawn(P);

			if(UTP != none && !UTP.bFeigningDeath && VSize(UTP.Velocity) > (UTP.GroundSpeed * RagdollThresholdScale))
			{
				UTP.ForceRagdoll();
			}
		}
	}
}

simulated function AffectRBPawn(Pawn P)
{
	local Vector HitLocation, HitNormal;
	local Vector AppliedForce;

	if(TraceComponent(HitLocation, HitNormal, P.CollisionComponent, P.CollisionComponent.GetPosition(), Location))
	{
		if(FastTrace(Location, HitLocation))
		{
			P.TakeDamage(0, InstigatorController, HitLocation, vect(0.0, 0.0, 0.0), MyDamageType,, self);
			Sentinel(Instigator).NotifyCausedDamage(P);

			AppliedForce = HitLocation - Location;
			AppliedForce = Normal(AppliedForce) * FMax(1.0 - (VSize(AppliedForce) / ForceRadius), 0.0);

			if(P.IsA('Vehicle'))
				AppliedForce *= VehicleForceMagnitude;
			else
				AppliedForce *= RBForceMagnitude;

			P.Mesh.AddForce(AppliedForce, HitLocation);
		}
	}
}

defaultproperties
{
	ForceRadius=200.0
	ProjectileForceMagnitude=5000.0
	PawnForceMagnitude=2000000.0
	VehicleForceMagnitude=2000.0
	RBForceMagnitude=500.0
	RagdollThresholdScale=2.0

	ProjFlightTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_MOG_Projectile'
	ProjExplosionTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_MOG_Explosion'
	ExplosionLightClass=class'UTExplosionLight_MOG'
	MaxEffectDistance=10000.0
	ExplosionSound=SoundCue'Sentinel_Resources.Sounds.Weapons.MOG_ExplosionCue'
	AmbientSound=SoundCue'Sentinel_Resources.Sounds.Weapons.MOG_AmbientCue'
	ProjectileLightClass=class'ProjectileLight_MOG'

	Speed=800.0
	MaxSpeed=1500.0
	AccelRate=0.0

	Damage=22
	DamageRadius=0.0
	MomentumTransfer=50000.0

	MyDamageType=class'UTDmgType_SentinelMOG'
	LifeSpan=10.0
	InitialState=Forcing
	bCollideWorld=true
}