//=============================================================================
// Base class for projectiles fired from the Multi-Mortar.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_Mortar extends SentinelProjectile
	abstract;

var() const StaticMeshComponent Mesh;
var() const LightEnvironmentComponent LightEnvironment;
var() SentinelProjectileComponent_Physics PhysicsComponent;
var() const AudioComponent WhistleSound;

var localized string FriendlyName;
var localized string Description;

/** When the shell starts falling down at a speed greater than this, the whistle sound triggers. */
var() float WhistleTriggerSpeed;

enum ETargetType
{
	TT_Infantry,
	TT_LightVehicle,
	TT_HeavyVehicle,
	TT_Sentinel
};

struct EffectivenessInfo
{
	var() ETargetType TargetType;
	/** A number representing how strong this shell is against the target type. 0.0 = average. */
	var() float Effectiveness;
};

/** Holds information on how strong this shell is against various target types. */
var() array<EffectivenessInfo> Effectiveness;

simulated function PostBeginPlay()
{
	super.PostBeginPlay();
}

function Init(Vector Direction)
{
	SetRotation(Rotator(Direction));
	Velocity = Direction;
}

simulated function PhysicsVolumeChange(PhysicsVolume NewVolume)
{
	super.PhysicsVolumeChange(NewVolume);

	PhysicsComponent.PhysicsVolumeChange(NewVolume);
}

simulated function TurnOffAmbientSound()
{
	super.TurnOffAmbientSound();

	WhistleSound.Stop();
}

/**
 * Called when the shell passes the apex of its trajectory.
 */
simulated function Whistle()
{
	WhistleSound.Play();
}

state WaitingToWhistle
{
	simulated function Tick(float DeltaTime)
	{
		super.Tick(DeltaTime);

		if(Velocity.Z < WhistleTriggerSpeed)
		{
			Whistle();
			GotoState('');
		}
	}
}

defaultproperties
{
	WhistleTriggerSpeed=-50.0

	Effectiveness[0]=(TargetType=TT_Infantry,Effectiveness=0.0)
	Effectiveness[1]=(TargetType=TT_LightVehicle,Effectiveness=0.0)
	Effectiveness[2]=(TargetType=TT_HeavyVehicle,Effectiveness=0.0)
	Effectiveness[3]=(TargetType=TT_Sentinel,Effectiveness=0.0)

	Speed=2500
	MaxSpeed=5000.0
	TerminalVelocity=5000.0
	LifeSpan=10.0
	bCollideWorld=true
	Physics=PHYS_Falling
	CheckRadius=36.0

	ProjFlightTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_SentinelMortar_Trail'

	bWaitForEffects=true
	bAttachExplosionToVehicles=false
	bRotationFollowsVelocity=true

	InitialState=WaitingToWhistle

	Begin Object Class=DynamicLightEnvironmentComponent Name=LightEnvironment0
 		bCastShadows=false
		bDynamic=true
		AmbientGlow=(R=0.2,G=0.2,B=0.2)
		MinTimeBetweenFullUpdates=0.2
 	End Object
 	LightEnvironment=LightEnvironment0
 	Components.Add(LightEnvironment0)

	Begin Object Class=StaticMeshComponent Name=ProjectileMesh
		StaticMesh=StaticMesh'VH_Goliath.Mesh.S_VH_Goliath_Tankshell'
		LightEnvironment=LightEnvironment0
		CullDistance=12000
		CastShadow=false
		bUseAsOccluder=false
		BlockRigidBody=false
		BlockNonZeroExtent=false
		BlockZeroExtent=false
		BlockActors=false
		CollideActors=false
	End Object
	Mesh=ProjectileMesh
	Components.Add(ProjectileMesh)

	Begin Object Class=SentinelProjectileComponent_Physics Name=PhysicsComponent0
		EnterWaterDamping=0.8
		WaterAccelRate=0.0
		WaterMaxSpeed=700.0
		WaterEntrySound=SoundCue'Sentinel_Resources.Sounds.Weapons.RocketSplashCue'
		WaterEntryEffect=ParticleSystem'Sentinel_Resources.Effects.Misc.P_Splash'
		WaterExitSound=SoundCue'Sentinel_Resources.Sounds.Weapons.RocketSplashCue'
		WaterExitEffect=ParticleSystem'Sentinel_Resources.Effects.Misc.P_Splash'
	End Object
	PhysicsComponent=PhysicsComponent0

	Begin Object Class=AudioComponent Name=WhistleSound0
		SoundCue=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_WhistleCue'
		bStopWhenOwnerDestroyed=true
	End Object
	WhistleSound=WhistleSound0
	Components.Add(WhistleSound0)
}