//=============================================================================
// A slightly heat-seeking shell that does a lot of damage to vehicles.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_Mortar_AT extends SentinelProjectile_Mortar;

/** Shell cannot seek any vehicle further than this. */
var() float SeekRange;
/** Shell cannot seek any vehicle more than this many degrees away from facing direction. */
var() float SeekAngle;
/** Rate of heading change (actual turn speed is this much when falling vertically, decreasing to zero as angle approaches horizontal). */
var() float TurnSpeed;
/** Time between target seeking updates. */
var() float SeekUpdatePeriod;

simulated function Whistle()
{
	super.Whistle();

	SetTimer(SeekUpdatePeriod, true, 'RecalculateTrajectory');
}

simulated function RecalculateTrajectory()
{
	local Vehicle V;
	local Vector NewVelocity;
	local float CurrentTurnSpeed;

	V = FindClosestVehicle();

	if(V != none)
	{
		CurrentTurnSpeed = TurnSpeed;
		CurrentTurnSpeed *= FMax(Vector(Rotation) Dot vect(0.0, 0.0, -1.0), 0.0);

		NewVelocity = (Normal(V.Location - Location) * CurrentTurnSpeed) + (Normal(Velocity) * (1.0 - CurrentTurnSpeed));
		NewVelocity *= VSize(Velocity);

		Velocity = NewVelocity;
	}
}

/**
 * Returns the closest vehicle within SeekRange and SeekAngle, lower than the current height, and visible.
 */
simulated function Vehicle FindClosestVehicle()
{
	local Vehicle V;
	local float Distance;
	local Vehicle ClosestVehicle;
	local float ClosestDistance;

	foreach WorldInfo.AllPawns(class'Vehicle', V, Location, SeekRange)
	{
		//Only seek vehicles being driven. Undriven vehicles are assumed to not give out enough heat to detect.
		//bPathfindsAsVehicle is false for the Hoverboard, which shouldn't be sought. With any luck it will also be true for any similar custom vehicles which also don't behave like normal vehicles.
		if(!V.bDriving || !V.bPathfindsAsVehicle)
		{
			continue;
		}

		if(RDiff(Rotation, Rotator(V.Location - Location)) > SeekAngle || V.Location.Z >= Location.Z)
		{
			continue;
		}

		if(!FastTrace(V.Location, Location, vect(0.0, 0.0, 0.0), true))
		{
			continue;
		}

		if(!V.IsValidTargetFor(InstigatorController) || Instigator.IsSameTeam(V))
		{
			continue;
		}

		Distance = VSize(V.Location - Location);

		if(ClosestVehicle == none || Distance < ClosestDistance)
		{
			ClosestDistance = Distance;
			ClosestVehicle = V;
		}
	}

	return ClosestVehicle;
}

simulated function Shutdown()
{
	ClearTimer('RecalculateTrajectory');

	super.Shutdown();
}

defaultproperties
{
	FriendlyName="Anti-tank"
	Description="The shaped-charge in this shell is highly effective at penetrating even the most advanced vehicle armour. The relatively small size of the payload leaves enough room for a simple infra-red sensor and control surfaces to increase the probability of a direct hit on a vehicle, as well as significantly increasing range."

	Effectiveness[0]=(TargetType=TT_Infantry,Effectiveness=-1.0)
	Effectiveness[1]=(TargetType=TT_LightVehicle,Effectiveness=0.0)
	Effectiveness[2]=(TargetType=TT_HeavyVehicle,Effectiveness=1.0)
	Effectiveness[3]=(TargetType=TT_Sentinel,Effectiveness=0.0)

	SeekRange=10000.0
	SeekAngle=45.0
	TurnSpeed=0.9
	SeekUpdatePeriod=0.09

	ProjExplosionTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_SentinelMortar_AT_Explosion'
	ExplosionLightClass=class'UTGame.UTRocketExplosionLight'

	Damage=150.0
	DamageRadius=100
	MomentumTransfer=100000
	MyDamageType=class'UTDmgType_SentinelMortar_AT'

	ExplosionSound=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_AT_ExplodeCue'
	ExplosionDecal=MaterialInstanceTimeVarying'WP_RocketLauncher.Decals.MITV_WP_RocketLauncher_Impact_Decal01'
	DecalWidth=180.0
	DecalHeight=180.0

	bNetTemporary=false
	bUpdateSimulatedPosition=true

	Begin Object Name=ProjectileMesh
		StaticMesh=StaticMesh'Sentinel_Resources.Meshes.Weapons.S_MortarShell_AT'
		Materials[0]=MaterialInstanceConstant'Sentinel_Resources.Materials.Weapons.MI_MortarShell_AT'
		Scale=0.7
	End Object
}