//=============================================================================
// Explodes once it reaches a certain altitude, spraying shrapnel around.
// Will also explode there is an obstacle the same distance in front of the shell.
// Damage is caused to everything in front of the projectile, the visible bits
// of shrapnel are only for effect.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_Mortar_Shrapnel extends SentinelProjectile_Mortar;

var() SentinelWeaponComponent_FireInfo ShrapnelFireInfo;
var() SentinelWeaponComponent_HitEffects HitEffects;

/** Height above ground at which to detonate. */
var() float AirburstAltitude;

/** Quantity of shrapnel sprayed when exploding. Should be divisible by ShrapnelDivisions, or the actual quantity will differ from this number. */
var() int NumShrapnel;
/** Shrapnel is divided into this many groups. A higher number means the shrapnel will be spread more evenly, but a number too high relative to NumShrapnel may cause the opposite effect. */
var() int ShrapnelDivisions;
/** Shrapnel will be sprayed within this angle. */
var() int ShrapnelAngle;
/** Simulated speed of shrapnel. Should be very fast to avoid players noticing that damage is instant even though effects are delayed. Should also match speed of shrapnel beam-emitter trail. */
var() float ShrapnelSpeed;

var float ExplodeTime;
var array<Vector> HitLocations;

simulated function Tick(float DeltaTime)
{
	super.Tick(DeltaTime);

	if(!bShuttingDown)
	{
		if(!FastTrace(Location - (vect(0.0, 0.0, 1.0) * AirburstAltitude), Location,, true) || !FastTrace(Location + (Vector(Rotation) * AirburstAltitude), Location,, true))
		{
			Explode(Location, Normal(Velocity));
		}
	}
}

simulated function Explode(Vector HitLocation, Vector HitNormal)
{
	if(WorldInfo.NetMode != NM_DedicatedServer)
	{
		GotoState('SpawningShrapnelEffects');
	}

	super.Explode(HitLocation, HitNormal);

	if(Role == ROLE_Authority)
	{
		HurtCone();
	}
}

/**
 * Causes damage to all actors in a cone in front of this projectile.
 */
function HurtCone()
{
	local Actor	Victim;
	local float AngularDistance;
	local float Distance;
	local float DamageScale;
	local float ActualDamage;
	local float ActualMomentum;

	foreach VisibleCollidingActors(class'Actor', Victim, ShrapnelFireInfo.MaxRange, Location)
	{
		AngularDistance = Vector(Rotation) Dot Normal(Victim.Location - Location);
		AngularDistance = Acos(AngularDistance);
		AngularDistance = (32768.0 * AngularDistance) / PI;

		if(AngularDistance <= ShrapnelAngle)
		{
			Distance = VSize(Victim.Location - Location);

			if(!Victim.bWorldGeometry && Victim != self && (Victim.bProjTarget || (NavigationPoint(Victim) == none)))
			{
				if(Pawn(Victim) == none || !Instigator.IsSameTeam(Pawn(Victim)))
				{
					DamageScale = 1.0 - (Square(AngularDistance / ShrapnelAngle) * 0.5);
					DamageScale *= 1.0 - Square(Distance / ShrapnelFireInfo.MaxRange);

					ActualDamage = FMax(1.0, ShrapnelFireInfo.Damage * DamageScale);
					ActualMomentum = ShrapnelFireInfo.Momentum * DamageScale;

					Victim.TakeDamage(ActualDamage, InstigatorController, Victim.Location, ActualMomentum * Normal(Victim.Location - Location), ShrapnelFireInfo.DamageType,, self);
					Sentinel(Instigator).NotifyCausedDamage(Victim);
				}
			}
		}
	}
}

simulated function bool EffectIsRelevant(Vector SpawnLocation, bool bForceDedicated, optional float CullDistance)
{
	return Instigator.EffectIsRelevant(SpawnLocation, bForceDedicated, CullDistance);
}

/**
 * Trail effects for the shrapnel are spawned immediately. The beam emitter used for the trail has a finite speed set so that it looks like the shrapnel is a real projectile.
 * To further this illusion, hit effects are delayed by an amount equal to the travel time for a projectile moving at the speed of the trail.
 */
state SpawningShrapnelEffects
{
	simulated function BeginState(name PreviousStateName)
	{
		GenerateHitLocations();
		ExplodeTime = WorldInfo.TimeSeconds;
		SpawnTrails();
	}

	simulated function Tick(float DeltaTime)
	{
		super.Tick(DeltaTime);

		if(!SpawnHitEffects())
		{
			GotoState('');
		}
	}

	simulated function GenerateHitLocations()
	{
		local int ShrapnelPerDivision;
		local int CurrentRoll;
		local int CurrentPitch;
		local Rotator ShrapnelRotation;
		local Vector TraceStart, TraceEnd;
		local Actor HitActor;
		local Vector HitLocation, HitNormal;
		local int i;

		ShrapnelPerDivision = NumShrapnel / ShrapnelDivisions;
		TraceStart = Location - Normal(Velocity);

		for(CurrentRoll = 0; CurrentRoll < 65536; CurrentRoll += 65536 / ShrapnelDivisions)
		{
			for(CurrentPitch = 0; CurrentPitch < ShrapnelAngle; CurrentPitch += ShrapnelAngle / ShrapnelPerDivision)
			{
				ShrapnelRotation = class'SUtils'.static.RotateRelative(Rotator(Velocity), 0.0, 0.0, CurrentRoll + (RandRange(-32768.0, 32768.0) / ShrapnelDivisions));
				ShrapnelRotation = class'SUtils'.static.RotateRelative(ShrapnelRotation, CurrentPitch + (RandRange(-ShrapnelAngle, ShrapnelAngle) / (2.0 * ShrapnelPerDivision)), 0.0, 0.0);
				TraceEnd = TraceStart + (Vector(ShrapnelRotation) * ShrapnelFireInfo.MaxRange);
				HitActor = Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true, vect(0.0, 0.0, 0.0),, TRACEFLAG_Bullet);

				if(HitActor == none)
				{
					HitLocation = TraceEnd;
				}

				HitLocations[i++] = HitLocation;

				if(i >= NumShrapnel)
				{
					break;
				}
			}
		}

		RandomizeHitLocations();
	}

	/**
	 * Randomizes the indices of the HitLocations array. Necessary because SpawnTrails may use only the first few of the array, then stop due to performance limits. If the array is not random, then the trails that are spawned will all be in similar directions instead of filling the whole cone.
	 */
	simulated function RandomizeHitLocations()
	{
		local array<Vector> TempHitLocations;
		local int i;

		TempHitLocations = HitLocations;
		HitLocations.length = 0;

		while(TempHitLocations.length > 0)
		{
			i = Rand(TempHitLocations.length - 1);
			HitLocations.AddItem(TempHitLocations[i]);
			TempHitLocations.Remove(i, 1);
		}
	}

	/**
	 * Spawns a beam-emitter trail for all shrapnel.
	 */
	simulated function SpawnTrails()
	{
		local int i;
		local ParticleSystemComponent PSC;

		for(i = 0; i < NumShrapnel; i++)
		{
			PSC = WorldInfo.MyEmitterPool.SpawnEmitter(ParticleSystem'Sentinel_Resources.Effects.Weapons.P_SentinelMortar_Shrapnel_Trail', Location, Rotation);

			if(PSC != none)
			{
				AttachComponent(PSC);
				PSC.SetVectorParameter('TrailEnd', HitLocations[i]);
			}
		}
	}

	/**
	 * Spawns a hit effect for any of HitLocations that are close enough for the shrapnel travelling at ShrapnelSpeed to have hit but have not hit already.
	 *
	 * @return	true if any of HitLocations are not close enough to have hit yet, false if all of HitLocations have been dealt with
	 */
	simulated function bool SpawnHitEffects()
	{
		local int i;
		local bool bShrapnelStillToHit;

		for(i = 0; i < NumShrapnel; i++)
		{
			if(!IsZero(HitLocations[i]))
			{
				if(VSize(HitLocations[i] - Location) <= (WorldInfo.TimeSeconds - ExplodeTime) * ShrapnelSpeed)
				{
					HitEffects.PlayImpactEffects(HitLocations[i] - Normal(HitLocations[i] - Location), HitLocations[i], self);
					HitLocations[i] = vect(0.0, 0.0, 0.0);
				}
				else
				{
					bShrapnelStillToHit = true;
				}
			}
		}

		return bShrapnelStillToHit;
	}
}

defaultproperties
{
	FriendlyName="Shrapnel"
	Description="This shell contains many small projectiles which are sprayed forward when the shell drops to a preset altitude. It is highly effective against infantry, but the light shrapnel is unable to penetrate vehicle armour."

	Effectiveness[0]=(TargetType=TT_Infantry,Effectiveness=1.0)
	Effectiveness[1]=(TargetType=TT_LightVehicle,Effectiveness=-0.8)
	Effectiveness[2]=(TargetType=TT_HeavyVehicle,Effectiveness=-1.6)
	Effectiveness[3]=(TargetType=TT_Sentinel,Effectiveness=-0.8)

	AirburstAltitude=1000.0
	NumShrapnel=64
	ShrapnelDivisions=8
	ShrapnelAngle=8000
	ShrapnelSpeed=5000.0

	ProjExplosionTemplate=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_SentinelMortar_Shrapnel_Explosion'
	ExplosionLightClass=class'UTGame.UTRocketExplosionLight'

	Damage=100.0
	DamageRadius=200
	MomentumTransfer=20000
	MyDamageType=class'UTDmgType_SentinelMortar_ShrapnelShell'

	ExplosionSound=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_Shrapnel_ExplodeCue'
	ExplosionDecal=MaterialInstanceTimeVarying'WP_RocketLauncher.Decals.MITV_WP_RocketLauncher_Impact_Decal01'
	DecalWidth=128.0
	DecalHeight=128.0

	Begin Object Name=ProjectileMesh
		StaticMesh=StaticMesh'Sentinel_Resources.Meshes.Weapons.S_MortarShell_Shrapnel'
		Materials[0]=MaterialInstanceConstant'Sentinel_Resources.Materials.Weapons.MI_MortarShell_Shrapnel'
		Scale=0.7
		Scale3D=(X=1.0,Y=0.9,Z=0.9)
	End Object

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		Damage=200.0
		Momentum=5000.0
		DamageType=class'UTDmgType_SentinelMortar_Shrapnel'
		MaxRange=2000.0
	End Object
	ShrapnelFireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_HitEffects Name=HitEffectsComp
	End Object
	HitEffects=HitEffectsComp
}