//=============================================================================
// A rocket with various target seeking capabilities.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelProjectile_Rocket extends SentinelProjectile_Rocket_Base;

enum ESeekType
{
	SEEK_Location, //Home in on a location. Only valid if a TargetLocation has been set, or there is a TargetActor.
	SEEK_Direct, //Home in on target actor's location. Only valid if there is a TargetActor.
	SEEK_WithinCollision, //Home in on a spot within the target actor's collision cylinder. Only valid if there is a TargetActor.
	SEEK_None //Don't seek anything.
};
/** Current method of homing in on target. */
var() ESeekType SeekType;

enum EWobbleType
{
	WOBBLE_Diverge, //Increase wobble on approaching target. Only valid if a TargetLocation has been set, or there is a TargetActor.
	WOBBLE_Converge, //High wobble before reaching target, then reduce wobble as distance decreases. Only valid if a TargetLocation has been set, or there is a TargetActor.
	WOBBLE_Constant, //Constant randomness.
	WOBBLE_None //Straight line.
};

struct WobbleStruct
{
	/** Type of wobble this is for. */
	var() EWobbleType WobbleType;
	/** Base wobble amount outside of start/stop dist. */
	var() float WobbleBaseAmount;
	/** Wobble when at WobbelStartDist. */
	var() float WobbleStartAmount;
	/** Wobble when at WobbleStopDist */
	var() float WobbleStopAmount;
	/** Distance from target at which to start interpolating wobble amount. */
	var() float WobbleStartDistance;
	/** Distance from target at which to stop interpolating wobble amount. */
	var() float WobbleStopDistance;
};

/** Ways to wobble for each wobble type. */
var() array<WobbleStruct> WobbleInfos;
/** Wibble wobble. */
var() WobbleStruct CurrentWobbleInfo;

/** If more than this distance from the target, then try not to hit things. */
var() float MinCollisionAvoidDistance;
/** Look this much farther ahead than the next update when avoiding collision. */
var() float CollisionAvoidSafetyMargin;

/** Class to switch to when rocket motor runs out. */
var() class<SentinelProjectile> ExtinguishedClass;
/** Initial and maximum speed will vary randomly by up to +/- this fraction. */
var() float SpeedVariation;
/** Manoeuvrability. */
var() float TurnSpeed; //1 -> very agile, 0 -> not a lot.

/** Time between target seeking updates. */
var() float SeekUpdatePeriod;
/** Time that rocket motor lasts for. */
var() float FlightTime;

/** Actor to move towards. */
var Actor TargetActor;
/** Location to move towards. */
var Vector TargetLocation;
/** Specific location relative to target that rocket aims for. */
var Vector AimSpot;
/** Distance from target within which rockets seek target's origin instead of AimSpot. */
var float TargetDistanceThreshold;
/** Velocity changes towards this over time. */
var Vector DesiredVelocity;

replication
{
	if(Role == ROLE_Authority && bNetInitial)
		TurnSpeed;

	if(Role == ROLE_Authority && bNetDirty)
		DesiredVelocity;
}

simulated function PostBeginPlay()
{
	super.PostBeginPlay();

	SetTimer(FlightTime, false, 'EndFlight');
}

function Init(Vector Direction)
{
	Speed *= 1.0 + SpeedVariation - (SpeedVariation * 2.0 * FRand());
	MaxSpeed *= 1.0 + SpeedVariation - (SpeedVariation * 2.0 * FRand());

	DesiredVelocity = Speed * Direction;

	super.Init(Direction);
}

simulated function PhysicsVolumeChange(PhysicsVolume NewVolume)
{
	super.PhysicsVolumeChange(NewVolume);

	if(NewVolume.bWaterVolume)
	{
		EndFlight();
	}
}

simulated function Shutdown()
{
	ClearTimer('EndFlight');
	ClearTimer('RecalculateTrajectory');

	super.Shutdown();
}

function SetTarget(Actor NewTarget, Vector NewTargetLocation, ESeekType NewSeekType=SEEK_Direct, EWobbleType NewWobbleType=WOBBLE_Constant)
{
	local int WobbleIndex;

	TargetActor = NewTarget;
	TargetLocation = NewTargetLocation;
	SeekType=NewSeekType;

	WobbleIndex = WobbleInfos.Find('WobbleType', NewWobbleType);

	if(WobbleIndex == INDEX_NONE)
	{
		`warn("Tried to set a target without a valid wobble type:"@NewWobbleType);
	}
	else
	{
		CurrentWobbleInfo = WobbleInfos[WobbleIndex];
	}

	if(TargetActor != none)
	{
		if(SeekType == SEEK_WithinCollision)
		{
			AimSpot = LocationWithinCollision(TargetActor, TargetDistanceThreshold);
		}
	}

	SetTimer(SeekUpdatePeriod, true, 'RecalculateTrajectory');
}

/**
 * Finds a random location relative to target's origin within its bounding cylinder.
 */
function Vector LocationWithinCollision(Actor A, out float CollisionSize)
{
	local float Radius, Height;
	local Vector V;
	
	A.GetBoundingCylinder(Radius, Height);
	CollisionSize = Radius + Height;
	V.X = (FRand() * Radius * 2.0) - Radius;
	V.Y = (FRand() * Radius * 2.0) - Radius;
	V.Z = (FRand() * Height * 2.0) - Height;

	return V;
}

/**
 * Shut down and spawn a new unpowered projectile, to give the appearance of running out of fuel.
 */
simulated function EndFlight()
{
	local SentinelProjectile P;

	if(Role == ROLE_Authority)
	{
		//Can't just make self fall because the trail emitter will keep following, so hide self and spawn a new falling projectile instead.
		P = Spawn(ExtinguishedClass);

		if(P != none)
		{
			P.Speed = VSize(Velocity);
			P.Init(Normal(Velocity));
			bSuppressExplosionFX = true;
		}
	}

	Shutdown();
}

simulated function Tick(float DeltaTime)
{
	local float CurrentSpeed;

	super.Tick(DeltaTime);

	CurrentSpeed = VSize(Velocity);
	Velocity *= 1.0 - (DeltaTime / SeekUpdatePeriod);
	Velocity += DesiredVelocity * (DeltaTime / SeekUpdatePeriod);
	Velocity = CurrentSpeed * Normal(Velocity);
}

/**
 * Calculates a new random wobble vector based on the current wobble type and the given distance from the target.
 */
function Vector CalculateWobble(float Distance)
{
	local float WobbleAmount;

	if(Distance > CurrentWobbleInfo.WobbleStartDistance || Distance <= CurrentWobbleInfo.WobbleStopDistance)
	{
		WobbleAmount = CurrentWobbleInfo.WobbleBaseAmount;
	}
	else
	{
		WobbleAmount = (Distance - CurrentWobbleInfo.WobbleStopDistance) / (CurrentWobbleInfo.WobbleStartDistance - CurrentWobbleInfo.WobbleStopDistance);
		WobbleAmount *= CurrentWobbleInfo.WobbleStopAmount - CurrentWobbleInfo.WobbleStartAmount;
		WobbleAmount += CurrentWobbleInfo.WobbleStopAmount;
	}

	return VRand() * WobbleAmount;
}

/**
 * Adjusts NewDirection in an attempt to avoid possible collisions with world geometry.
 * How succesful this is depends on TurnSpeed and CollisionAvoidSafetyMargin.
 */
function AvoidCollision(out Vector NewDirection, float CurrentSpeed)
{
	local Vector TraceStart, TraceEnd;
	local Vector HitLocation, HitNormal;
	local Actor HitActor;
	local Vector HitPerpendicular;

	TraceStart = Location;
	TraceEnd = Location + (NewDirection * CurrentSpeed * SeekUpdatePeriod * CollisionAvoidSafetyMargin);
	HitActor = Trace(HitLocation, HitNormal, TraceEnd, TraceStart);

	if(HitActor != none && HitActor.bWorldGeometry)
	{
		HitPerpendicular = HitNormal Cross NewDirection;
		NewDirection = HitPerpendicular Cross HitNormal;
	}
}

/**
 * Calculates a new heading based on current behaviour and situation.
 */
function RecalculateTrajectory()
{
	local float Distance;
	local Vector NewDirection;
	local float CurrentSpeed;

	if(SeekType == SEEK_None)
	{
		//If not homing in on something, then rockets go vaguely in the direction that they happen to be pointing in.
		TargetLocation = Location + (Normal(Velocity) * CurrentWobbleInfo.WobbleStartDistance);
	}
	else if(TargetActor != none) 
	{
		if(SeekType == SEEK_WithinCollision)
		{
			TargetLocation = TargetActor.GetTargetLocation();

			//If close to target, home in on its origin, otherwise keep going towards AimSpot.
			if(VSize(Location - TargetLocation) > TargetDistanceThreshold)
			{
				TargetLocation += AimSpot >> TargetActor.Rotation;
			}
		}
		else if(SeekType == SEEK_Direct)
		{
			TargetLocation = TargetActor.GetTargetLocation();
		}
	}

	Distance = VSize(Location - TargetLocation);
	CurrentSpeed = VSize(Velocity);

	NewDirection = Normal(TargetLocation - Location);
	NewDirection += CalculateWobble(Distance);
	NewDirection = Normal(NewDirection);

	if(Distance > MinCollisionAvoidDistance)
	{
		AvoidCollision(NewDirection, CurrentSpeed);
	}

	NewDirection = (NewDirection * TurnSpeed) + (Normal(Velocity) * (1.0 - TurnSpeed));

	DesiredVelocity = CurrentSpeed * NewDirection;
	Acceleration = AccelRate * NewDirection;

	bForceNetUpdate = true;
}

defaultproperties
{
	SeekType=SEEK_None

	WobbleInfos[0]=(WobbleType=WOBBLE_Diverge,WobbleBaseAmount=0.05,WobbleStartAmount=0.05,WobbleStopAmount=0.4,WobbleStartDistance=1000.0,WobbleStopDistance=0.0)
	WobbleInfos[1]=(WobbleType=WOBBLE_Converge,WobbleBaseAmount=0.1,WobbleStartAmount=0.5,WobbleStopAmount=0.0,WobbleStartDistance=3000.0,WobbleStopDistance=300.0)
	WobbleInfos[2]=(WobbleType=WOBBLE_Constant,WobbleBaseAmount=0.1)
	WobbleInfos[3]=(WobbleType=WOBBLE_None)

	TurnSpeed=0.4
	SeekUpdatePeriod=0.08

	MinCollisionAvoidDistance=500.0
	CollisionAvoidSafetyMargin=2.0
	FlightTime=7.0
	ExtinguishedClass=class'SentinelProjectile_Rocket_Falling'

	SpeedVariation=0.02
	bRotationFollowsVelocity=true

	ProjectileLightClass=class'ProjectileLight_Rocket'
	bProjectileLight=false

	LifeSpan=9.0
	bNetTemporary=false
	NetUpdateFrequency=10.0 //Updates are requested manually every direction change, so don't need to be fast here.

	Begin Object Name=PhysicsComponent0
		WaterEntryEffect=ParticleSystem'Sentinel_Resources.Effects.Misc.P_Steam-Splash'
	End Object
}
