//=============================================================================
// Makes the Sentinel invisible whilst not firing.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_Cloak extends SentinelUpgrade;

/** Rate at which cloak changes between off and fully on. */
var() float CloakingRate;
/** Time for which the cloak is deactivated upon the Sentinel firing. */
var() float DecloakTime;
/** Name of the material parameter that controls visibility. */
var() name VisibilityParamName;

/** Set true when invisible, for replication purposes. */
var repnotify bool bInvisible;
var float Visibility;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		bInvisible;

	//Zero visibility replicated to prevent re-cloaking every time this upgrade becomes relevant.
	if(Role == ROLE_Authority && bNetDirty && Visibility == 0.0)
		Visibility;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'bInvisible')
	{
		if(bInvisible)
		{
			Cloak();
		}
		else
		{
			DeCloak();
		}
	}

	super.ReplicatedEvent(VarName);
}

function InitializeFor(Sentinel S)
{
	super.InitializeFor(S);

	Cloak();
}

function AddUpgrade(SentinelUpgrade NewUpgrade)
{
	super.AddUpgrade(NewUpgrade);

	if(NewUpgrade.IsA('SentinelWeapon'))
	{
		Decloak();
		SetTimer(DecloakTime, false, 'Cloak');
	}
}

function NotifyFired()
{
	super.NotifyFired();

	Decloak();
	SetTimer(DecloakTime, false, 'Cloak');
}

simulated function Cloak()
{
	if(Role == ROLE_Authority)
	{
		bInvisible = true;
	}

	GotoState('Cloaking');
}

simulated function Decloak()
{
	if(Role == ROLE_Authority)
	{
		bInvisible = false;
	}
	
	Visibility = 1.0;

	if(Cannon != none)
	{
		Cannon.bIsInvisible = false;
		Cannon.bAmbientCreature = Cannon.default.bAmbientCreature;
		Cannon.SWeapon.NotifyInvisible(false);

		Cannon.BaseComponent.SetVisibility(Visibility);
		Cannon.RotatorComponent.SetVisibility(Visibility);
		Cannon.WeaponComponent.SetVisibility(Visibility);
	}

	GotoState('');
}

function OnDetach()
{
	Decloak();

	super.OnDetach();
}

simulated state Cloaking
{
	simulated function Tick(float DeltaTime)
	{
		Visibility = FMax(Visibility - (DeltaTime * CloakingRate), 0.0);

		if(Cannon != none)
		{
			Cannon.BaseComponent.SetVisibility(Visibility);
			Cannon.RotatorComponent.SetVisibility(Visibility);
			Cannon.WeaponComponent.SetVisibility(Visibility);

			if(Visibility <= 0.5 && !Cannon.bIsInvisible)
			{
				Cannon.bIsInvisible = true;
				Cannon.bAmbientCreature = true; //Hack to prevent bots from seeing invisible Sentinels.
				Cannon.SWeapon.NotifyInvisible(true);
			}

			if(Visibility <= 0.0)
			{
				GotoState('');
			}
		}
	}
}

defaultproperties
{
	FriendlyName="Cloaking Device"
	Description="A metamaterial cloaking device which refracts light around the Sentinel, effectively making it almost totally transparent.\n\nBecause it interferes with the active scanning required for accurate targeting, the cloaking device is automatically disabled when the Sentinel needs to fire."

	CloakingRate=0.3
	DecloakTime=2.0

	Visibility=1.0

	AmmoCost=80
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.InvisibleCue'

	ForbiddenUpgrades.Add(class'SentinelUpgrade_Shield')
}
