//=============================================================================
// Communicates with other networked Sentinels.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_Network extends SentinelUpgrade;

/** Time between repeat broadcasts to peers. */
var() float UpdatePeriod;

/** Pawn most recently broadcast over the network. */
var Pawn LastBroadcastEnemy;
/** Keeps track of the number of Sentinels connected to this one. */
var int NumPeers;
/** Set to true while receiving something from the network, to prevent recursion. */
var private bool bReceiving;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		NumPeers;
}

function NotifyNewTarget(Actor NewTarget)
{
	LastBroadcastEnemy = Pawn(NewTarget);

	//Only alert Sentinels of mobile pawns.
	if(LastBroadcastEnemy != none && !NewTarget.IsStationary())
	{
		BroadcastEnemy(LastBroadcastEnemy, Cannon.TargetName, Cannon.SController.bEnemyIsVisible);
		SetTimer(UpdatePeriod, true, 'RepeatBroadcast');
	}
	else
	{
		ClearTimer('RepeatBroadcast');
	}

	super.NotifyNewTarget(NewTarget);
}

function NotifyWaiting()
{
	ClearTimer('RepeatBroadcast');

	super.NotifyWaiting();
}

function RepeatBroadcast()
{
	if(LastBroadcastEnemy != none)
	{
		BroadcastEnemy(LastBroadcastEnemy, Cannon.TargetName, Cannon.SController.bEnemyIsVisible);
	}
	else
	{
		ClearTimer('RepeatBroadcast');
	}
}

/**
 * Broadcasts to all connected Sentinels. Also sets NumPeers to the number of network-equipped Sentinels found.
 */
function BroadcastEnemy(Pawn NewEnemy, string NewEnemyName, bool bNewEnemyVisible)
{
	local Sentinel S;
	local SentinelUpgrade_Network Node;

	NumPeers = 0;

	foreach WorldInfo.AllPawns(class'Sentinel', S)
	{
		if(S != Cannon && S.SController != none && S.IsSameTeam(Cannon))
		{
			Node = SentinelUpgrade_Network(S.UpgradeManager.GetUpgrade(Class));

			if(Node != none)
			{
				NumPeers++;

				//If something is being received then there is no point in re-broadcasting it, but still want to find all connected Sentinels so NumPeers can be updated.
				if(!bReceiving)
				{
					Node.ReceiveEnemy(NewEnemy, NewEnemyName, bNewEnemyVisible);
				}
			}
		}
	}
}

/**
 * Called by BroadcastEnemy. Notifies the owning Sentinel of the potential new enemy.
 */
function ReceiveEnemy(Pawn NewEnemy, string NewEnemyName, bool bNewEnemyVisible)
{
	bReceiving = true;

	if(Cannon.SController.Enemy != NewEnemy || !Cannon.SController.bEnemyIsVisible)
	{
		if(bNewEnemyVisible && Cannon.SController.LineOfSightTo(NewEnemy))
		{
			Cannon.SController.SeePlayer(NewEnemy);
		}
		else
		{
			Cannon.SController.HearNoise(1.0, NewEnemy);
		}

		//Ensure notified Sentinels know what they are targeting, for display purposes.
		if(Cannon.Target == NewEnemy && NewEnemyName != "")
		{
			Cannon.TargetName = NewEnemyName;
		}
	}

	bReceiving = false;
}

//Show how many Sentinels are networked to this one.
simulated function GetStatusText(out array<String> StatusStrings)
{
	local String StatusLine;

	StatusLine = class'UIStrings'.default.Peers$": ";
	class'SUtils'.static.PadWithSpaces(StatusLine, 15);

	if(NumPeers < 1)
		StatusLine $= class'SUtils'.static.TextColourMarkup(1.0, 1.0, 0.0);
	else
		StatusLine $= class'SUtils'.static.TextColourMarkup(0.0, 1.0, 0.0);

	StatusLine $= NumPeers;
	StatusLine $= class'SUtils'.static.TextColourMarkup(1.0, 1.0, 1.0);
	StatusStrings[StatusStrings.Length] = StatusLine;

	super.GetStatusText(StatusStrings);
}

defaultproperties
{
	FriendlyName="Mesh Network Interface"
	Description="Radio network link allowing Sentinels to form an ad-hoc network to coordinate their efforts. Quantum encryption ensures that hacking the network is as close to impossible as is theoretically possible.\n\nSentinel networks are most effective when one Sentinel is positioned and equipped to view a wide area from where it can alert slower Sentinels to bring the bulk of the collective firepower to bear on a target."

	UpdatePeriod=1.21

	AmmoCost=40
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.NetworkActivateCue'
}
