//=============================================================================
// Allows Sentinel to target shootable objectives and flag carriers.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_ObjectiveSoftware extends SentinelUpgrade;

/** How important it should be to attack flag carriers. */
var() float FlagCarrierWeight;
/** How frequently to check for nearby objectives to attack. */
var() float ObjectiveCheckTime;
/** All objectives found in the map. */
var array<UTGameObjective> Objectives;

function InitializeFor(Sentinel S)
{
	super.InitializeFor(S);

	FindAllObjectives();

	if(Cannon.SController.Enemy == none)
	{
		CheckObjectives();
		SetTimer(ObjectiveCheckTime, true, 'CheckObjectives');
	}
}

//Give priority to attacking flag/orb carriers.
function ModifyPawnTargetPriority(out float Weight, Pawn PawnTarget)
{
	if(PawnTarget.PlayerReplicationInfo.bHasFlag)
		Weight += FlagCarrierWeight;

	super.ModifyPawnTargetPriority(Weight, PawnTarget);
}

function NotifyNewTarget(Actor NewTarget)
{
	super.NotifyNewTarget(NewTarget);

	ClearTimer('CheckObjectives');
}

function NotifyWaiting()
{
	super.NotifyWaiting();

	CheckObjectives();
	SetTimer(ObjectiveCheckTime, true, 'CheckObjectives');
}

/**
 * Finds all objectives in the map.
 */
function FindAllObjectives()
{
	local UTGameObjective Objective;

	foreach WorldInfo.AllNavigationPoints(class'UTGameObjective', Objective)
	{
		Objectives.AddItem(Objective);
	}
}

/**
 * Looks for an objective that can be attacked.
 *
 * @return	the closest highest priority valid objective
 */
function UTGameObjective BestObjective()
{
	local UTGameObjective Objective, BestObjective;

	foreach Objectives(Objective)
	{
		if(!Cannon.SController.ObjectiveTargetingBehaviour.IsValidObjective(Objective) || !Cannon.SController.LineOfSightTo(Objective))
			continue;

		if(BestObjective == none)
		{
			BestObjective = Objective;
		}
		else if(Cannon.SController.ObjectiveTargetingBehaviour.ObjectiveTargetPriority(Objective) > Cannon.SController.ObjectiveTargetingBehaviour.ObjectiveTargetPriority(BestObjective))
		{
			BestObjective = Objective;
		}
	}

	return BestObjective;
}

function CheckObjectives()
{
	local UTGameObjective Objective;

	Objective = BestObjective();

	if(Objective != none && Cannon.SController.Enemy == none)
	{
		Cannon.SController.ObjectiveTargetingBehaviour.Objective = Objective;
		Cannon.SController.ChangeBehaviourTo(Cannon.SController.ObjectiveTargetingBehaviour);
	}
}

static function float BotDesirability(UTBot B, Sentinel S)
{
	local float Desirability;
	local UTGameObjective Objective;
	local bool bNearObjective;

	Desirability = super.BotDesirability(B, S);

	if(Desirability >= 0.0)
	{
		foreach B.WorldInfo.AllNavigationPoints(class'UTGameObjective', Objective)
		{
			if(VSize(Objective.Location - S.Location) < S.GetRange() && S.SController.LineOfSightTo(Objective))
			{
				bNearObjective = true;
				break;
			}
		}

		//Low but not negative if no objectives nearby because the Sentinel may come into range in the future, or there may be carried objects that will pass by.
		if(!bNearObjective)
		{
			Desirability = 0.0;
		}
	}

	return Desirability;
}

defaultproperties
{
	FriendlyName="Objective ID"
	Description="This software imparts the ability to identify and attack static objectives, such as Power Nodes, and preferentially attack enemies carrying objectives."

	FlagCarrierWeight=2.0
	ObjectiveCheckTime=3.3

	MaxDesirability=0.6
	AmmoCost=15
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.beep9Cue'
	RequiredUpgrades.Add(class'SentinelUpgrade_CPU')
}
