//=============================================================================
// Allows adjustment of a Sentinel's idle rotation parameters.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_ScanAngleLimiter extends SentinelUpgrade;

var() const StaticMeshComponent DrawAngleComponent;
var MaterialInstanceConstant DrawAngleMaterialInstance;

var UTUIScene_ScanAngleLimiterMenu ScanAngleLimiterMenu;

var localized string ScanAngleRange;
var localized string ScanAngleCentre;

simulated function ClientInitializeFor()
{
	super.ClientInitializeFor();

	DrawAngleMaterialInstance = DrawAngleComponent.CreateAndSetMaterialInstanceConstant(0);
	Cannon.BaseComponent.AttachComponent(DrawAngleComponent, Cannon.BaseComponent.RootBone);
}

function OnDetach()
{
	super.OnDetach();

	if(DrawAngleComponent != none)
	{
		DrawAngleComponent.SetHidden(true);
	}
}

simulated function OpenConfigurationMenuFor(UTPlayerController UTPC)
{
	local Vector DrawAngleOrigin;

	super.OpenConfigurationMenuFor(UTPC);

	ScanAngleLimiterMenu = UTUIScene_ScanAngleLimiterMenu(ConfigurationMenu);

	if(ScanAngleLimiterMenu != none && DrawAngleComponent != none)
	{
		DrawAngleComponent.SetHidden(false);

		DrawAngleOrigin = Cannon.WeaponComponent.Translation;
		DrawAngleOrigin.Z += Cannon.CylinderComponent.CollisionHeight;
		DrawAngleComponent.SetTranslation(DrawAngleOrigin / Cannon.BaseComponent.Scale);
	}
}

simulated function ConfigurationMenuClosed()
{
	if(DrawAngleComponent != none)
	{
		DrawAngleComponent.SetHidden(true);
	}
}

simulated function ConfigurationMenuDeactivated()
{
	if(DrawAngleComponent != none)
	{
		DrawAngleComponent.SetHidden(true);
	}
}

simulated function SetAutoRotateParams(float NewAutoRotateRange, float NewAutoRotateCentre)
{
	ServerSetAutoRotateParams(NewAutoRotateRange, NewAutoRotateCentre);
}

reliable server function ServerSetAutoRotateParams(float NewAutoRotateRange, float NewAutoRotateCentre)
{
	Cannon.AutoRotateHalfRange = NewAutoRotateRange * 0.5;
	Cannon.AutoRotateCentre = NewAutoRotateCentre;
	Cannon.bAutoRotateYawReversed = false;
}

simulated function Tick(float DeltaTime)
{
	local float AutoRotateRange;
	local Rotator AutoRotateCentre;

	if(DrawAngleComponent != none && !DrawAngleComponent.HiddenGame && Cannon != none && ScanAngleLimiterMenu != none)
	{
		AutoRotateRange = ScanAngleLimiterMenu.GetAutoRotateRange() / 65536.0;
		DrawAngleMaterialInstance.SetScalarParameterValue('RotationAmount', AutoRotateRange);

		AutoRotateCentre.Yaw = ScanAngleLimiterMenu.GetAutoRotateCentre();
		DrawAngleComponent.SetRotation(AutoRotateCentre);
	}
}

defaultproperties
{
	FriendlyName="Scan Angle Limiter"
	Description="This software upgrade exposes the Sentinel's automatic scanning parameters for user configuration, allowing the Sentinel to be instructed to concentrate its search on a specific area instead of the default 360 degree scan."

	ScanAngleRange="Scan Angle Range"
	ScanAngleCentre="Scan Angle Centre"
	ConfigurationMenuTemplate=UTUIScene_ScanAngleLimiterMenu'Sentinel_Resources.Menus.ScanAngleLimiterMenu'

	AmmoCost=15
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.mactiv50Cue'
	RequiredUpgrades.Add(class'SentinelUpgrade_CPU')
	MaxDesirability=-1.0 //Bots don't really have a way to use this.

	Begin Object Class=StaticMeshComponent Name=StaticMeshComponent0
		StaticMesh=StaticMesh'Sentinel_Resources.Meshes.Upgrades.XMirroredSquare'
		Materials[0]=Material'Sentinel_Resources.Materials.Upgrades.M_DrawAutoYawLimits'
		Scale=2.0
		DepthPriorityGroup=SDPG_Foreground //Arrow is drawn on top of the Sentinel so it doesn't become hidden when viewed from some angles. The material makes it look different when viewed through a solid object so it's clear when it's being occluded.
		HiddenGame=true
		CollideActors=false
		BlockActors=false
		BlockZeroExtent=false
		BlockNonZeroExtent=false
		BlockRigidBody=false
		bAcceptsDecals=false
	End Object
	DrawAngleComponent=StaticMeshComponent0
	Components.Add(StaticMeshComponent0)
}
