//=============================================================================
// Sweet revenge.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_SelfDestruct extends SentinelUpgrade;

var Controller InstigatorController;

/** Damage to cause when exploding. */
var() float Damage;
/** Size of explosion when exploding. */
var() float DamageRadius;
/** Momentum of explosion when exploding. */
var() float Momentum;
/** DamageType to inflict when exploding. */
var() class<DamageType> MyDamageType;
/** Effect to spawn when exploding. */
var() ParticleSystem ExplosionTemplate;
/** Light to spawn when exploding */
var() class<UTExplosionLight> ExplosionLightClass;
/** Sound to play when exploding. */
var() SoundCue ExplosionSound;

var() CameraAnim ExplosionShake;

var bool bTriggered;

//ExplosionTemplate etc. need to be set on client too.
simulated function ClientInitializeFor()
{
	Cannon.DiedDamage = Damage;
	Cannon.DiedDamageRadius = DamageRadius;
	Cannon.DiedMomentum = Momentum;
	Cannon.DiedDamageType = MyDamageType;

	if(Cannon.ExplosionEffectComponent != none)
	{
		Cannon.ExplosionEffectComponent.ExplosionTemplate = ExplosionTemplate;
		Cannon.ExplosionEffectComponent.ExplosionLightClass = ExplosionLightClass;
		Cannon.ExplosionEffectComponent.ExplosionSound = ExplosionSound;
		Cannon.ExplosionEffectComponent.DecalWidth = Cannon.ExplosionEffectComponent.default.DecalWidth * 2.0;
		Cannon.ExplosionEffectComponent.DecalHeight = Cannon.ExplosionEffectComponent.default.DecalHeight * 2.0;
	}

	Cannon.DeadLifeSpan = Cannon.BurnTime;

	super.ClientInitializeFor();
}

function SentinelUpgrade DetachUpgrade(SentinelUpgrade UpgradeToDetach)
{
	//Can't remove the self destruct upgrade, instead cause cannon to explode.
	if(UpgradeToDetach == self && !bTriggered)
	{
		bTriggered = true;
		Cannon.Suicide();

		return none;
	}

	return super.DetachUpgrade(UpgradeToDetach);
}

//Remove self from list to avoid the usual destruction, so the explosion can be done after.
function NotifyDied(Controller Killer, out class<DamageType> DamageType, Vector HitLocation)
{
	super.NotifyDied(Killer, DamageType, HitLocation);

	if(Killer != none && Cannon.InstigatorController != none)
	{
		//Make the Sentinel explode, even if it was a suicide (i.e. the owner tried to recycle it).
		if(DamageType == class'DmgType_Suicided')
		{
			DamageType = MyDamageType;
		}

		InstigatorController = Cannon.InstigatorController;
		LifeSpan = Cannon.DeadLifeSpan;
		bTriggered = true;
		Cannon.UpgradeManager.DetachUpgrade(self);
		GotoState('Dying');
	}
}

static function float BotDesirability(UTBot B, Sentinel S)
{
	local float Desirability;

	//Bots should never apply this to human-owned Sentinels, since it could kill the owner.
	Desirability = PlayerController(S.InstigatorController) != none ? -1.0 : super.BotDesirability(B, S);

	return Desirability;
}

//Copied from UTProj_RedeemerBase.ShakeView with some alterations
function ShakeView()
{
	local UTPlayerController PC;
	local float Dist, Scale;

	foreach WorldInfo.AllControllers(class'UTPlayerController', PC)
	{
		if (PC.ViewTarget != None)
		{
			Dist = VSize(Location - PC.ViewTarget.Location);
			if (Dist < DamageRadius * 2.0)
			{
				if (Dist < DamageRadius)
				{
					Scale = 1.0;
				}
				else
				{
					Scale = (DamageRadius * 2.0 - Dist) / DamageRadius;
				}
				PC.ClientPlayCameraAnim(ExplosionShake, Scale);
			}
		}
	}
}

//Copied from UTProj_RedeemerBase.DoKnockdown with some alterations
function DoKnockdown()
{
	local Controller C;
	local vector Dir;
	local float DamageScale;
	local UTPawn P;

	foreach WorldInfo.AllControllers(class'Controller', C)
	{
		P = UTPawn(C.Pawn);
		if (P != None && P.Health > 0 && !P.bIgnoreForces && !P.InGodMode() && !WorldInfo.GRI.OnSameTeam(P, InstigatorController))
		{
			Dir = P.Location - Location;
			DamageScale = FClamp(1.0 - (VSize(Dir) / DamageRadius), 0.0, 1.0);
			if (P.Health < Damage * DamageScale)
			{
				P.AddVelocity(((Momentum * DamageScale) / P.Mass) * 0.5 * Normal(Dir), P.Location, MyDamageType);
				P.ForceRagdoll();
			}
		}
	}
}

//Copied from UTProj_RedeemerBase.RedeemerHurtRadius with some alterations
function SelfDestructHurtRadius(float RadiusFactor)
{
	local Actor Victim, HitActor;
	local bool bCauseDamage;
	local float HurtRadius, Radius, Height;
	local vector Side, Up, ExplosionCenter, HitNormal, HitLocation;

	HurtRadius = DamageRadius * RadiusFactor;

	// move explosion center up a bit to address not hitting pawns
	ExplosionCenter = Location + vect(0,0,200);
	HitActor = Trace(HitLocation, HitNormal, ExplosionCenter, Location, true,,,TRACEFLAG_Blocking);
	ExplosionCenter = (HitActor == None) ? Location + vect(0,0,100) : 0.5 * (ExplosionCenter + HitLocation);

	foreach CollidingActors(class'Actor', Victim, HurtRadius, ExplosionCenter)
	{
		if (!Victim.bStatic)
		{
			Victim.GetBoundingCylinder(Radius, Height);
			Up = vect(0,0,1) * Height;
			bCauseDamage = FastTrace(Victim.Location + Up, ExplosionCenter);
			if (!bCauseDamage)
			{
				bCauseDamage = FastTrace(Victim.Location , ExplosionCenter);
				if ( !bCauseDamage && (Radius > 100.0) )
				{
					// try sides
					Side = (Normal(Victim.Location - ExplosionCenter) Cross vect(0,0,1)) * Radius;
					bCauseDamage = FastTrace(Victim.Location + Side + Up, ExplosionCenter) ||
							FastTrace(Victim.Location - Side + Up, ExplosionCenter);
				}
			}
			if (bCauseDamage)
			{
				Victim.TakeRadiusDamage(InstigatorController, Damage, HurtRadius, MyDamageType, Momentum, ExplosionCenter, false, self);
			}
		}
	}
}

state Dying
{
	function BeginState(Name PreviousStateName)
	{
		ShakeView();
	}

Begin:
	SelfDestructHurtRadius(0.125);
	Sleep(0.3);
	SelfDestructHurtRadius(0.300);
	Sleep(0.2);
	SelfDestructHurtRadius(0.475);
	Sleep(0.2);
	SelfDestructHurtRadius(0.650);

	DoKnockdown();

	Sleep(0.2);
	SelfDestructHurtRadius(0.825);
	Sleep(0.2);
	SelfDestructHurtRadius(1.0);
	Destroy();
}

defaultproperties
{
	FriendlyName="Self Destruct"
	Description="This device monitors the status of various internal systems. If it detects critical damage, it will disconnect the safety mechanisms on the Sentinel's internal power core and initiate an uncontrolled chain reaction. The end result is a big bang which will take out just about anything within a large radius.\n\nBe warned that there is no way to disable the mechanism once installed, so you will not be able to retrieve the Sentinel for spare parts."

	Damage=250.0
	DamageRadius=2000.0
	Momentum=200000.0
	MyDamageType=class'UTDmgType_SentinelSelfDestruct'
	ExplosionTemplate=ParticleSystem'Sentinel_Resources.Effects.Upgrades.P_SelfDestruct'
	ExplosionLightClass=class'UTExplosionLight_SelfDestruct'
	ExplosionSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.WarExploCue'
	ExplosionShake=CameraAnim'Camera_FX.WP_Redeemer.C_WP_Redeemer_Shake'

	AmmoCost=60
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.mactiv70Cue'
}
