//=============================================================================
// Protects the Sentinel from 100% of damage.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_Shield extends SentinelUpgrade;

var() const StaticMeshComponent SphereComponent;
var() const StaticMeshComponent FrameComponent;
var() const ParticleSystemComponent HitEffectComponent;

/** Amount of damage the shield can absorb before being destroyed. */
var() float ShieldAmount;
/** Played when shield is hit. */
var() SoundCue AbsorbSound;
/** Distance from centre to place hit effect. */
var() float HitEffectDist;
var() Vector HitEffectParticleAcceleration;

var repnotify SentinelTakeHitInfo LastTakeHitInfo;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		LastTakeHitInfo;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'LastTakeHitInfo')
	{
		PlayTakeHitEffects();
	}
}

function InitializeFor(Sentinel S)
{
	super.InitializeFor(S);
	SetOwner(S); //Must be owned by the Sentinel, or traces from it will be blocked by the shield.

	SetDrawScale(S.CylinderComponent.CollisionRadius / 34.0);
}

simulated function Tick(float DeltaTime)
{
	super.Tick(DeltaTime);

	if(Cannon != none)
	{
		SphereComponent.SetRotation(Cannon.BaseComponent.Rotation);
		FrameComponent.SetRotation(SphereComponent.Rotation);
	}
}

simulated function bool StopsProjectile(Projectile P)
{
	//Don't block projectiles fired by this Sentinel.
	return super.StopsProjectile(P) && P.Instigator != Cannon;
}

function TakeDamage(int Damage, Controller InstigatedBy, vector HitLocation, vector Momentum, class<DamageType> DamageType, optional TraceHitInfo HitInfo, optional Actor DamageCauser)
{
	if(Cannon == none || (DamageCauser != none && DamageCauser.Instigator == Cannon))
	{
		return;
	}

	Damage = (InstigatedBy == Cannon.InstigatorController) ? 0 : Max(Damage, 0);
	WorldInfo.Game.ReduceDamage(Damage, Cannon, InstigatedBy, HitLocation, Momentum, DamageType);

	super.TakeDamage(Damage, InstigatedBy, HitLocation, Momentum, DamageType, HitInfo, DamageCauser);

	ShieldAmount -= Damage;

	if((WorldInfo.TimeSeconds - Cannon.LastDamageEffectTime) > Cannon.MinDamageEffectInterval)
	{
		LastTakeHitInfo.Damage = Damage;
		LastTakeHitInfo.HitLocation = HitLocation;
		LastTakeHitInfo.DamageType = DamageType;
		Cannon.LastDamageEffectTime = WorldInfo.TimeSeconds;

		PlayTakeHitEffects();
	}

	//Note that the shield always blocks the last shot completely, even if it didn't have enough capacity to absorb all the damage without collapsing.
	if(ShieldAmount <= 0)
	{
		Cannon.UpgradeManager.RemoveUpgrade(self);
	}

	//Let controller know that someone wants a fight.
	Cannon.NotifyTakeHit(InstigatedBy, HitLocation, 0, DamageType, Momentum);
}

/**
 * Displays an effect and plays a sound where damage was inflicted.
 */
simulated function PlayTakeHitEffects()
{
	local Vector EffectTranslation;
	local Vector EffectScale;

	EffectTranslation = LastTakeHitInfo.HitLocation - SphereComponent.GetPosition();
	EffectTranslation = Normal(EffectTranslation) * HitEffectDist;
	HitEffectComponent.SetRotation(Rotator(EffectTranslation));

	EffectTranslation += SphereComponent.Translation;
	HitEffectComponent.SetTranslation(EffectTranslation);

	//Not quite sure why this needs to be scaled in this manner.
	EffectScale.X = 1.0;
	EffectScale.Y = 1.0 / DrawScale;
	EffectScale.Z = EffectScale.Y;
	HitEffectComponent.SetScale3D(EffectScale);
	
	HitEffectComponent.ActivateSystem();
	PlaySound(AbsorbSound, true,,, LastTakeHitInfo.HitLocation);
}

//Block any damage that might have been dealt to bits of the Sentinel that are somehow outside the shield, but allow damage caused from inside the shield.
function NotifyTakeDamage(out int Damage, Controller InstigatedBy, Vector HitLocation, Vector Momentum, class<DamageType> DamageType)
{
	local Vector TraceHitLocation, TraceHitNormal;
	local Vector TraceStart, TraceEnd;
	local Actor A;

	TraceStart = HitLocation;
	TraceEnd = CollisionComponent.GetPosition();
	//Move start out a bit since external hits sometimes have a HitLocation that is a tiny bit inside the shield.
	TraceStart += Normal(TraceStart - TraceEnd) * 2.0;

	foreach WorldInfo.TraceActors(class'Actor', A, TraceHitLocation, TraceHitNormal, TraceEnd, TraceStart,,, TRACEFLAG_Bullet)
	{
		//Shot was from outside the shield, block it entirely.
		if(A == self)
		{
			TakeDamage(Damage, InstigatedBy, HitLocation, Momentum, DamageType);
			Damage = 0;
			break;
		}
	}

	super.NotifyTakeDamage(Damage, InstigatedBy, HitLocation, Momentum, DamageType);
}

function int AmmoValue()
{
	return super.AmmoValue() * (ShieldAmount / default.ShieldAmount);
}

function AdjustStrength(out float Strength)
{
	super.AdjustStrength(Strength);

	Strength += ExtraStrength * (ShieldAmount / default.ShieldAmount);
}

simulated function GetStatusText(out array<String> StatusStrings)
{
	local String StatusLine;

	StatusLine = class'UIStrings'.default.Shield$": ";
	class'SUtils'.static.PadWithSpaces(StatusLine, 15);
	StatusLine $= class'SUtils'.static.RedGreenTextColourMarkup(ShieldAmount, default.ShieldAmount);
	StatusLine $= int((ShieldAmount / default.ShieldAmount) * 100.0)$"%";
	StatusLine $= class'SUtils'.static.TextColourMarkup(1.0, 1.0, 1.0);
	StatusStrings[StatusStrings.Length] = StatusLine;

	super.GetStatusText(StatusStrings);
}

defaultproperties
{
	FriendlyName="Molecular Shield"
	Description="A molecular shield to provide complete protection from any and all weapons a Sentinel is likely to be faced with, whilst allowing the Sentinel to continue firing unaffected. The standard battery fitted provides enough power to absorb substantial punishment before running out.\n\nThe shield does not hinder the passage of most lifeforms, allowing the owner to pass through to carry out repairs and upgrades, without disrupting the integrity of the shield as a whole."

	ShieldAmount=1000.0
	AbsorbSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.ShieldAbsorbCue'
	HitEffectDist=90.0
	HitEffectParticleAcceleration=(X=-220.0,Y=0.0,Z=0.0)

	ExtraStrength=0.5
	AmmoCost=80
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.ShieldActivateCue'
	DeactivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.ShieldDrainedCue'

	ForbiddenUpgrades.Add(class'SentinelUpgrade_Cloak')

	bCollideActors=true
	bBlockActors=false
	bCollideWorld=false
	bProjTarget=true
	bIgnoreRigidBodyPawns=true
	bIgnoreBaseRotation=true

	Begin Object Class=StaticMeshComponent Name=SphereComponent0
		StaticMesh=StaticMesh'Sentinel_Resources.Meshes.Upgrades.ShieldSphere'
		AbsoluteRotation=true
		CollideActors=false
		BlockActors=false
		BlockZeroExtent=false
		BlockNonZeroExtent=false
		CastShadow=false
		BlockRigidBody=false
		bAcceptsLights=false
		bAcceptsDecals=false
		bAcceptsDecalsDuringGameplay=false
		bUseAsOccluder=false
		Scale=1.33
	End Object
	SphereComponent=SphereComponent0
	Components.Add(SphereComponent0)

	Begin Object Class=StaticMeshComponent Name=FrameComponent0
		StaticMesh=StaticMesh'Sentinel_Resources.Meshes.Upgrades.ShieldSphere'
		Materials[0]=Material'Sentinel_Resources.Materials.Upgrades.M_Shield_Wireframe'
		AbsoluteRotation=true
		CollideActors=true
		BlockActors=false
		BlockZeroExtent=true
		BlockNonZeroExtent=true
		BlockRigidBody=false
		CastShadow=false
		bAcceptsLights=false
		bAcceptsDecals=false
		bAcceptsDecalsDuringGameplay=false
		bUseAsOccluder=false
		Scale=1.35
	End Object
	FrameComponent=FrameComponent0
	CollisionComponent=FrameComponent0
	Components.Add(FrameComponent0)

	Begin Object Class=ParticleSystemComponent Name=ParticleSystemComponent0
		Template=ParticleSystem'Sentinel_Resources.Effects.Upgrades.P_Shield_Hit'
		AbsoluteRotation=true
		bAutoActivate=false
		Scale=1.33
	End Object
	HitEffectComponent=ParticleSystemComponent0
	Components.Add(ParticleSystemComponent0)
}
