//=============================================================================
// Gives the owning player's Deployer access to health and  target information.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelUpgrade_Telemeter extends SentinelUpgrade;

/** Interval between updates of target information. */
var() float TargetUpdatePeriod;
/** Interval between searches for Deployers to render for. */
var() float DeployerSearchPeriod;
/** Distance beyond which only health will be rendered on the beacon. */
var() float BeaconTextMaxDistance;
/** Disance beyond which owner's name will not be rendered on the beacon. */
var() float BeaconOwnerTextMaxDistance;

/** Colour to draw line in when target is visible. */
var() Color VisibleColour;
/** Colour to draw line in when target is hidden. */
var() Color HiddenColour;
/** Current line colour */
var Color LineColour;
/** WTF does this do? */
var() float LinePerc;
/** Space between line and dropshadow line. */
var() float LineSpacing;
/** Width of line as a fraction of screen width. */
var() float LineWidth;

/** Distance away from text that shadow text is drawn relative to the font size. */
var() float TextShadowDistance;
/** Size of target text and beacon text font. */
var() int TextSize;

/** How much wider the beacon background is than its contents, as a fraction of text height. */
var() float BeaconBackgroundWidth;
/** How much higher the beacon background is than its contents, as a fraction of text height. */
var() float BeaconBackgroundHeight;
/** How much space there is between lines of text in the beacon relative to the text height. */
var() float BeaconTextSpacing;
/** How much wider the beacon health bar is than the text width, as a fraction of text height. */
var() float BeaconHealthWidth;
/** How much higher the beacon health bar is relative to the text height, as a fraction of text height. */
var() float BeaconHealthHeight;

var Vector TargetLocation;
var String TargetText;
var int TargetTeamIndex;
var bool bValidTarget;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		bValidTarget;

	if(Role == ROLE_Authority && bNetDirty && bValidTarget)
		TargetLocation, TargetText, TargetTeamIndex, LineColour;
}

function NotifyNewTarget(Actor NewTarget)
{
	GotoState('Tracking');

	super.NotifyNewTarget(NewTarget);
}

function NotifyWaiting()
{
	GotoState('');

	super.NotifyWaiting();
}

simulated function ClientInitializeFor()
{
	super.ClientInitializeFor();

	FindDeployersToRenderFor();
	SetTimer(DeployerSearchPeriod, true, 'FindDeployersToRenderFor');
	bRenderOverlays = true;
}

simulated function RenderOverlays(HUD H)
{
	local UTHUD UTH;
	local PlayerController PC;
	local Vector ViewLocation;
	local Rotator ViewRotation;

	UTH = UTHUD(H);

	if(UTH == none)
		return;

	PC = PlayerController(UTH.PawnOwner.Controller);

	if(PC == none)
		return;

	ViewLocation = UTH.PawnOwner.GetPawnViewLocation();
	ViewRotation = UTH.PawnOwner.GetViewRotation();

	//Set font here for both target text and beacon text.
	H.Canvas.Font = H.GetFontSizeIndex(TextSize);

	//Draw current target indicator.
	if(bValidTarget)
	{
		DrawTargetLine(H.Canvas, ViewLocation, ViewRotation, PC.FOVAngle);

		//Only bother if TargetLocation is within player's field of view.
		if(RDiff(ViewRotation, Rotator(TargetLocation - ViewLocation)) < PC.FOVAngle)
		{
			DrawTargetText(H.Canvas);
		}
	}

	//Draw the beacon.
	if(RDiff(ViewRotation, Rotator(Cannon.Location - ViewLocation)) < PC.FOVAngle)
	{
		DrawBeacon(H.Canvas, VSize(Cannon.Location - ViewLocation));
	}
}

/**
 * Draws the targets name.
 *
 * @param	Canvas	canvas to draw the information on
 */
simulated function DrawTargetText(Canvas Canvas)
{
	local float TextXL, TextYL;
	local Vector TextPosition;
	local LinearColor TextColour;
	local Color DropColour;

	Canvas.TextSize(TargetText, TextXL, TextYL);
	TextPosition = Canvas.Project(TargetLocation);
	TextPosition.X -= TextXL * 0.5;
	TextPosition.Y -= TextYL;

	class'UTHUD'.static.GetTeamColor(TargetTeamIndex, TextColour, DropColour);
	TextColour.A = 1.0;

	//Drop shadow text.
	Canvas.DrawColor = DropColour;
	Canvas.TextSize("X", TextXL, TextYL);
	Canvas.SetPos(TextPosition.X + (TextXL * TextShadowDistance), TextPosition.Y + (TextXL * TextShadowDistance));
	Canvas.DrawTextClipped(TargetText);

	//Target text.
	Canvas.DrawColor = class'SUtils'.static.LinearColorToColor(TextColour);
	Canvas.SetPos(TextPosition.X, TextPosition.Y);
	Canvas.DrawTextClipped(TargetText);
}

simulated function DrawTargetLine(Canvas Canvas, Vector ViewLocation, Rotator ViewRotation, float FOV)
{
	local Vector CannonLocation;
	local float AngleToTarget, AngleToCannon;
	local float HalfFOV;
	local Vector LineStartPosition, LineEndPosition;
	local float AbsoluteLineWidth;

	CannonLocation = Cannon.GetPawnViewLocation();
	AngleToCannon = RDiff(ViewRotation, Rotator(CannonLocation - ViewLocation));
	AngleToTarget = RDiff(ViewRotation, Rotator(TargetLocation - ViewLocation));
	HalfFOV = FOV * 0.5;

	if(AngleToCannon > HalfFOV)
	{
		if(AngleToTarget > HalfFOV)
		{
			//Both outside FOV.
			return;
		}
		else
		{
			//Only target within FOV
			LineStartPosition = Canvas.Project(TargetLocation - Normal(TargetLocation - CannonLocation));
			LineStartPosition.Z = 0.0;
			LineEndPosition = Canvas.Project(TargetLocation);
			LineEndPosition.Z = 0.0;

			LineStartPosition = LineEndPosition - ((Canvas.ClipX + Canvas.ClipY) * Normal(LineEndPosition - LineStartPosition));

			//DrawTextureLine is funny about positions below the bottom of the screen, so clip it.
			if(LineStartPosition.Y > Canvas.ClipY)
			{
				LineStartPosition = Normal(LineStartPosition - LineEndPosition) * ((Canvas.ClipY - LineEndPosition.Y) / ((Canvas.ClipY - LineEndPosition.Y) + LineStartPosition.Y - Canvas.ClipY)) * VSize(LineStartPosition - LineEndPosition);
				LineStartPosition += LineEndPosition;
			}
		}
	}
	else
	{
		if(AngleToTarget > HalfFOV)
		{
			//Only cannon within FOV.
			LineStartPosition = Canvas.Project(CannonLocation);
			LineStartPosition.Z = 0.0;
			LineEndPosition = Canvas.Project(CannonLocation - Normal(CannonLocation - TargetLocation));
			LineEndPosition.Z = 0.0;

			LineEndPosition = LineStartPosition - ((Canvas.ClipX + Canvas.ClipY) * Normal(LineStartPosition - LineEndPosition));

			//DrawTextureLine is funny about positions below the bottom of the screen, so clip it.
			if(LineEndPosition.Y > Canvas.ClipY)
			{
				LineEndPosition = Normal(LineEndPosition - LineStartPosition) * ((Canvas.ClipY - LineStartPosition.Y) / ((Canvas.ClipY - LineStartPosition.Y) + LineEndPosition.Y - Canvas.ClipY)) * VSize(LineEndPosition - LineStartPosition);
				LineEndPosition += LineStartPosition;
			}

			//DrawTextureLine is funny about end positions off the right of the screen, so clip it.
			if(LineEndPosition.X > Canvas.ClipX)
			{
				LineEndPosition = Normal(LineEndPosition - LineStartPosition) * ((Canvas.ClipX - LineStartPosition.X) / ((Canvas.ClipX - LineStartPosition.X) + LineEndPosition.X - Canvas.ClipX)) * VSize(LineEndPosition - LineStartPosition);
				LineEndPosition += LineStartPosition;
			}
		}
		else
		{
			//Both inside FOV.
			LineStartPosition = Canvas.Project(CannonLocation);
			LineEndPosition = Canvas.Project(TargetLocation);
		}
	}

	AbsoluteLineWidth = FMax(LineWidth * Canvas.ClipX, 1.0);

	//Target line and drop shadow line drawn at once.
	//Canvas.DrawTextureDoubleLine(LineStartPosition, LineEndPosition, LinePerc, LineSpacing, AbsoluteLineWidth, LineColour, DropColour, Texture2D'EngineResources.White', 0.0, 0.0, 1.0, 1.0);
	Canvas.DrawTextureLine(LineStartPosition, LineEndPosition, LinePerc, AbsoluteLineWidth, LineColour, Texture2D'EngineResources.White', 0.0, 0.0, 1.0, 1.0);
}

/**
 * Renders a beacon showing a name and health.
 *
 * @param	Canvas			canvas to render the beacon on
 * @param	Distance		distance from camera of the beacon's owner
 */
simulated function DrawBeacon(Canvas Canvas, float Distance)
{
	local Vector ScreenLocation;
	local float HealthState;
	local string Text;
	local float TextX, TextY;
	local float TextXL, TextYL;
	local string OwnerText;
	local float OwnerTextX, OwnerTextY;
	local float OwnerTextXL, OwnerTextYL;
	local float HealthX, HealthY;
	local float HealthXL, HealthYL;
	local float BackgroundX, BackgroundY;
	local float BackgroundXL, BackgroundYL;
	local LinearColor DropColour;
	local Color	TextColour;

	ScreenLocation = Canvas.Project(Cannon.Location + (vect(0.0, 0.0, 1.0) * Cannon.CylinderComponent.CollisionHeight));
	class'UTHUD'.Static.GetTeamColor(Cannon.GetTeamNum(), DropColour, TextColour);
	DropColour.A = 1.0;
	HealthState = float(Cannon.Health) / float(Cannon.HealthMax);

	if(Distance < BeaconTextMaxDistance)
	{
		Text = Cannon.ShortMenuName;
		Canvas.StrLen(Text, TextXL, TextYL);
		TextX = ScreenLocation.X - (TextXL * 0.5);
		TextY = ScreenLocation.Y - (TextYL * 0.5);

		if(Distance < BeaconOwnerTextMaxDistance)
		{
			OwnerText = SentinelReplicationInfo(Cannon.PlayerReplicationInfo).OwnerPRI.GetPlayerAlias();
			Canvas.StrLen(OwnerText, OwnerTextXL, OwnerTextYL);
			OwnerTextYL = TextYL;
			OwnerTextX = ScreenLocation.X - (OwnerTextXL * 0.5);
			OwnerTextY = ScreenLocation.Y - (OwnerTextYL * 0.5) + (OwnerTextYL * BeaconTextSpacing);

			TextY -= OwnerTextYL * BeaconTextSpacing;
		}
	}
	else
	{
		Canvas.StrLen("XX", TextXL, TextYL);
		TextYL *= 0.5; //Reduce height of health bar at very long range.
	}

	HealthXL = FMax(TextXL, OwnerTextXL) + (TextYL * (BeaconHealthWidth - 1.0));
	HealthYL = (TextYL * BeaconHealthHeight) + (OwnerTextYL * (0.5 + BeaconTextSpacing));
	HealthX =ScreenLocation.X - (HealthXL * 0.5);
	HealthY = ScreenLocation.Y - (HealthYL * 0.5);

	BackgroundXL = HealthXL + (TextYL * (BeaconBackgroundWidth - 1.0));
	BackgroundYL = HealthYL + ((BeaconBackgroundHeight - 1.0) * TextYL);
	BackgroundX = ScreenLocation.X - (BackgroundXL * 0.5);
	BackgroundY = ScreenLocation.Y - (BackgroundYL * 0.5);

	//Shift everything up so it doesn't cover the Sentinel too much.
	TextY -= BackgroundYL * 0.5;
	OwnerTextY -= BackgroundYL * 0.5;
	HealthY -= BackgroundYL * 0.5;
	BackgroundY -= BackgroundYL * 0.5;

	//Draw background.
	DrawBackground(BackgroundX, BackgroundY, BackgroundXL, BackgroundYL, Cannon.TeamColour, Canvas);

	//Draw health bar.
	class'UTHUD'.static.DrawHealth(HealthX, HealthY, HealthState * HealthXL, HealthXL, HealthYL, Canvas);

	//Draw text.
	if(Text != "")
	{
		//Drop shadow.
		Canvas.DrawColor = class'SUtils'.static.LinearColorToColor(DropColour);
		Canvas.TextSize("X", TextXL, TextYL);
		Canvas.SetPos(TextX + (TextXL * TextShadowDistance), TextY + (TextXL * TextShadowDistance));
		Canvas.DrawTextClipped(Text, true);

		//Text.
		Canvas.DrawColor = TextColour;
		Canvas.SetPos(TextX, TextY);
		Canvas.DrawTextClipped(Text, true);

		//Draw owner text.
		if(OwnerText != "")
		{
			//Drop shadow.
			Canvas.DrawColor = class'SUtils'.static.LinearColorToColor(DropColour);
			Canvas.SetPos(OwnerTextX + (TextXL * TextShadowDistance), OwnerTextY + (TextXL * TextShadowDistance));
			Canvas.DrawTextClipped(OwnerText, true);

			Canvas.DrawColor = TextColour;
			Canvas.SetPos(OwnerTextX, OwnerTextY);
			Canvas.DrawTextClipped(OwnerText, true);
		}
	}
}

/**
 * Draws a translucent rectangle with an opaque outline.
 *
 * @param	X			X coordinate in pixele of top left corner
 * @param	Y			Y coordinate in pixels of top left corner
 * @param	Width		width in pixels to draw background
 * @param	Height		height in pixles to draw background
 * @param	DrawColour	colour to draw background
 */
simulated function DrawBackground(float X, float Y, float Width, float Height, LinearColor DrawColour, Canvas DrawCanvas)
{
	//Background.
	DrawCanvas.SetPos(X, Y);
	DrawCanvas.DrawColorizedTile(Texture2D'Sentinel_Resources.Textures.UI.T_Beacon_Background', Width, Height, 0, 32, 128, 64, DrawColour);

	//Upper bracket.
	DrawCanvas.SetPos(X, Y);
	DrawCanvas.DrawColorizedTile(Texture2D'Sentinel_Resources.Textures.UI.T_Beacon_Background', Width, Width * 0.25, 0, 0, 128, 32, DrawColour);

	//Lower bracket.
	DrawCanvas.SetPos(X, Y + Height - (Width * 0.25));
	DrawCanvas.DrawColorizedTile(Texture2D'Sentinel_Resources.Textures.UI.T_Beacon_Background', Width, Width * 0.25, 0, 96, 128, 32, DrawColour);
}

state Tracking
{
	function BeginState(Name PreviousStateName)
	{
		TargetTimer();
		bValidTarget = true;
		SetTimer(TargetUpdatePeriod, true, 'TargetTimer');
	}

	/**
	 * Updates target information.
	 */
	function TargetTimer()
	{
		if(Cannon.Target != none)
		{
			if(Cannon.SController.bEnemyIsVisible)
			{
				TargetLocation = Cannon.Target.Location;
				LineColour = VisibleColour;
			}
			else
			{
				TargetLocation = Cannon.SController.LastDetectedLocation;
				LineColour = HiddenColour;
			}

			if(Cannon.TargetName != "")
				TargetText = Cannon.TargetName;
			else
				TargetText = class'UIStrings'.default.Scanning;

			TargetTeamIndex = Cannon.Target.GetTeamNum();
		}
	}

	function EndState(Name NextStateName)
	{
		bValidTarget = false;
		ClearTimer('TargetTimer');
	}
}

defaultproperties
{
	FriendlyName="Telemeter"
	Description="Transmits various information to allied Deployers, greatly aiding tactical decision making:\n\nThe Sentinel's current health and owner's name are displayed above its location.\n\nIf the Sentinel has line of sight to its target, then the target's location will be indicated with a green line green, otherwise the last known location will be indicated in red."

	TargetUpdatePeriod=0.2
	DeployerSearchPeriod=2.0
	BeaconTextMaxDistance=1500.0
	BeaconOwnerTextMaxDistance=500.0
	TargetTeamIndex=255

	BeaconBackgroundWidth=1.0
	BeaconBackgroundHeight=1.0
	BeaconTextSpacing=0.5
	BeaconHealthWidth=2.0
	BeaconHealthHeight=1.3

	VisibleColour=(R=0,G=255,B=0,A=255)
	HiddenColour=(R=255,G=0,B=0,A=255)
	LinePerc=0.0
	LineSpacing=0.0
	LineWidth=0.001

	TextShadowDistance=0.1
	TextSize=0

	AmmoCost=20
	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Upgrades.TelemeterActivateCue'

	bAlwaysRelevant=true
}
