//=============================================================================
// Handles hit effects for different materials.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeaponComponent_HitEffects extends Component;

/** impact effects by material type */
var() array<MaterialImpactEffect> ImpactEffects;
/** default impact effect to use if a material specific one isn't found */
var() MaterialImpactEffect DefaultImpactEffect;
/** Distance from viewer beyond which effects will not be spawned. */
var() float MaxImpactEffectDistance;
/** Controls decal fade out. */
var() InterpCurveFloat DecalMaterialParameterCurve;

//Mostly copied from UTWeaponAttachment.uc
function MaterialImpactEffect GetImpactEffect(PhysicalMaterial HitMaterial, optional name MaterialType)
{
	local int i;
	local UTPhysicalMaterialProperty PhysicalProperty;

	if(HitMaterial != none)
	{
		PhysicalProperty = UTPhysicalMaterialProperty(HitMaterial.GetPhysicalMaterialProperty(class'UTPhysicalMaterialProperty'));

		if(PhysicalProperty != none)
		{
			MaterialType = PhysicalProperty.MaterialType;
		}
	}

	if(MaterialType != 'None')
	{
		i = ImpactEffects.Find('MaterialType', MaterialType);

		if(i != INDEX_NONE)
		{
			return ImpactEffects[i];
		}
	}

	return DefaultImpactEffect;
}

//Mostly copied from UTWeaponAttachment.uc
function PlayImpactEffects(Vector FireLocation, Vector HitLocation, Actor Impactor)
{
	local WorldInfo WI;
	local Vector NewHitLoc, HitNormal, WaterHitNormal;
	local Actor HitActor;
	local PhysicsVolume HitVolume;
	local TraceHitInfo HitInfo;
	local MaterialImpactEffect ImpactEffect;
	local Pawn P;
	local Vehicle V;

	WI = Impactor.WorldInfo;

	HitNormal = Normal(FireLocation - HitLocation);

	if(Impactor.EffectIsRelevant(HitLocation, false, MaxImpactEffectDistance))
	{
		if(!WI.bDropDetail)
		{
			HitVolume = PhysicsVolume(Impactor.Trace(NewHitLoc, WaterHitNormal, HitLocation, FireLocation, true,, HitInfo, Impactor.TRACEFLAG_PhysicsVolumes | Impactor.TRACEFLAG_Bullet));

			if(HitVolume != none && HitVolume.bWaterVolume)
			{
				ImpactEffect = GetImpactEffect(none, 'Water');

				if(ImpactEffect != DefaultImpactEffect)
				{
					if(ImpactEffect.Sound != none)
					{
						Impactor.PlaySound(ImpactEffect.Sound, true,,, NewHitLoc);
					}

					if(ImpactEffect.ParticleTemplate != none)
					{
						WI.MyEmitterPool.SpawnEmitter(ImpactEffect.ParticleTemplate, NewHitLoc, Rotator(WaterHitNormal));
					}
				}
			}
		}

		HitActor = Impactor.Trace(NewHitLoc, HitNormal, (HitLocation - (HitNormal * 32)), HitLocation + (HitNormal * 32), true,, HitInfo, Impactor.TRACEFLAG_Bullet);

		if(HitActor != none && PortalTeleporter(HitActor) == none)
		{
			P = Pawn(HitActor);

			if(P != none)
			{
				Impactor.CheckHitInfo(HitInfo, P.Mesh, -HitNormal, NewHitLoc);

				V = Vehicle(HitActor);
			}

			ImpactEffect = GetImpactEffect(HitInfo.PhysMaterial);

			if(ImpactEffect.Sound != none)
			{
				Impactor.PlaySound(ImpactEffect.Sound, true,,, HitLocation);
			}

			//Pawns handle their own hit effects
			if(P == none || V != none)
			{
				if(!WI.bDropDetail && (P == none) && (WI.GetDetailMode() != DM_Low))
				{
					SpawnImpactDecal(ImpactEffect, HitLocation, HitNormal, HitInfo, WI);
				}

				if(ImpactEffect.ParticleTemplate != none)
				{
					WI.MyEmitterPool.SpawnEmitter(ImpactEffect.ParticleTemplate, HitLocation, Rotator(HitNormal), HitActor);
				}
			}
		}
	}
}

function SpawnImpactDecal(MaterialImpactEffect ImpactEffect, Vector HitLocation, Vector HitNormal, TraceHitInfo HitInfo, WorldInfo WI)
{
	local int DecalMaterialsLength;
	local MaterialInterface MI;
	local MaterialInstanceTimeVarying MITV_Decal;
	local float DecalDissolveStartTime;

	DecalMaterialsLength = ImpactEffect.DecalMaterials.length;

	if(DecalMaterialsLength > 0)
	{
		MI = ImpactEffect.DecalMaterials[Rand(DecalMaterialsLength)];

		if(MI != none)
		{
			MITV_Decal = new(Outer) class'MaterialInstanceTimeVarying';
			MITV_Decal.SetParent(MI);

			WI.MyDecalManager.SpawnDecal(MITV_Decal, HitLocation, Rotator(-HitNormal), ImpactEffect.DecalWidth,
				ImpactEffect.DecalHeight, 10.0, false,, HitInfo.HitComponent, true, false, HitInfo.BoneName, HitInfo.Item, HitInfo.LevelIndex);

			MITV_Decal.SetScalarCurveParameterValue(ImpactEffect.DecalDissolveParamName, DecalMaterialParameterCurve);
			//Start fading out just soon enough to finish fading when the decal's lifespan runs out.
			DecalDissolveStartTime = WI.MyDecalManager.DecalLifeSpan;
			DecalDissolveStartTime -= DecalMaterialParameterCurve.Points[DecalMaterialParameterCurve.Points.length - 1].InVal;
			DecalDissolveStartTime = FMax(DecalDissolveStartTime, 0.0);
			MITV_Decal.SetScalarStartTime(ImpactEffect.DecalDissolveParamName, DecalDissolveStartTime);
		}
	}
}

defaultproperties
{
	ImpactEffects(0)=(MaterialType=Dirt,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.DirtHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default3'),DecalWidth=6.0,DecalHeight=6.0)
	ImpactEffects(1)=(MaterialType=Gravel,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.DirtHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default3'),DecalWidth=6.0,DecalHeight=6.0)
	ImpactEffects(2)=(MaterialType=Sand,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.DirtHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow3'),DecalWidth=8.0,DecalHeight=8.0)
	ImpactEffects(3)=(MaterialType=Dirt_Wet,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactMud_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow3'),DecalWidth=8.0,DecalHeight=8.0)
	ImpactEffects(4)=(MaterialType=Energy,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactEnergy_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Metal')
	ImpactEffects(5)=(MaterialType=WorldBoundary,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactEnergy_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default')
	ImpactEffects(6)=(MaterialType=Stone,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.StoneHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default3'),DecalWidth=6.0,DecalHeight=6.0)
	ImpactEffects(7)=(MaterialType=Flesh_Human,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactFlesh_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow3'),DecalWidth=8.0,DecalHeight=8.0)
	ImpactEffects(8)=(MaterialType=Foliage,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactFoliage_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Plant',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow3'),DecalWidth=8.0,DecalHeight=8.0)
	ImpactEffects(9)=(MaterialType=Glass,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.GlassHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Glass',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Glass1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Glass2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Glass3'),DecalWidth=16.0,DecalHeight=16.0)
	ImpactEffects(10)=(MaterialType=Liquid,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactWater_Cue',ParticleTemplate=ParticleSystem'WP_Enforcers.Effects.P_WP_Enforcer_Impact')
	ImpactEffects(11)=(MaterialType=Water,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactWater_Cue',ParticleTemplate=ParticleSystem'Envy_Effects.Particles.P_WP_Water_Splash_Small')
	ImpactEffects(12)=(MaterialType=ShallowWater,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactWater_Cue',ParticleTemplate=ParticleSystem'Envy_Effects.Particles.P_WP_Water_Splash_Small')
	ImpactEffects(13)=(MaterialType=Lava,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactWater_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Metal')
	ImpactEffects(14)=(MaterialType=Slime,Sound=SoundCue'A_Weapon_Enforcer.Cue.A_Weapon_Enforcer_ImpactWater_Cue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default')
	ImpactEffects(15)=(MaterialType=Metal,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.MetalHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Metal',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Metal1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Metal2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Metal3'),DecalWidth=12.0,DecalHeight=12.0)
	ImpactEffects(16)=(MaterialType=Snow,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.SnowHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Snow',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Snow3'),DecalWidth=8.0,DecalHeight=8.0)
	ImpactEffects(17)=(MaterialType=Wood,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.WoodHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Wood',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Wood1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Wood2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Wood3'),DecalWidth=5.0,DecalHeight=5.0)
	ImpactEffects(18)=(MaterialType=NecrisVehicle,Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.MetalHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Metal',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Metal1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Metal2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Metal3'),DecalWidth=12.0,DecalHeight=12.0)

	DefaultImpactEffect=(Sound=SoundCue'Sentinel_Resources.Sounds.HitEffects.StoneHitCue',ParticleTemplate=ParticleSystem'Sentinel_Resources.Effects.HitEffects.P_Impact_Default',DecalMaterials=(MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default1',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default2',MaterialInterface'Sentinel_Resources.Materials.Decals.MITV_Decal_Default3'),DecalWidth=6.0,DecalHeight=6.0)

	MaxImpactEffectDistance=4000.0

	DecalMaterialParameterCurve=(Points=((InVal=0.0,OutVal=0.0),(InVal=4.0,OutVal=1.0)))
}