//=============================================================================
// Oh yeah!
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeapon_Disturber extends SentinelWeapon;

/** FireInfo for main projectile. */
var() const SentinelWeaponComponent_FireInfo ProjectileFireInfo;
/** FireInfo for zap beam. */
var() const SentinelWeaponComponent_FireInfo ZapFireInfo;

/** Controls muzzle flash light when zapping. */
var() array<LightValues> ZapLightTimeShift;
/** Time between zap checks. */
var() float ZapCheckInterval;

/** Last projectile fired. */
var SentinelProjectile_Disturber CurrentProj;
/** Set true when in WaitingToZap state, so FlashMuzzleFlash knows when to show the beam effect. */
var repnotify bool bZapping;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		CurrentProj, bZapping;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'bZapping')
	{
		SetFireMode();
	}
	else
	{
		super.ReplicatedEvent(VarName);
	}
}

simulated function SetFireMode()
{
	if(bZapping)
	{
		FireInfo = ZapFireInfo;
	}
	else
	{
		FireInfo = ProjectileFireInfo;
	}

	super.SetFireMode();
}

function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
{
	local Vector Aim;
	local Vector X, Y, Z;
	local Vector SpawnLocation;
	local bool bFired;

	Aim = Normal(End - Start);
	GetAxes(BarrelDir, X, Y, Z);

	SpawnLocation = Start;
	SpawnLocation += X * FireInfo.FireOffset.X;
	SpawnLocation += Y * FireInfo.FireOffset.Y;
	SpawnLocation += Z * FireInfo.FireOffset.Z;

	CurrentProj = SentinelProjectile_Disturber(Cannon.Spawn(FireInfo.ProjectileClass, Cannon,, SpawnLocation, Rotator(Aim)));

	if(CurrentProj != none)
	{
		InitializeProjectile(CurrentProj, Aim);
		bFired = true;
		super.FireAt(Start, BarrelDir, End);
		GotoState('WaitingToZap');
	}

	return bFired;
}

simulated function FlashMuzzleFlash()
{
	MuzzleFlash.ResetToDefaults();

	if(bZapping)
	{
		MuzzleFlashLight.TimeShift = ZapLightTimeShift;

		if(CurrentProj != none && !CurrentProj.bDeleteMe)
		{
			MuzzleFlash.SetActorParameter('BeamTarget', CurrentProj);
			//MuzzleFlash.SetVectorParameter('BeamEnd', CurrentProj.Location);
			MuzzleFlash.bShouldFlash = true;
		}
		else
		{
			MuzzleFlash.bShouldFlash = false;
		}
	}
	else
	{
		MuzzleFlashLight.TimeShift = MuzzleFlashLight.default.TimeShift;
		MuzzleFlash.bShouldFlash = false;
	}

	super.FlashMuzzleFlash();
}

function ModifyPawnTargetPriority(out float Weight, Pawn PawnTarget)
{
	super.ModifyPawnTargetPriority(Weight, PawnTarget);

	//Prefer targets that can be zapped without having to fire again.
	if(CouldZapWithCurrentProj(PawnTarget))
	{
		Weight += 0.5;

		//Even more likely to change target if current enemy can't be zapped.
		if(Cannon.SController.Enemy != none && !CouldZapWithCurrentProj(Cannon.SController.Enemy))
		{
			Weight += 0.5;
		}
	}
}

//Check to see if CurrentProj could be used to zap new target, or if need to fire a new one.
function NotifyNewTarget(Actor NewTarget)
{
	local Pawn P;

	super.NotifyNewTarget(NewTarget);

	P = Pawn(NewTarget);

	if(P != none && CouldZapWithCurrentProj(P))
	{
		if(!IsInState('WaitingToZap'))
		{
			GotoState('WaitingToZap');
		}
	}
	else
	{
		GotoState('');
	}
}

function NotifyDied(Controller Killer, out class<DamageType> DamageType, vector HitLocation)
{
	super.NotifyDied(Killer, DamageType, HitLocation);

	//Make sure no more zapping occurs.
	CurrentProj = none;
	GotoState('');
}

/**
 * Returns true if CurrentProj could still be used to zap something, i.e. it still exists, is visible and is in range.
 */
function bool CurrentProjValid()
{
	local bool bValid;

	bValid = true;

	if(CurrentProj == none || CurrentProj.bDeleteMe || CurrentProj.Physics == PHYS_None)
	{
		bValid = false;
	}
	else if(VSize(CurrentProj.Location - Cannon.Location) > Cannon.GetRange())
	{
		bValid = false;
	}
	else if(!Cannon.FastTrace(CurrentProj.Location, Cannon.GetPawnViewLocation(),, true))
	{
		bValid = false;
	}

	return bValid;
}

/**
 * Returns true if the current projectile could be used to zap the given pawn now or in the future.
 */
function bool CouldZapWithCurrentProj(Pawn PotentialZapee)
{
	local bool bCouldZap;
	local float ClosestTime;
	local float ClosestDistance;
	local Vector ClosestPoint1, ClosestPoint2;

	bCouldZap = true;

	if(!CurrentProjValid())
	{
		CurrentProj = none;
		bCouldZap = false;
	}
	else
	{
		class'SUtils'.static.ClosestPointOfApproach(CurrentProj, PotentialZapee, ClosestTime, ClosestPoint1, ClosestPoint2, ClosestDistance);

		if(ClosestTime > CurrentProj.LifeSpan)
		{
			bCouldZap = false;
		}
		else if(ClosestTime <= 0.0 && VSize(PotentialZapee.Location - CurrentProj.Location) > CurrentProj.ZapFireInfo.MaxRange)
		{
			bCouldZap = false;
		}
		else if(ClosestDistance > CurrentProj.ZapFireInfo.MaxRange)
		{
			bCouldZap = false;
		}
		else if(VSize(ClosestPoint1 - Cannon.Location) > Cannon.GetRange())
		{
			bCouldZap = false;
		}
		else if(!Cannon.FastTrace(ClosestPoint1, Cannon.GetPawnViewLocation(),, true))
		{
			bCouldZap = false;
		}
	}

	return bCouldZap;
}

state WaitingToZap
{
	function BeginState(Name PreviousStateName)
	{
		Cannon.SController.ForceTarget(CurrentProj);
		SetTimer(ZapCheckInterval, true, 'ZapTimer');
	}

	function ZapTimer()
	{
		if(!bZapping)
		{
			//Done here because BeginState runs before the FlashMuzzleFlash that goes with the projectile.
			bZapping = true;
			SetFireMode();
		}

		if(!CurrentProjValid())
		{
			CurrentProj = none;
			GotoState('');
		}
		else if(bCanFire)
		{
			if(CurrentProj.FindZapee() != none)
			{
				Cannon.FireAt(CurrentProj.Location);
			}
		}
	}

	function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
	{
		local Vector X, Y, Z;
		local Vector HitLocation, HitNormal;
		local bool bFired;

		GetAxes(BarrelDir, X, Y, Z);

		//Only fire if it will hit the projectile.
		if(Cannon.TraceComponent(HitLocation, HitNormal, CurrentProj.CollisionComponent, End, Start))
		{
			//TODO: Add a bullet whip, same as or similar to fire sound.
			if(CurrentProj.Zap())
			{
				bFired = true;
				super(SentinelWeapon).FireAt(Start, BarrelDir, End);
			}
		}

		return bFired;
	}

	function NotifyWaiting()
	{
		super.NotifyWaiting();

		GotoState('');
	}

	function EndState(Name NextStateName)
	{
		Cannon.SController.ForceTarget(none);
		bZapping = false;
		SetFireMode();

		ClearTimer('ZapTimer');
	}
}

defaultproperties
{
	FriendlyName="Disturber"
	Description="The \"standard\" Transdimensional Disturber requires more power to operate continuously than even the heaviest class of Sentinel can supply, but innovations in micro-wormhole manipulation have led to this reduced-power version that retains much of the destructive power you have come to expect.\n\nLike its predecessor, this Disturber fires wormholes linked to a specific point in a parallel dimension that, in an as-yet undetermined manner, reflects and magnifies the directed energy beam subsequently fired at it from the Disturber."

	WeaponMesh=SkeletalMesh'Sentinel_Resources.Meshes.Weapons.SK_Disturber'
	PhysicsAsset=PhysicsAsset'Sentinel_Resources.Physics.SK_Disturber_Physics'
	WeaponMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel'
	WeaponSpawnMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Spawn'
	WeaponDeadMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Dead'
	WeaponScale=0.6

	ZapCheckInterval=0.2

	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Weapons.redeemer_changeCue'

	ExtraStrength=0.5
	AmmoCost=150
	CompatibleSentinels.Remove(class'Sentinel_Ceiling')

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		FireInterval=2.0
		ProjectileClass=class'SentinelProjectile_Disturber'
		FireOffset=(X=20.0,Y=0.0,Z=0.0)
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.DisturberFireCue'
	End Object
	ProjectileFireInfo=FireInfo0
	FireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo1
		FireInterval=0.6
		FireOffset=(X=10.0,Y=0.0,Z=0.0)
		MaxRange=16384.0
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.electricity1Cue'
	End Object
	ZapFireInfo=FireInfo1

	Begin Object Class=SentinelWeaponComponent_MuzzleFlash Name=MuzzleFlash0
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Disturber_Beam'
		bShouldFlash=false
	End Object
	MuzzleFlash=MuzzleFlash0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlashLight Name=MuzzleFlashLightComponent0
		LightOffset=(X=20.0,Y=0.0,Z=0.0)
		TimeShift=((StartTime=0.0,Radius=500,Brightness=-1.0,LightColor=(R=255,G=255,B=255,A=255)),(StartTime=1.5,Radius=200,Brightness=0,LightColor=(R=255,G=255,B=255,A=255)))
	End Object
	MuzzleFlashLight=MuzzleFlashLightComponent0

	ZapLightTimeShift=((StartTime=0.0,Radius=200,Brightness=5,LightColor=(R=255,G=170,B=42,A=255)),(StartTime=0.2,Radius=64,Brightness=0,LightColor=(R=255,G=170,B=42,A=255)))
}
