//=============================================================================
// Ordinary rapid-fire, instant-hit weapon.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeapon_Minigun extends SentinelWeapon;

var() SentinelWeaponComponent_MuzzleFlash Tracer;
var() SentinelWeaponComponent_HitEffects HitEffects;
var() const AudioComponent WindUpSound;
var() const AudioComponent RunningSound;
var() const AudioComponent WindDownSound;

var() int NumBarrels;
/** Array of barrel distances radially from the centre. The array size must be equal to NumBarrels. */
var() array<float> BarrelOffsets;
/** A tracer will be spawned every this many shots. */
var() int TracerInterval;
/** This many seconds must pass after Sentinel goes idle before the minigun spins down. */
var() float WindDownDelay;

/** How much more likely to attack flying targets. */
var() float FlyingTargetPreference;

/** Location of last hit, or 0 if not fired. */
var Vector LastHitLocation; //Might be wrong if packets arrive out of order.
/** 0 = Idle, 1 = WindingUp, 2 = Running, 3 = WindingDown */
var protected repnotify byte FiringState;

var SkelControlSingleBone SkelControl;
var float CurrentBarrelRoll; //Do a barrel roll!

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		LastHitLocation, FiringState;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'FiringState')
	{
		//n.b. This may behave strangely if network latency is extremely high.
		switch FiringState
		{
			case 0:
				GotoState('Idle');
				break;
			case 1:
				GotoState('WindingUp');
				break;
			case 2:
				GotoState('Running');
				break;
			case 3:
				GotoState('WindingDown');
				break;
			default:
				GotoState('Idle');
		}
	}
	else
	{
		super.ReplicatedEvent(VarName);
	}
}

simulated function InitMuzzleFlash()
{
	super.InitMuzzleFlash();

	if(Cannon != none && Tracer != none)
	{
		Tracer.Initialize(Cannon.WeaponComponent, MuzzleSocketName);
	}

	SetMuzzleFlashOffset();
}

function ModifyPawnTargetPriority(out float Weight, Pawn PawnTarget)
{
	//The minigun, being instant-hit, is particularly useful against flying targets.
	if(PawnTarget.Physics == PHYS_Flying || PawnTarget.bCanFly)
	{
		Weight += FlyingTargetPreference;
	}

	super.ModifyPawnTargetPriority(Weight, PawnTarget);
}

simulated function FlashMuzzleFlash()
{
	if(Cannon != none)
	{
		SetMuzzleFlashOffset();

		super.FlashMuzzleFlash();

		if(!IsZero(LastHitLocation))
		{
			FireInfo.CheckBulletWhip(Cannon.GetPawnViewLocation(), LastHitLocation);

			if(HitEffects != none)
			{
				HitEffects.PlayImpactEffects(Cannon.GetPawnViewLocation(), LastHitLocation, self);
			}

			if(FlashCount % TracerInterval == 0 && Tracer != none)
			{
				Tracer.SetVectorParameter('TracerEnd', LastHitLocation);
				Tracer.Flash();
				Tracer.Initialize(Cannon.WeaponComponent, MuzzleSocketName); //Hack: Ensures tracer start location is in the right place.
			}
		}
	}
}

/**
 * Finds the nearest barrel to the current rotation and aligns MuzzleFlashOffset.Z with it.
 */
simulated function SetMuzzleFlashOffset()
{
	local float RollSpacing;
	local int NearestBarrelIndex;

	CurrentBarrelRoll = CurrentBarrelRoll % 65536.0;
	RollSpacing = 65536.0 / NumBarrels;
	NearestBarrelIndex = Round(Abs(CurrentBarrelRoll) / RollSpacing);
	NearestBarrelIndex = NearestBarrelIndex % NumBarrels;

	MuzzleFlash.MuzzleFlashOffset.Z = BarrelOffsets[NearestBarrelIndex];
	MuzzleFlash.SetTranslation(MuzzleFlash.MuzzleFlashOffset);

	if(Tracer != none)
	{
		Tracer.MuzzleFlashOffset.Z = MuzzleFlash.MuzzleFlashOffset.Z;
		Tracer.SetTranslation(MuzzleFlash.MuzzleFlashOffset);
	}
}

auto state Idle
{
	function BeginState(name PreviousStateName)
	{
		FiringState = 0;
		LastHitLocation = vect(0.0, 0.0, 0.0);
	}

	//Start winding up as soon as a target appears.
	function NotifyNewTarget(Actor NewTarget)
	{
		super.NotifyNewTarget(NewTarget);

		GotoState('WindingUp');
	}

	function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
	{
		GotoState('WindingUp');

		return false;
	}
}

state WindingUp
{
	simulated function BeginState(name PreviousStateName)
	{
		if(Role == ROLE_Authority)
		{
			FiringState = 1;
		}

		WindUpSound.Play();
		//Can't use the AudioComponent.OnAudioFinished delegate because it doesn't work on dedicated servers, so use a timer instead.
		SetTimer(WindUpSound.SoundCue.GetCueDuration(), false);
		SkelControl = SkelControlSingleBone(Cannon.WeaponComponent.FindSkelControl('BarrelRotate'));
	}

	simulated function Tick(float DeltaTime)
	{
		local float RollRate;

		if(SkelControl != none)
		{
			//Simple linear acceleration.
			RollRate = -65536.0 / (NumBarrels * FireInfo.FireInterval);
			RollRate *= GetTimerCount() / GetTimerRate();
			CurrentBarrelRoll += RollRate * DeltaTime;
			SkelControl.BoneRotation.Roll = CurrentBarrelRoll;
		}
	}

	function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
	{
		return false;
	}

	simulated function Timer()
	{
		if(Cannon != none && Cannon.Health > 0 && Cannon.bTracking)
		{
			GotoState('Running');
		}
		else
		{
			GotoState('WindingDown');
		}
	}
}

state Running
{
	simulated function BeginState(Name PreviousStateName)
	{
		if(Role == ROLE_Authority)
		{
			FiringState = 2;
		}

		RunningSound.Play();
	}

	simulated function Tick(float DeltaTime)
	{
		if(SkelControl != none)
		{
			CurrentBarrelRoll += (-65536.0 / (NumBarrels * FireInfo.FireInterval)) * DeltaTime;
			SkelControl.BoneRotation.Roll = CurrentBarrelRoll;
		}
	}

	function NotifyWaiting()
	{
		global.NotifyWaiting();

		SetTimer(WindDownDelay, false, 'WindDown');
	}

	function NotifyNewTarget(Actor NewTarget)
	{
		global.NotifyNewTarget(NewTarget);

		ClearTimer('WindDown');
	}

	function NotifyDied(Controller Killer, out class<DamageType> DamageType, vector HitLocation)
	{
		global.NotifyDied(Killer, DamageType, HitLocation);

		GotoState('WindingDown');
	}

	function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
	{
		local Vector TraceStart, TraceEnd;
		local Actor HitActor;
		local Vector HitLocation, HitNormal;

		TraceStart = Start;
		TraceEnd = Vector(BarrelDir) + (Normal(VRand()) * FireInfo.Spread);
		TraceEnd = Normal(TraceEnd);
		TraceEnd *= FireInfo.MaxRange;
		TraceEnd += TraceStart;

		HitActor = Cannon.Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true, vect(0,0,0),, TRACEFLAG_Bullet);

		if(HitActor == none)
		{
			HitLocation	= TraceEnd;
		}
		else if(Pawn(HitActor) == none || !Cannon.IsSameTeam(Pawn(HitActor)))
		{
			HitActor.TakeDamage(FireInfo.Damage, Cannon.InstigatorController, HitLocation, FireInfo.Momentum * Normal(TraceEnd - TraceStart), FireInfo.DamageType,, Cannon);
			Cannon.NotifyCausedDamage(HitActor);
		}

		LastHitLocation = HitLocation;

		super.FireAt(Start, BarrelDir, End);

		return true;
	}

	function WindDown()
	{
		GotoState('WindingDown');
	}

	simulated function EndState(name NextStateName)
	{
		RunningSound.Stop();
	}
}

state WindingDown
{
	simulated function BeginState(name PreviousStateName)
	{
		if(Role == ROLE_Authority)
		{
			FiringState = 3;
		}

		WindDownSound.Play();
		SetTimer(WindDownSound.SoundCue.GetCueDuration(), false);
	}

	simulated function Tick(float DeltaTime)
	{
		local float RollRate;

		if(SkelControl != none)
		{
			//Quadratic deceleration.
			RollRate = -65536.0 / (NumBarrels * FireInfo.FireInterval);
			RollRate *= Square((GetTimerRate() - GetTimerCount()) / GetTimerRate());
			CurrentBarrelRoll += RollRate * DeltaTime;
			SkelControl.BoneRotation.Roll = CurrentBarrelRoll;
		}
	}

	function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
	{
		return false;
	}

	simulated function Timer()
	{
		if(Cannon.bTracking)
		{
			GotoState('WindingUp');
		}
		else
		{
			GotoState('Idle');
		}
	}
}

defaultproperties
{
	FriendlyName="Minigun"
	Description="In acknowledgement of the continuing popularity of kinetic weapons, Evil Corp. has developed a new minigun attachment for the Mk III Sentinel. Incorporating patented \"3-wave\" technology, the new minigun exhibits far higher barrel strength and thermal capacity compared to previous models, allowing for an indefinitely sustainable high rate of fire*.\n\nThe 15mm aluminium rounds fired by this weapon exhibit hypersonic muzzle velocity and maintain that velocity with the use of miniature gravitomagnetic thrusters, giving exceptional damage potential against lightly armoured targets at all ranges. Standard configuration is one tracer every fourth round.\n\n\n*Non-military model limited to 600rpm to meet NEG safety guidelines."

	NumBarrels=9
	BarrelOffsets[0]=15.5
	BarrelOffsets[1]=11.0
	BarrelOffsets[2]=7.0
	BarrelOffsets[3]=15.5
	BarrelOffsets[4]=11.0
	BarrelOffsets[5]=7.0
	BarrelOffsets[6]=15.5
	BarrelOffsets[7]=11.0
	BarrelOffsets[8]=7.0

	TracerInterval=4
	WindDownDelay=3.0
	FlyingTargetPreference=0.5

	WeaponMesh=SkeletalMesh'Sentinel_Resources.Meshes.Weapons.SK_Minigun'
	PhysicsAsset=PhysicsAsset'Sentinel_Resources.Physics.SK_Minigun_Physics'
	AnimTreeTemplate=AnimTree'Sentinel_Resources.Meshes.Weapons.AT_Minigun'
	WeaponMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel'
	WeaponSpawnMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Spawn'
	WeaponDeadMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Dead'
	WeaponScale=0.6

	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Weapons.SwitchToMiniGunCue'

	ExtraStrength=0.1
	AmmoCost=50
	CompatibleSentinels.Remove(class'Sentinel_Ceiling')

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		FireInterval=0.1
		FireOffset=(X=70.0,Y=0.0,Z=10.0)
		Damage=6.0
		Momentum=5000.0
		DamageType=class'UTDmgType_SentinelMinigun'
		MaxRange=20000.0
		Spread=0.03
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.BMiniGunAltFireCue'
	End Object
	FireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlash Name=MuzzleFlash0
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Minigun_MuzzleFlash'
		MuzzleFlashDuration=1.25
		MuzzleFlashOffset=(X=2.0,Y=0.0,Z=0.0)
	End Object
	MuzzleFlash=MuzzleFlash0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlashLight Name=MuzzleFlashLightComponent0
		LightOffset=(X=30.0,Y=0.0,Z=0.0)
		TimeShift=((StartTime=0.0,Radius=200,Brightness=5,LightColor=(R=255,G=190,B=64,A=255)),(StartTime=0.8,Radius=64,Brightness=0,LightColor=(R=255,G=190,B=64,A=255)))
	End Object
	MuzzleFlashLight=MuzzleFlashLightComponent0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlash Name=TracerComp
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_MiniGun_Tracer'
	End Object
	Tracer=TracerComp

	Begin Object Class=SentinelWeaponComponent_HitEffects Name=HitEffectsComp
	End Object
	HitEffects=HitEffectsComp

	Begin Object Class=AudioComponent Name=WindUpSound0
		SoundCue=SoundCue'Sentinel_Resources.Sounds.Weapons.SMinigunWindUpCue'
		bStopWhenOwnerDestroyed=true
	End Object
	WindUpSound=WindUpSound0
	Components.Add(WindUpSound0)

	Begin Object Class=AudioComponent Name=RunningSound0
		SoundCue=SoundCue'Sentinel_Resources.Sounds.Weapons.SMinigunLoopCue'
		bStopWhenOwnerDestroyed=true
	End Object
	RunningSound=RunningSound0
	Components.Add(RunningSound0)

	Begin Object Class=AudioComponent Name=WindDownSound0
		SoundCue=SoundCue'Sentinel_Resources.Sounds.Weapons.SMinigunWindDownCue'
		bStopWhenOwnerDestroyed=true
	End Object
	WindDownSound=WindDownSound0
	Components.Add(WindDownSound0)
}
