//=============================================================================
// Mortar that fires projectiles with variable velocity instead of the usual
// variable elevation. It can also be user-configured to fire different types
// of shell.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeapon_MultiMortar extends SentinelWeapon;

var() const StaticMeshComponent CurrentShellComponent;
var() const StaticMeshComponent NextShellComponent;

/** One FireInfo per type of shell that the mortar can fire. */
var() array<SentinelWeaponComponent_FireInfo> ShellTypes;

/** Launch pitch relative to Sentinel, in radians. */
var() float DesiredElevation;
/** Highest elevation settable via menu, in radians. */
var() float MaxElevation;
/** Lowest elevation settable via menu, in radians. */
var() float MinElevation;
/** Maximum rate at which Elevation can rotate to DesiredElevation, in radians per second. */
var() float ElevationRate;
/** Current pitch of barrel relative to Sentinel in radians. */
var float Elevation;

var() float MinimumLaunchSpeed;
var() float MaximumLaunchSpeed;

/** Targets that the current shell type are effective against will be preferred by this much. */
var() float TargetEffectivenessPreference;

var() int AdjustAimIterations;
var() int AdjustAimMaxError;

var SkelControlSingleBone BaseSkelControl;

var repnotify bool bHideShells;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		DesiredElevation, bHideShells;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'bHideShells')
	{
		if(bHideShells)
		{
			HideShells();
		}
	}
	else
	{
		super.ReplicatedEvent(VarName);
	}
}

simulated function PostBeginPlay()
{
	super.PostBeginPlay();

	if(Role < ROLE_Authority)
	{
		//Hack warning! Done on a timer because Owner isn't replicated immediately.
		SetTimer(1.05, false, 'GetNewShellTypes');
	}
}

simulated function InitializeWeaponComponent(SentinelComponent_Mesh WeaponComponent)
{
	super.InitializeWeaponComponent(WeaponComponent);

	InitializeShellComponents();
}

simulated function InitializeShellComponents()
{
	local StaticMeshComponent ProjectileMesh;

	if(WorldInfo.NetMode == NM_DedicatedServer)
	{
		return;
	}

	ProjectileMesh = class<SentinelProjectile_Mortar>(FireInfo.ProjectileClass).default.Mesh;

	CurrentShellComponent.SetStaticMesh(none); //Set mesh to none while scale is changed to avoid "cooking convex".
	CurrentShellComponent.SetMaterial(0, ProjectileMesh.Materials[0]);
	CurrentShellComponent.SetScale(ProjectileMesh.Scale);
	CurrentShellComponent.SetScale3D(ProjectileMesh.Scale3D);
	CurrentShellComponent.SetLightEnvironment(Cannon.WeaponComponent.LightEnvironment);
	CurrentShellComponent.SetStaticMesh(ProjectileMesh.StaticMesh);

	NextShellComponent.SetStaticMesh(none);
	NextShellComponent.SetMaterial(0, ProjectileMesh.Materials[0]);
	NextShellComponent.SetScale(ProjectileMesh.Scale);
	NextShellComponent.SetScale3D(ProjectileMesh.Scale3D);
	NextShellComponent.SetLightEnvironment(Cannon.WeaponComponent.LightEnvironment);
	NextShellComponent.SetStaticMesh(ProjectileMesh.StaticMesh);

	Cannon.WeaponComponent.AttachComponentToSocket(CurrentShellComponent, 'CurrentShell');
	Cannon.WeaponComponent.AttachComponentToSocket(NextShellComponent, 'NextShell');
}

function NotifyDied(Controller Killer, out class<DamageType> DamageType, Vector HitLocation)
{
	super.NotifyDied(Killer, DamageType, HitLocation);

	bHideShells = true;
	HideShells();
}

simulated function NotifyInvisible(bool bInvisible)
{
	HideShells(bInvisible);
}

simulated function HideShells(optional bool bHide=true)
{
	CurrentShellComponent.SetHidden(bHide);
	NextShellComponent.SetHidden(bHide);
}

/**
 * Instructs the server to synchronize the owning client's ShellTypes array with the server's version.
 */
reliable server function GetNewShellTypes()
{
	local int i;

	SetShellTypesLength(ShellTypes.Length);

	for(i = 0; i < ShellTypes.Length; i++)
	{
		if(i >= default.ShellTypes.Length || ShellTypes[i].ProjectileClass != default.ShellTypes[i].ProjectileClass)
		{
			ClientAddShellType(SentinelWeaponComponent_FireInfo(ShellTypes[i].ObjectArchetype), i, ShellTypes[i] == FireInfo);
		}
	}
}

/**
 * Synchronizes the length of the owning client's ShellTypes array with the server's.
 */
protected reliable client function SetShellTypesLength(int NewLength)
{
	ShellTypes.Length = NewLength;
}

/**
 * Sends a new shell type to the owning client. In order for the replication to work, it must be a static, i.e. default object.
 */
protected reliable client function ClientAddShellType(SentinelWeaponComponent_FireInfo NewShellType, int NewIndex, bool bSetAsCurrent)
{
	AddShellType(NewShellType, NewIndex, bSetAsCurrent);
}

/**
 * Adds a copy of NewShellType to the ShellTypes array, if one is not already present.
 *
 * @param		NewShellType	type of shell to add. Must be a static (default) object so it can be replicated. Only replicated to the owner, so FireSound and other client-side variables should not be changed.
 * @param		NewIndex		index into ShellTypes array to add the new shell. Should usually be left blank (defaults to -1), which results in the shell being added to the end of the array.
 * @param		bSetAsCurrent	if true, NewShellType will be assigned to FireInfo, even if it already existed in the array.
 */
simulated function AddShellType(SentinelWeaponComponent_FireInfo NewShellType, optional int NewIndex=-1, optional bool bSetAsCurrent)
{
	local SentinelWeaponComponent_FireInfo ShellType;
	local bool bAlreadyAdded;
	local SentinelWeaponComponent_FireInfo NewShellTypeClone;

	if(NewShellType == none)
	{
		`warn("Tried to add none");
		return;
	}

	foreach ShellTypes(ShellType)
	{
		if(ShellType != none && ShellType.ProjectileClass == NewShellType.ProjectileClass)
		{
			if(bSetAsCurrent)
			{
				FireInfo = ShellType;
			}

			bAlreadyAdded = true;
			break;
		}
	}

	if(!bAlreadyAdded)
	{
		NewShellTypeClone = new NewShellType.Class (NewShellType);

		if(NewShellTypeClone != none)
		{
			if(NewIndex < 0)
			{
				NewIndex = ShellTypes.Length;
			}
			else
			{
				if(NewIndex >= ShellTypes.Length)
				{
					ShellTypes.Length = NewIndex + 1;
				}
			}

			ShellTypes[NewIndex] = NewShellTypeClone;

			if(bSetAsCurrent)
			{
				FireInfo = ShellTypes[NewIndex];
			}
		}
		else
		{
			`warn("Failed to clone NewShellType:"@NewShellType);
		}
	}
}

/**
 * Changes the current FireInfo to the one from ShellTypes with the given index. Does nothing if the given index is invalid.
 */
reliable server function SetShellType(int NewShellTypeIndex)
{
	if(NewShellTypeIndex < 0 || NewShellTypeIndex > ShellTypes.length - 1)
	{
		`warn("Tried to select invalid shell type:"@NewShellTypeIndex);
		return;
	}

	FireInfo = ShellTypes[NewShellTypeIndex];
	SetFireMode();
	ClientSetShellType(NewShellTypeIndex);
}

/**
 * Called by the server to ensure that the client's menu can retrieve the correct current FireInfo.
 */
reliable client function ClientSetShellType(int NewShellTypeIndex)
{
	FireInfo = ShellTypes[NewShellTypeIndex];
	InitializeShellComponents();
}

/**
 * Sets DesiredElevation, checking that the value is within the allowed range.
 */
reliable server function SetDesiredElevation(float NewDesiredElevation)
{
	DesiredElevation = FClamp(NewDesiredElevation, MinElevation, MaxElevation);
}

simulated function Tick(float DeltaTime)
{
	local float DeltaElevation;
	local float URotElevation;

	super.Tick(DeltaTime);

	DeltaElevation = DesiredElevation - Elevation;

	if(Abs(DeltaElevation) > ElevationRate * DeltaTime)
	{
		DeltaElevation /= Abs(DeltaElevation);
		DeltaElevation *= ElevationRate * DeltaTime;
	}

	Elevation += DeltaElevation;

	if(Cannon != none)
	{
		URotElevation = Elevation * (32768.0 / PI);

		Cannon.MaxPitch = URotElevation;
		Cannon.MinPitch = URotElevation;

		if(BaseSkelControl != none)
		{
			BaseSkelControl.BoneRotation.Pitch = 16384 - Cannon.WeaponComponent.Rotation.Pitch;
		}
	}
}

simulated function Destroyed()
{
	if(Cannon != none)
	{
		Cannon.MaxPitch = Cannon.default.MaxPitch;
		Cannon.MinPitch = Cannon.default.MinPitch;
	}
}

/**
 * If the mortar is not on perfectly level ground, it will be pointed the wrong way due to the barrel's elevation. This function corrects the aim.
 * Unfortunately I can't work out how to calculate the answer algebraically, so it is solved iteratively instead. It's slower and less accurate this way, but good enough.
 */
function AdjustAimToHit(Actor A, out Vector AimSpot, out Rotator AimRotation)
{
	local Vector X, Y, Z;
	local Rotator BarrelDir;
	local int YawIncrement;
	local int DeltaYaw;
	local int i;

	DeltaYaw = 0;

	for(i = 0; i < AdjustAimIterations; i++)
	{
		X = class'SUtils'.static.ClosestPointToPlane(Vector(AimRotation), Cannon.BaseComponent.Rotation);
		Z = vect(0.0, 0.0, 1.0) >> Cannon.BaseComponent.Rotation;
		Y = Z cross X;
		BarrelDir = OrthoRotation(X, Y, Z);
		BarrelDir = class'SUtils'.static.RotateRelative(BarrelDir, 0.0, DeltaYaw, 0.0);
		BarrelDir = class'SUtils'.static.RotateRelative(BarrelDir, (Elevation / (2.0 * PI)) * 65536.0, 0.0, 0.0);

		if(Abs(Normalize(BarrelDir - AimRotation).Yaw) < AdjustAimMaxError || Abs(DeltaYaw) >= 16384)
		{
			break;
		}

		YawIncrement = Normalize(AimRotation - BarrelDir).Yaw;
		DeltaYaw += YawIncrement;
	}

	AimRotation.Yaw += DeltaYaw;
}

function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
{
	local Vector X, Y, Z;
	local Vector SpawnLocation;
	local float CurrentElevation;
	local float LaunchSpeed;
	local SentinelProjectile_Mortar Proj;
	local bool bFired;

	CurrentElevation = (BarrelDir.Pitch / 65536.0) * 2.0 * PI;
	GetAxes(BarrelDir, X, Y, Z);

	SpawnLocation = Start;
	SpawnLocation += X * FireInfo.FireOffset.X;
	SpawnLocation += Y * FireInfo.FireOffset.Y;
	SpawnLocation += Z * FireInfo.FireOffset.Z;

	LaunchSpeed = CalculateSpeedToReach(SpawnLocation, End, CurrentElevation);

	if(LaunchSpeed >= MinimumLaunchSpeed && LaunchSpeed <= MaximumLaunchSpeed)
	{
		if(ProjectileCanHit(End, SpawnLocation, LaunchSpeed, CurrentElevation, vect(0.0, 0.0, 0.0), FireInfo.ProjectileClass.default.MaxSpeed))
		{
			Proj = SentinelProjectile_Mortar(Cannon.Spawn(FireInfo.ProjectileClass, Cannon,, SpawnLocation, BarrelDir));

			if(Proj != none)
			{
				InitializeProjectile(Proj, Vector(BarrelDir) * LaunchSpeed);
				bFired = super.FireAt(Start, BarrelDir, End);
			}
		}
	}

	return bFired;
}

simulated function FlashMuzzleFlash()
{
	super.FlashMuzzleFlash();

	if(Cannon != none)
	{
		Cannon.WeaponComponent.PlayAnim('Firing');
	}
}

simulated function PostInitAnimTree(SkeletalMeshComponent SkelComp)
{
	BaseSkelControl = SkelControlSingleBone(SkelComp.FindSkelControl('BaseRotate'));
}

simulated function OnAnimEnd(AnimNodeSequence SeqNode, float PlayedTime, float ExcessTime)
{
	if(SeqNode.AnimSeqName == 'Firing')
	{
		Cannon.WeaponComponent.PlayAnim('Reloading');
		PlaySound(SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_ReloadCue', true,, true, Cannon.WeaponComponent.GetPosition());
	}
}

/**
 * @return	true if there is a path from Start to Destination for a ballistic projectile fired at TossSpeed and TossElevation.
 */
final function bool ProjectileCanHit(Vector Destination, vector Start, float TossSpeed, float TossElevation, optional vector CollisionSize, optional float TerminalVelocity)
{
	local Vector TossVelocity;
	local float DesiredZPct;

	//SuggestTossVelocity will fail erroneously if TossElevation is close to 45 degrees and TossSpeed is precisely enough to reach Destination, so a fudge-factor must be applied to help it.
	TossSpeed *= (1.05 - (0.05 * FMin((Abs(TossElevation - 0.785) / 0.065), 1.0)));
	DesiredZPct = Sin(TossElevation) / (Sin(TossElevation) + Cos(TossElevation));

	return SuggestTossVelocity(TossVelocity, Destination, Start, TossSpeed, 0.0, DesiredZPct, CollisionSize, TerminalVelocity);
}


/**
 * Calculates the speed that a projectile needs to reach End from Start if launched at Angle and influenced only by gravity.
 *
 * @param	Start	location projectile is fired from
 * @param	End		target location
 * @param	Angle	launch angle in radians
 */
final function float CalculateSpeedToReach(Vector Start, Vector End, float Angle)
{
	//Taken from
	//http://en.wikipedia.org/wiki/Range_of_a_projectile
	//and solved for v instead of d.

	local float d; //Horizontal distance to target.
	local float h; //Initial height above target.
	local float g; //Acceleration due to gravity.
	local float v; //Required speed at the given angle.

	d = VSize((Start - End) * vect(1.0, 1.0, 0.0));
	h = Start.Z - End.Z;
	g = -Cannon.GetGravityZ() * 2.0; //I don't know why it needs the extra factor of two. Maybe I messed up the algebra somewhere.

	v = 2.0 * Cos(Angle);
	v *= (h * Cos(Angle)) + (d * Sin(Angle));
	v = (d * d * g) / v;
	v = v >= 0.0 ? Sqrt(v) : 0.0; //If calculated v is not a real number, then return 0 because it's impossible to actually fire.

	return v;
}

function ModifyPawnTargetPriority(out float Weight, Pawn PawnTarget)
{
	local float LaunchSpeed;
	local array<EffectivenessInfo> ShellEffectiveness;
	local int i;

	LaunchSpeed = CalculateSpeedToReach(Cannon.Location, PawnTarget.Location, Elevation);

	//Don't want to fire on targets that are too close or too distant.
	if(LaunchSpeed < MinimumLaunchSpeed || LaunchSpeed > MaximumLaunchSpeed)
	{
		Weight -= 2.0;
	}
	//Don't want to fire on targets that cannot be hit. This is a guess, since some targets may only appear to be unreachable due to the Sentinel being tilted and pointed away). A more accurate test is possible, but I don't think it would be worth it.
	else if(!ProjectileCanHit(PawnTarget.Location, Cannon.Location, LaunchSpeed, Elevation, vect(0.0, 0.0, 0.0), FireInfo.ProjectileClass.default.MaxSpeed))
	{
		Weight -= 2.0;
	}
	//Prefer targets that the currently selected shell type is effective against.
	else
	{
		ShellEffectiveness = class<SentinelProjectile_Mortar>(FireInfo.ProjectileClass).default.Effectiveness;

		if(Vehicle(PawnTarget) != none)
		{
			if(UTVehicle(PawnTarget) != none && UTVehicle(PawnTarget).bLightArmor)
			{
				i = ShellEffectiveness.Find('TargetType', TT_LightVehicle);
			}
			else
			{
				i = ShellEffectiveness.Find('TargetType', TT_HeavyVehicle);
			}
		}
		else if(Sentinel(PawnTarget) != none)
		{
			i = ShellEffectiveness.Find('TargetType', TT_Sentinel);
		}
		else
		{
			//Assume that anything that is not a vehicle or a Sentinel must be infantry.
			i = ShellEffectiveness.Find('TargetType', TT_Infantry);
		}

		if(i != INDEX_NONE)
		{
			Weight += ShellEffectiveness[i].Effectiveness;
		}
	}

	super.ModifyPawnTargetPriority(Weight, PawnTarget);
}

/**
 * Calculates time accounting for arcing path, unlike the standard version which assumes the projectile travels directly to the target.
 */
simulated function float GetProjectileTimeToLocation(Vector TargetLoc, Vector StartLoc, Controller RequestedBy)
{
	local float LaunchSpeed;
	local float HorizontalDistance;
	local float HorizontalSpeed;
	local float TravelTime;

	LaunchSpeed = CalculateSpeedToReach(StartLoc, TargetLoc, Elevation);

	if(LaunchSpeed > 0.0)
	{
		HorizontalDistance = VSize((TargetLoc - StartLoc) * vect(1.0, 1.0, 0.0));
		HorizontalSpeed = Cos(Elevation) * LaunchSpeed;
		TravelTime = HorizontalDistance / HorizontalSpeed;
	}

	return TravelTime;
}

defaultproperties
{
	AdjustAimIterations=10
	AdjustAimMaxError=32

	FriendlyName="Multi-Mortar"
	Description="The Multi-Mortar is a fully automated artillery piece, easily reconfigurable in the field. Advanced micro-fabrication technology allows the Multi-Mortar to assemble shells on demand using a variety of payloads, at the cost of a reduced firing rate for the more complex types.\n\nShells are launched with variable velocity, minimizing the quantity of propellant needed per shot. There is a minimum launch speed, and thus a minimum range, so the Multi-Mortar should not be used where enemies are expected to be encountered at close range. Elevation can be adjusted manually if required.\n\nThe Multi-mortar can be fired indirectly at targets that are hidden or outside the Sentinel's normal sensor range. Targets can be designated by other Sentinels via the \"Mesh Network" upgrade.\n\nPlease note that if the Multi-Mortar is not deployed on level ground then its accuracy may be impaired, or it may be unable to fire at all."

	ConfigurationMenuTemplate=UTUIScene_MultiMortarMenu'Sentinel_Resources.Menus.MultiMortarMenu'

	DesiredElevation=1.1
	MaxElevation=1.4
	MinElevation=0.44
	ElevationRate=0.625
	MinimumLaunchSpeed=1000.0
	MaximumLaunchSpeed=5000.0
	TargetEffectivenessPreference=1.0

	WeaponMesh=SkeletalMesh'Sentinel_Resources.Meshes.Weapons.SK_MultiMortar'
	PhysicsAsset=PhysicsAsset'Sentinel_Resources.Physics.SK_MultiMortar_Physics'
	AnimTreeTemplate=AnimTree'Sentinel_Resources.Meshes.Weapons.AT_MultiMortar'
	AnimSet=AnimSet'Sentinel_Resources.Meshes.Weapons.AS_MultiMortar'
	WeaponMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel'
	WeaponSpawnMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Spawn'
	WeaponDeadMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Dead'
	WeaponScale=0.6

	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_ActivateCue'

	bCanFireBlind=true
	AmmoCost=70
	CompatibleSentinels.Remove(class'Sentinel_Ceiling')
	MaxDesirability=-1.0 //TODO: Try to work out when bots can use this. For now, just don't let them use it at all.

	Begin Object Class=StaticMeshComponent Name=ShellComponent0
		AbsoluteScale=true
		CollideActors=false
		BlockActors=false
		BlockZeroExtent=false
		BlockNonZeroExtent=false
		CastShadow=false
		BlockRigidBody=false
		bAcceptsDecals=false
		bAcceptsDecalsDuringGameplay=false
		bUseAsOccluder=false
	End Object
	CurrentShellComponent=ShellComponent0
	Components.Add(ShellComponent0)

	Begin Object Class=StaticMeshComponent Name=ShellComponent1
		AbsoluteScale=true
		CollideActors=false
		BlockActors=false
		BlockZeroExtent=false
		BlockNonZeroExtent=false
		CastShadow=false
		BlockRigidBody=false
		bAcceptsDecals=false
		bAcceptsDecalsDuringGameplay=false
		bUseAsOccluder=false
	End Object
	NextShellComponent=ShellComponent1
	Components.Add(ShellComponent1)

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		FireInterval=3.0
		ProjectileClass=class'SentinelProjectile_Mortar_HE'
		MaxAimError=20.0
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_FireCue'
	End Object
	ShellTypes.Add(FireInfo0)
	FireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo1
		FireInterval=3.0
		ProjectileClass=class'SentinelProjectile_Mortar_AT'
		MaxAimError=30.0
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_FireCue'
	End Object
	ShellTypes.Add(FireInfo1)

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo2
		FireInterval=4.0
		ProjectileClass=class'SentinelProjectile_Mortar_Shrapnel'
		MaxAimError=20.0
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.Mortar_FireCue'
	End Object
	ShellTypes.Add(FireInfo2)

	Begin Object Class=SentinelWeaponComponent_MuzzleFlash Name=MuzzleFlash0
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_Mortar_MuzzleFlash'
		MuzzleFlashDuration=0.15
		MuzzleFlashOffset=(X=10.0,Y=0.0,Z=1.0)
	End Object
	MuzzleFlash=MuzzleFlash0
}
