//=============================================================================
// Link beam capable of healing vehicles/nodes, but cannot damage enemies.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeapon_PulseToLinkConverter extends SentinelWeapon;

/** Sound played while beam is active. */
var() UTAmbientSoundComponent BeamSoundComponent;
/** Some tube-shaped mesh emitters to make this weapon visually different to the Pulse Cannon. */
var() ParticleSystemComponent TubesComponent;

var() editconst SentinelControllerComponent_Behaviour_TargetingPawn OriginalPawnTargetingBehaviour;
var() editconst SentinelControllerComponent_Behaviour_TargetingObjective OriginalObjectiveTargetingBehaviour;
var() SentinelControllerComponent_Behaviour_HealingVehicle NewPawnTargetingBehaviour;
var() SentinelControllerComponent_Behaviour_HealingObjective NewObjectiveTargetingBehaviour;

/** Time after firing ceases before unlinking from target. */
var() float UnlinkDelay;
/** Name of PSC parameter controlling location of end of beam. */
var() name BeamEndParameterName;

/** Holds the actor that this weapon is linked to. */
var repnotify Actor LinkedTo;

/** Set to Cannon.TeamColour in InitMuzzleFlash. Used for beam. */
var LinearColor EffectColour;

/** Set true when the Sentinel has died, so that tube component can be disabled. */
var repnotify bool bKillTubes;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		LinkedTo, bKillTubes;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'LinkedTo')
	{
		if(LinkedTo != none)
			Link();
		else
			UnLink();
	}
	if(VarName == 'bKillTubes')
	{
		if(bKillTubes && TubesComponent != none)
		{
			TubesComponent.DeactivateSystem();
		}
	}
	else
	{
		super.ReplicatedEvent(VarName);
	}
}

function InitializeFor(Sentinel S)
{
	super.InitializeFor(S);

	NewPawnTargetingBehaviour = new(S.SController) class'SentinelControllerComponent_Behaviour_HealingVehicle';
	NewObjectiveTargetingBehaviour = new(S.SController) class'SentinelControllerComponent_Behaviour_HealingObjective';

	//Set controller to idle and swap the normal enemy-attacking behaviour for new ally-healing behaviour.
	S.SController.ChangeBehaviourTo(S.SController.IdleBehaviour);
	OriginalPawnTargetingBehaviour = S.SController.PawnTargetingBehaviour;
	OriginalObjectiveTargetingBehaviour = S.SController.ObjectiveTargetingBehaviour;
	S.SController.PawnTargetingBehaviour = NewPawnTargetingBehaviour;
	S.SController.ObjectiveTargetingBehaviour = NewObjectiveTargetingBehaviour;
	S.SController.bSeeFriendly = true;
}

simulated function ClientInitializeFor()
{
	super.ClientInitializeFor();

	if(TubesComponent != none)
	{
		Cannon.WeaponComponent.AttachComponent(TubesComponent, Cannon.WeaponComponent.RootBone);
	}
}

function OnDetach()
{
	super.OnDetach();

	Unlink();

	//Set controller to idle and revert to default behaviour.
	Cannon.SController.ChangeBehaviourTo(Cannon.SController.IdleBehaviour);
	Cannon.SController.bSeeFriendly = Cannon.SController.default.bSeeFriendly;
	Cannon.SController.PawnTargetingBehaviour = OriginalPawnTargetingBehaviour;
	Cannon.SController.ObjectiveTargetingBehaviour = OriginalObjectiveTargetingBehaviour;
	NewPawnTargetingBehaviour = none;
	NewObjectiveTargetingBehaviour = none;
}

simulated function NotifyTeamChanged()
{
	InitMuzzleFlash();
}

simulated function NotifyInvisible(bool bInvisible)
{
	if(TubesComponent != none)
	{
		if(bInvisible)
		{
			TubesComponent.DeactivateSystem();
		}
		else
		{
			TubesComponent.ActivateSystem();
		}
	}
}

function NotifyDied(Controller Killer, out class<DamageType> DamageType, Vector HitLocation)
{
	super.NotifyDied(Killer, DamageType, HitLocation);

	TubesComponent.DeactivateSystem();
	bKillTubes = true;
}

simulated function InitMuzzleFlash()
{
	super.InitMuzzleFlash();

	EffectColour = Cannon.TeamColour;

	if(Cannon != none)
	{
		if(MuzzleFlash != none)
		{
			MuzzleFlash.SetColour(EffectColour);
		}

		if(MuzzleFlashLight != none)
		{
			MuzzleFlashLight.SetColour(EffectColour);
		}

		if(TubesComponent != none)
		{
			TubesComponent.SetColorParameter('TubeColour', MakeColor(EffectColour.R * 255, EffectColour.G * 255, EffectColour.B * 255, EffectColour.A * 255));
		}
	}
}

function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
{
	local Vector TraceStart, TraceEnd;
	local Actor HitActor;
	local Vector HitLocation, HitNormal;
	local bool bFired;

	TraceStart = Start;
	TraceEnd = TraceStart;
	TraceEnd += Normal(End - Start) * FireInfo.MaxRange;

	HitActor = Cannon.Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true, vect(0,0,0),, TRACEFLAG_Bullet);

	if(HitActor != none && (NewPawnTargetingBehaviour.CanHeal(UTVehicle(HitActor)) || NewObjectiveTargetingBehaviour.IsValidObjective(UTGameObjective(HitActor))))
	{
		if(LinkedTo != HitActor)
		{
			UnLink();
			LinkedTo = HitActor;
			Link();
		}

		LinkedTo.HealDamage(FireInfo.Damage, Cannon.InstigatorController, FireInfo.DamageType);

		bFired = true;
		super.FireAt(Start, BarrelDir, End);
	}
	else
	{
		Unlink();
	}

	return bFired;
}

function FireTimer()
{
	super.FireTimer();

	SetTimer(UnlinkDelay, false, 'UnLink');
}

simulated function Link()
{
	if(Role == ROLE_Authority)
	{
		if(UTVehicle(LinkedTo) != none)
		{
			UTVehicle(LinkedTo).IncrementLinkedToCount();
		}
	}

	BeamSoundComponent.Play();
}

simulated function Unlink()
{
	if(Role == ROLE_Authority)
	{
		if(UTVehicle(LinkedTo) != none)
		{
			UTVehicle(LinkedTo).DecrementLinkedToCount();
		}

		LinkedTo = none;
	}

	BeamSoundComponent.Stop();
}

static function float BotDesirability(UTBot B, Sentinel S)
{
	local float Desirability;
	local UTOnslaughtObjective Objective;
	local bool bNearObjective;

	Desirability = super.BotDesirability(B, S);

	if(Desirability >= 0.0)
	{
		foreach B.WorldInfo.AllNavigationPoints(class'UTOnslaughtObjective', Objective)
		{
			if(VSize(Objective.Location - S.Location) < default.FireInfo.MaxRange && S.SController.LineOfSightTo(Objective))
			{
				bNearObjective = true;
				break;
			}
		}

		if(!bNearObjective)
		{
			Desirability = 0.0;
		}
	}

	return Desirability;
}

simulated function Tick(float DeltaTime)
{
	super.Tick(DeltaTime);

	if(LinkedTo != none)
	{
		MuzzleFlash.SetVectorParameter(BeamEndParameterName, GetLinkedToLocation());
	}
}

simulated function Vector GetLinkedToLocation()
{
	local Vector TraceStart, TraceEnd;
	local Vector HitLocation, HitNormal;
	local bool bHitVehicleAimingAt;
	local UTOnslaughtPowerNode PN;
	local float BestScore, Score;
	local name SocketName;
	local name BestSocket;
	local Vector LinkedToLocation;

	//Quite a bit of this is copied from UTWeap_LinkGun.uc
	Cannon.WeaponComponent.GetSocketWorldLocationAndRotation(MuzzleSocketName, TraceStart);

	if(Vehicle(LinkedTo) != none)
	{
		TraceEnd = LinkedTo.GetTargetLocation();

		if(UTVehicle_Walker(LinkedTo) != none && UTVehicle_Walker(LinkedTo).BodyActor != none)
		{
			// check to see if we are hitting the legs
			bHitVehicleAimingAt = TraceComponent(HitLocation, HitNormal, UTVehicle_Walker(LinkedTo).BodyActor.SkeletalMeshComponent, TraceEnd, TraceStart);
		}

		// if we didn't it the legs so maybe we hit the body or it wasn't a Walker at all!
		if(!bHitVehicleAimingAt)
		{
			bHitVehicleAimingAt = TraceComponent(HitLocation, HitNormal, UTVehicle(LinkedTo).Mesh, TraceEnd, TraceStart);
		}

		// check here to make certain we hit what we were tracing
		if(bHitVehicleAimingAt)
		{
			LinkedToLocation = HitLocation;
		}
		// else something terrible occurred so we will just have the effect be on the MuzzleFlashSocket otherwise it will bend around and look bad
		else
		{
			LinkedToLocation = TraceStart;
		}
	}
	else if(UTOnslaughtPowerNode(LinkedTo) != none && UTOnslaughtPowerNode(LinkedTo).LinkToSockets.length > 0)
	{
		PN = UTOnslaughtPowerNode(LinkedTo);

		foreach PN.LinkToSockets(SocketName)
		{
			PN.EnergySphere.GetSocketWorldLocationAndRotation(SocketName, TraceStart);
			TraceEnd = Cannon.GetPawnViewLocation() - TraceStart;
			Score = Abs(TraceEnd / VSize(TraceEnd) dot Vector(Cannon.GetViewRotation()));

			if(BestSocket == '' || Score > BestScore)
			{
				BestSocket = SocketName;
				HitLocation = TraceStart;
				BestScore = Score;
			}
		}

		LinkedToLocation = HitLocation;
	}
	else
	{
		LinkedToLocation = TraceStart;
	}

	return LinkedToLocation;
}

simulated function Destroyed()
{
	super.Destroyed();

	Unlink();
}

defaultproperties
{
	FriendlyName="Pulse-to-Link Converter"
	Description="This device remodulates the pulsed output of the standard Pulse Cannon into a continuous stream, similar to Riordan Dynamic Weapon Systems' \"Link Gun\". The beam operates in carrier mode only, useful for repairing vehicles etc. but not for offensive purposes. It is a permanent modification; you must install a new Pulse Cannon to revert to projectile mode.\n\nPlease note that this device is not officially supported and its use will void your warranty."

	UnlinkDelay=0.3
	BeamEndParameterName=BeamEnd

	WeaponMesh=SkeletalMesh'Sentinel_Resources.Meshes.Weapons.SK_PulseCannon'
	PhysicsAsset=PhysicsAsset'Sentinel_Resources.Physics.SK_PulseCannon_Physics'
	WeaponMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel'
	WeaponSpawnMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Spawn'
	WeaponDeadMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Dead'
	WeaponScale=0.6

	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Weapons.SwitchToLinkGunCue'

	AmmoCost=55
	RequiredUpgrades.Add(class'SentinelWeapon_PulseCannon')
	CompatibleSentinels.Remove(class'Sentinel_Ceiling')
	MaxDesirability=0.55

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		FireInterval=0.2
		Damage=12.0
		DamageType=class'UTDmgType_SentinelLink'
		MaxRange=800.0
		MaxAimError=25.0
	End Object
	FireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlash Name=MuzzleFlash0
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_LinkCannon_Beam'
		AbsoluteScale=true
		Scale=0.6
		MuzzleFlashDuration=0.25
		MuzzleFlashColourName=BeamColour
		bConstantFlash=true
	End Object
	MuzzleFlash=MuzzleFlash0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlashLight Name=MuzzleFlashLightComponent0
		LightOffset=(X=60.0,Y=0.0,Z=0.0)
		TimeShift=((StartTime=0.0,Radius=200,Brightness=5,LightColor=(R=200,G=200,B=200,A=255)),(StartTime=0.21,Radius=200,Brightness=5,LightColor=(R=200,G=200,B=200,A=255)),(StartTime=0.3,Radius=64,Brightness=0,LightColor=(R=32,G=32,B=32,A=255)))
	End Object
	MuzzleFlashLight=MuzzleFlashLightComponent0

	Begin Object Class=UTAmbientSoundComponent name=AmbientSoundComponent0
		SoundCue=SoundCue'Sentinel_Resources.Sounds.Weapons.BLinkGunBeam1Cue'
	End Object
	BeamSoundComponent=AmbientSoundComponent0
	Components.Add(AmbientSoundComponent0)

	Begin Object Class=ParticleSystemComponent Name=ParticleSystemComponent0
		Template=ParticleSystem'Sentinel_Resources.Effects.Weapons.P_LinkCannon_Tubes'
	End Object
	TubesComponent=ParticleSystemComponent0
	Components.Add(ParticleSystemComponent0)
}
