//=============================================================================
// Fires multiple rockets, like the old PepperPot.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeapon_RocketLauncher extends SentinelWeapon;

/** Rockets fired in a circle of this radius. */
var() float FireRadius;
/** Total number of rockets to fire in each volley. */
var() int NumRockets;
/** Targets with CollisionHeight + CollisionRadius greater than this can be homed in on by rockets. */
var() float MinimumSizeForLock;

function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
{
	local Vector Aim;
	local Vector X, Y, Z;
	local Vector SpawnLocation, RealSpawnLocation;
	local Quat FireRotation;
	local Projectile Proj;
	local int i;
	local bool bFired;

	Aim = Normal(End - Start);
	GetAxes(BarrelDir, X, Y, Z);

	SpawnLocation = Start;
	SpawnLocation += X * FireInfo.FireOffset.X;
	SpawnLocation += Y * FireInfo.FireOffset.Y;
	SpawnLocation += Z * FireInfo.FireOffset.Z;

	for(i = 0; i < NumRockets; i++)
	{
		RealSpawnLocation = SpawnLocation;
		FireRotation = QuatFromAxisAndAngle(X, (2.0 * Pi / NumRockets) * i);
		RealSpawnLocation += QuatRotateVector(FireRotation, vect(0.0, 0.0, 1.0) * FireRadius);

		Proj = Cannon.Spawn(FireInfo.ProjectileClass, Cannon,, RealSpawnLocation, BarrelDir);

		if(Proj != none)
		{
			InitializeProjectile(Proj, Aim, End);
			bFired = true;
		}
	}

	super.FireAt(Start, BarrelDir, End);

	return bFired;
}

function InitializeProjectile(Projectile Proj, Vector ProjDir, optional Vector TargetLocation)
{
	local Actor TargetActor;
	local SentinelProjectile_Rocket RocketProj;
	local float TargetRadius, TargetHeight;
	local ESeekType SeekType;
	local EWobbleType WobbleType;

	Proj.Init(ProjDir);

	RocketProj = SentinelProjectile_Rocket(Proj);

	if(RocketProj != none)
	{
		TargetActor = Cannon.Target;

		if(TargetActor != none)
		{
			TargetActor.GetBoundingCylinder(TargetRadius, TargetHeight);

			if(TargetRadius + TargetHeight <= MinimumSizeForLock)
			{
				//Rocket's can't track small targets, so home in on their expected location and spread out to maximize the chance of hitting something.
				TargetActor = none;
				SeekType = SEEK_Location;
				WobbleType = WOBBLE_Diverge;
			}
			else if(TargetActor.IsStationary())
			{
				//Stationary targets don't need any special handling.
				SeekType = SEEK_Direct;
				WobbleType = WOBBLE_Converge;
			}
			else
			{
				//For other large targets, spread rockets out over entire target actor.
				SeekType = SEEK_WithinCollision;
				WobbleType = WOBBLE_Constant;
			}
		}
		else
		{
			//If there is no target, let rockets follow their own wobbly path to nowhere.
			SeekType = SEEK_None;
			WobbleType = WOBBLE_Constant;
		}

		RocketProj.SetTarget(TargetActor, TargetLocation, SeekType, WobbleType);
	}
}

defaultproperties
{
	FriendlyName="Rocket Launcher"
	Description="A derivative of Evil Corp's popular 'PepperPot' portable rocket launcher, this weapon attachment fires a cluster of small, fast rockets. Each equipped with its own processor and miniature transponder, the rockets are capable of interfacing with the Sentinel's own sensors while in flight, actively tracking larger targets and coordinating with each other to maximize effectiveness.\n\nThere is limited space inside a Sentinel to store rockets, but your subscription to the ADS-1 support service means that replacement rockets will be translocated in from the nearest Evil Corp. warehouse as needed, freeing you from the logistics of resupplying Sentinels in the field."

	FireRadius=12.0
	NumRockets=8
	MinimumSizeForLock=100.0

	WeaponMesh=SkeletalMesh'Sentinel_Resources.Meshes.Weapons.SK_RocketLauncher'
	PhysicsAsset=PhysicsAsset'Sentinel_Resources.Physics.SK_RocketLauncher_Physics'
	WeaponMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel'
	WeaponSpawnMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Spawn'
	WeaponDeadMaterial=MaterialInstance'Sentinel_Resources.Materials.Sentinel.MI_Sentinel_Dead'
	WeaponScale=0.6

	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Weapons.SwitchToRocketLauncherCue'

	ExtraStrength=0.3
	AmmoCost=40
	CompatibleSentinels.Remove(class'Sentinel_Ceiling')

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		FireInterval=1.0
		ProjectileClass=class'SentinelProjectile_Rocket'
		FireOffset=(X=40.0,Y=0.0,Z=0.0)
		FireSound=SoundCue'Sentinel_Resources.Sounds.Weapons.PepperPotFireCue'
		MaxAimError=22.0
	End Object
	FireInfo=FireInfo0

	Begin Object Class=SentinelWeaponComponent_MuzzleFlashLight Name=MuzzleFlashLightComponent0
		LightOffset=(X=40.0,Y=0.0,Z=0.0)
		TimeShift=((StartTime=0.0,Radius=200,Brightness=5,LightColor=(R=255,G=170,B=42,A=255)),(StartTime=1.0,Radius=64,Brightness=0,LightColor=(R=255,G=170,B=42,A=255)))
	End Object
	MuzzleFlashLight=MuzzleFlashLightComponent0
}
