//=============================================================================
// Not a weapon, instead it allows a Sentinel to act as a lookout.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class SentinelWeapon_SurveillanceModule extends SentinelWeapon;

var() const SceneCapture2DComponent CameraComponent;
/** Material to draw on the screen */
var() Material CameraMaterial;
var MaterialInstance CameraMaterialInstance;
var TextureRenderTarget2D TextureTarget;

/** Resolution to create TextureTarget. */
var() int TextureResolution;
/** Relative location on screen of top left corner of TextureTarget. */
var() float DrawLocationX, DrawLocationY;
/** Relative size on screen of TextureTarget. */
var() float DrawSize;

/** Desaturation to apply to camera material. 0.0 = normal colour, 1.0 = black and white. This is a maximum amount - the actual desaturation decreases to 0 when view is centred. */
var() float DesaturationAmout;
/** Adjustment to camera material gamma. 1.0 = normal brightness, <1.0 = brighter, >1.0 = darker. */
var() float GammaAdjustment;
/** Amount of noise to add to camera material. 0.0 = no noise >0.0 = more noise. This is a maximum amount - the actual noise decreases to 0 when view is centred. */
var() float NoiseAmount;

/** Interval between searches for Deployers to render for. */
var() float DeployerSearchPeriod;

var repnotify bool bCameraActivated;

var() Rotator TestRotator;

replication
{
	if(Role == ROLE_Authority && bNetDirty)
		bCameraActivated;
}

simulated event ReplicatedEvent(name VarName)
{
	if(VarName == 'bCameraActivated')
	{
		if(bCameraActivated)
		{
			ActivateCamera();
		}
		else
		{
			DeactivateCamera();
		}
	}
}

simulated function ClientInitializeFor()
{
	Cannon.TargetingSound = none;
	Cannon.WaitingSound = none;

	super.ClientInitializeFor();

	CheckCamera();
}

/**
 * Activates the camera if it should be on (i.e. a Telemeter is present), or deactivates it if it should not be on.
 */
simulated function CheckCamera()
{
	if(Cannon.UpgradeManager != none && Cannon.UpgradeManager.HasUpgrade(class'SentinelUpgrade_Telemeter', true))
	{
		ActivateCamera();
	}
	else
	{
		DeactivateCamera();
	}
}

simulated function ActivateCamera()
{
	FindDeployersToRenderFor();
	SetTimer(DeployerSearchPeriod, true, 'FindDeployersToRenderFor');

	if(TextureTarget == none)
	{
		TextureTarget = class'TextureRenderTarget2D'.static.Create(TextureResolution, TextureResolution);

		if(CameraMaterialInstance == none)
		{
			CameraMaterialInstance = new(self) class'MaterialInstanceConstant';
			CameraMaterialInstance.SetParent(CameraMaterial);
		}
	
		CameraMaterialInstance.SetTextureParameterValue('CameraTexture', TextureTarget);
		CameraComponent.SetCaptureParameters(TextureTarget);
		Cannon.WeaponComponent.AttachComponentToSocket(CameraComponent, Cannon.SWeapon.CameraSocketName);
	}

	bRenderOverlays = true;
	bCameraActivated = true;
}

simulated function DeactivateCamera()
{
	bCameraActivated = false;
	bRenderOverlays = false;

	if(Cannon != none)
	{
		Cannon.WeaponComponent.DetachComponent(CameraComponent); //n.b. not detaching the component when Sentinel destroyed results in a GPF.
	}
}

function AddUpgrade(SentinelUpgrade NewUpgrade)
{
	super.AddUpgrade(NewUpgrade);

	CheckCamera();
}

function SentinelUpgrade DetachUpgrade(SentinelUpgrade UpgradeToDetach)
{
	local SentinelUpgrade DetachedUpgrade;

	DetachedUpgrade = super.DetachUpgrade(UpgradeToDetach);

	if(DetachedUpgrade == self)
	{
		DeactivateCamera();
	}
	else
	{
		CheckCamera();
	}

	return DetachedUpgrade;
}

function bool FireAt(Vector Start, Rotator BarrelDir, Vector End)
{
	return false; //false so that NotifyFired is not called, and thus the Cloak upgrade does not decloak on "firing".
}

simulated function RenderOverlays(HUD H)
{
	local float Angle;
	local float X, Y; //Canvas location in pixels of upper left corner of texture.
	local float XL, YL; //Dimensions in pixels of texture on canvas.
	local Rotator CameraRotation; //Rotation needed for the camera image to appear upright.

	if(ShouldRenderFor(H.PlayerOwner, Angle))
	{
		Cannon.WeaponComponent.AttachComponentToSocket(CameraComponent, Cannon.SWeapon.CameraSocketName);
		CameraMaterialInstance.SetScalarParameterValue('Gamma', GammaAdjustment);
		CameraMaterialInstance.SetScalarParameterValue('Noise', NoiseAmount * (Angle / 360.0));
		CameraMaterialInstance.SetScalarParameterValue('Desaturation', DesaturationAmout * (Angle / 360.0));

		X = DrawLocationX * H.Canvas.ClipX;
		Y = DrawLocationY * H.Canvas.ClipY;
		XL = FMin(H.Canvas.ClipX, H.Canvas.ClipY) * DrawSize;
		YL = XL;
		CameraRotation.Yaw = Cannon.WeaponComponent.Rotation.Roll;

		H.Canvas.SetPos(X, Y);
		H.Canvas.DrawRotatedMaterialTile(CameraMaterialInstance, CameraRotation, XL, YL, 0.0, 0.0, 1.0, 1.0);
	}
}

/**
 * Determines if this Sentinel should have its camera drawn on the player's HUD.
 *
 * @param	PC		the PlayerController that wants a camera view
 * @param	Angle	set to the angle in degrees between the player's view and the Sentinel
 * @return	true if this Sentinel is close enough to the centre of the player's view and is closer than any other camera-equiped Sentinels.
 */
simulated function bool ShouldRenderFor(PlayerController PC, out float Angle)
{
	local Vector ViewLocation;
	local Rotator ViewRotation;
	local Sentinel S;
	local bool bShouldRender;

	PC.GetActorEyesViewPoint(ViewLocation, ViewRotation);
	Angle = RDiff(ViewRotation, Rotator(Cannon.Location - ViewLocation));

	bShouldRender = true;

	foreach WorldInfo.AllPawns(class'Sentinel', S)
	{
		if(Cannon.IsSameTeam(S) && S.UpgradeManager != none && S.UpgradeManager.HasUpgrade(Class, true))
		{
			if(RDiff(ViewRotation, Rotator(S.Location - ViewLocation)) < Angle)
			{
				bShouldRender = false;
				break;
			}
		}
	}

	return bShouldRender;
}

simulated function Destroyed()
{
	if(Cannon != none)
	{
		Cannon.TargetingSound = Cannon.default.TargetingSound;
		Cannon.WaitingSound = Cannon.default.WaitingSound;
	}

	DeactivateCamera();
	CameraMaterialInstance = none;

	super.Destroyed();
}

defaultproperties
{
	FriendlyName="Surveillance Module"
	Description="The Surveillance Module is fitted in place of a Sentinel's weapon, but it is not a weapon itself. It is intended for use in combination with the Mesh Network Interface, so that a suitably positioned Sentinel can act as a passive lookout, coordinating other Sentinels without drawing attention to itself.\n\nIf a telemeter is present, then the module's built-in wide-angle camera can transmit a live video feed to your Sentinel Deployer. Because of the high bandwidth nature of the transmission, then Deployer must be pointed directly at the Sentinel for optimum reception. If the Deployer detects multiple signals, then it will automatically lock on to the strongest one."

	CameraMaterial=Material'Sentinel_Resources.Materials.Misc.CameraMaterial'
	TextureResolution=256
	DrawLocationX=0.0
	DrawLocationY=0.05
	DrawSize=0.33
	DesaturationAmout=0.9
	GammaAdjustment=0.4
	NoiseAmount=1.0

	DeployerSearchPeriod=2.1

	WeaponMesh=SkeletalMesh'Sentinel_Resources.Meshes.Weapons.SK_SurveillanceModule'
	AnimTreeTemplate=AnimTree'Sentinel_Resources.Meshes.Weapons.AT_SurveillanceModule'
	WeaponMaterial=MaterialInstance'Sentinel_Resources.Materials.Weapons.MI_SurveillanceModule'
	WeaponSpawnMaterial=MaterialInstance'Sentinel_Resources.Materials.Weapons.MI_SurveillanceModule_Spawn'
	WeaponDeadMaterial=MaterialInstance'Sentinel_Resources.Materials.Weapons.MI_SurveillanceModule_Dead'
	WeaponScale=0.5
	RootBone=b_root

	ActivateSound=SoundCue'Sentinel_Resources.Sounds.Weapons.SwitchToLinkGunCue'

	ExtraStrength=-1.0 //No weapon = useless in combat.
	AmmoCost=25
	CompatibleSentinels.Remove(class'Sentinel_Floor')
	RequiredUpgrades.Add(class'SentinelUpgrade_Network')

	bAlwaysRelevant=true

	Begin Object Class=SentinelWeaponComponent_FireInfo Name=FireInfo0
		FireInterval=1.0
		FireLoudness=0.0
		MaxRange=10000.0
	End Object
	FireInfo=FireInfo0

	Begin Object Class=SceneCapture2DComponent Name=SceneCapture2DComponent0
		FrameRate=10.0
		NearPlane=10.0
		FarPlane=0.0
		FieldOfView=120
		ViewMode=SceneCapView_Lit
		bEnableFog=true
	End Object
	CameraComponent=SceneCapture2DComponent0
	Components.Add(SceneCapture2DComponent0)
}
