//=============================================================================
// Replicates configurable things from server to client.
// TODO: This will fail if there are multiple local players (split screen).
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class UTLRI_Sentinel extends UTLinkedReplicationInfo;

/** All sizes of ammo available. Used by the Deployer when throwing ammo. */
var() array< class<UTAmmo_SentinelDeployer> > AmmoClasses;

/** Object that provides access to values of configurable settings. */
var SentinelConfigurationProvider ConfigProvider;
var bool bAllowUpgrades;
var int MaxCannon;
var int MaxTeamCannon;
var int MaxTeamAdvantageCannon;

/** Set when upgrade lists have been emptied in preparation for replicated upgrade classes. */
var bool bUpgradeListsCleared;

replication
{
	if(Role == ROLE_Authority && bNetInitial)
		bAllowUpgrades, MaxCannon, MaxTeamCannon, MaxTeamAdvantageCannon;
}

function PostBeginPlay()
{
	local PlayerController PlayerOwner;

	super.PostBeginPlay();

	FindConfigProvider();

	if(ConfigProvider != none)
	{
		bAllowUpgrades = ConfigProvider.GetAllowUpgrades();
		MaxCannon = ConfigProvider.GetMaxCannon();
		MaxTeamCannon = ConfigProvider.GetMaxTeamCannon();
		MaxTeamAdvantageCannon = ConfigProvider.GetMaxTeamAdvantageCannon();
	}
	else
	{
		`warn("No ConfigProvider!");
	}

	PlayerOwner = PlayerController(Owner);

	//Replicate upgrade classes only to remote players.
	if(PlayerOwner != none && LocalPlayer(PlayerOwner.Player) == none)
	{
		ReplicateUpgradeClasses();
	}

	SetTimer(9.99, true);
}

/**
 * Looks for all SentinelConfigurationProviders and selects the highest priority one to assign to ConfigProvider.
 */
function FindConfigProvider()
{
	local Actor A;
	local SentinelConfigurationProvider FoundConfigProvider;

	foreach DynamicActors(class'Actor', A)
	{
		FoundConfigProvider = SentinelConfigurationProvider(A);

		if(FoundConfigProvider != none)
		{
			//`log("ConfigProvider:"@FoundConfigProvider@FoundConfigProvider.GetProviderPriority());

			if(ConfigProvider == none || FoundConfigProvider.GetProviderPriority() > ConfigProvider.GetProviderPriority())
			{
				ConfigProvider = FoundConfigProvider;
			}
		}
	}
}

function Timer()
{
	//Make sure the LRI doesn't hang around too long after the owner has gone.
	if(Owner == none)
	{
		Destroy();
	}
}

/**
 * Replicate the list of allowed upgrades to the client.
 */
function ReplicateUpgradeClasses()
{
	local class<UpgradeList> List;
	local String ClassName;

	foreach class'UTUIDataStore_SentinelUpgrades'.default.Upgrades(List)
	{
		foreach List.default.ClassNames(ClassName)
		{
			ReplicateUpgradeClass(ClassName, List.default.Tag);
		}
	}
}

/**
 * Add an upgrade class to the list of allowed upgrades.
 */
reliable client function ReplicateUpgradeClass(String NewUpgradeClassName, name ListTag)
{
	local class<UpgradeList> List;

	if(!bUpgradeListsCleared)
	{
		ClearUpgradeClasses();
		bUpgradeListsCleared = true;
	}

	foreach class'UTUIDataStore_SentinelUpgrades'.default.Upgrades(List)
	{
		if(List.default.Tag == ListTag)
		{
			List.default.ClassNames.AddItem(NewUpgradeClassName);
			break;
		}
	}
}

/**
 * Remove all allowed upgrade classes so that only the ones replicated later are valid.
 */
simulated function ClearUpgradeClasses()
{
	local class<UpgradeList> List;

	foreach class'UTUIDataStore_SentinelUpgrades'.default.Upgrades(List)
	{
		List.default.ClassNames.Remove(0, List.default.ClassNames.Length);
	}
}

defaultproperties
{
	AmmoClasses(0)=class'UTAmmo_SentinelDeployer'
	AmmoClasses(1)=class'UTAmmo_SentinelDeployer_Medium'
	AmmoClasses(2)=class'UTAmmo_SentinelDeployer_Small'
}