//=============================================================================
// Gives all players the Sentinel Deployer and uses various methods to add ammo
// for it to the game.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class UTMutator_Sentinel extends UTMutator_SentinelBaseMutator
	config(Mutator_Sentinel)
	implements(SentinelConfigurationProvider);

/**
 * A pickup to replace and how.
 */
struct ReplacementInfo
{
	/** Pickup to be replaced. */
	var class<PickupFactory> PickupClass;
	/** Size of ammo to replace it with. */
	var class<UTAmmo_SentinelDeployer> ReplacementClass;
	/** Percentage of pickups replaced. */
	var int ReplacementPercent;

	structdefaultproperties
	{
		ReplacementClass=class'UTAmmo_SentinelDeployer'
		ReplacementPercent=100
	}
};

/**
 * A NavigationPoint for spawning ammo at.
 */
struct AmmoPointInfo
{
	/** The point itself. */
	var NavigationPoint Point;
	/** A pickup that has been spawned here. */
	var UTDropped_SentinelAmmo Pickup;
};

/** Set true to give every player a Deployer when they spawn. */
var config bool bStartWithDeployer;
/** Deployer is not allowed to upgrade if this is false. */
var config bool bAllowUpgrades;
/** Set true to use settings from the mutator in preference to those set by the mapper in maps which have a Deployer placed. */
var config bool bOverrideMapSettings;

/** Maximum number of Sentinels each player can have deployed at once. */
var config int MaxCannon;
/** Maximum number of Sentinels for all players in a team */
var config int MaxTeamCannon;
/** Biggest difference allowed in number of Sentinels deployed for each team. */
var config int MaxTeamAdvantageCannon;

/** Pickups to be replaced. */
var config array<ReplacementInfo> ReplacedPickups;
/** If true, ammo crates will be spawned at random NavigationPoints. */
var config bool bSpawnRandomAmmo;
/** Random ammo may be added until the total ammo (replacement + random) is up to this amount. */
var config int InitialMaxPickupAmmo;
/** MaxAmmoQuantity is increased by this much for every player in the game. */
var config int ExtraMaxPickupAmmoPerPlayer;
/** Ammo will not be randomly spawned at NavigationPoints of these classes. */
var array< class<NavigationPoint> > NoAmmoPointClasses;
/** All NavigationPoints found to be suitable for spawning ammo at. */
var array<AmmoPointInfo> AmmoPoints;

/** Class of GameRules to add. */
var class<GameRules> GRClass;

/** Set to true if a SentinelConfigurationProvider was found in the map. */
var bool bMapHasOwnSettings;
/** Quantity of ammo spawned so far. */
var int CurrentAmmoQuantity;

function InitMutator(string Options, out string ErrorMessage)
{
	WorldInfo.Game.AddGameRules(GRClass);
	bMapHasOwnSettings = MapHasConfigActor();

	if(bSpawnRandomAmmo)
	{
		//This is done on a timer because AllNavigationPoints doesn't work immediately.
		SetTimer(1.0, false, 'InitializeRandomAmmo');
	}

	if(bStartWithDeployer)
	{
		AddToDefaultInventory(UTGame(WorldInfo.Game), DeployerClass);
	}

	super.InitMutator(Options, ErrorMessage);
}

static final function AddToDefaultInventory(UTGame UTG, class<Weapon> NewClass)
{
	local int i;
	local bool bDeployerAlreadyAdded;

	if(UTG != none)
	{
		for(i = 0; i < UTG.DefaultInventory.Length; i++)
		{
			if(ClassIsChildOf(UTG.DefaultInventory[i], NewClass))
			{
				bDeployerAlreadyAdded = true;
				break;
			}
		}

		if(!bDeployerAlreadyAdded)
		{
			UTG.DefaultInventory.AddItem(NewClass);
		}
	}
	else
	{
		`warn("Failed to find a UTGame.");
	}
}

function bool MapHasConfigActor()
{
	local Actor A;
	local bool bConfigActorFound;

	foreach DynamicActors(class'Actor', A)
	{
		if(SentinelConfigurationProvider(A) != none && A != self)
		{
			bConfigActorFound = true;
			break;
		}
	}

	return bConfigActorFound;
}

/**
 * Calculates current maximum ammo quantity.
 *
 * @return InitialMaxPickupAmmo + (total players * ExtraMaxPickupAmmoPerPlayer)
 */
function int GetMaxAmmoQuantity()
{
	local int MaxAmmoQuantity;

	MaxAmmoQuantity = InitialMaxPickupAmmo;
	MaxAmmoQuantity += WorldInfo.Game.NumPlayers * ExtraMaxPickupAmmoPerPlayer;
	MaxAmmoQuantity += WorldInfo.Game.NumBots * ExtraMaxPickupAmmoPerPlayer;

	return MaxAmmoQuantity;
}

/**
 * Spawns initial allocation of random ammo.
 */
function InitializeRandomAmmo()
{
	FindRandomAmmoPoints(AmmoPoints);

	`log("Started spawning random ammo from intial quantity:"@CurrentAmmoQuantity);
	SpawnRandomAmmo();
	`log("Finished spawning random ammo with final quantity:"@CurrentAmmoQuantity);
}

/**
 * Delegate for random ammo to notify the mutator that it has been destroyed.
 */
function RandomAmmoDestroyed(UTDropped_SentinelAmmo_Random DestroyedPickup)
{
	CurrentAmmoQuantity -= DestroyedPickup.AmmoClass.default.AmmoAmount;

	SpawnRandomAmmo();
}

/**
 * Spawns ammo pickups at random NavigationPoints until MaxAmmoQuantity is reached.
 */
function SpawnRandomAmmo()
{
	local int MaxAmmoQuantity;
	local array<int> EmptyAmmoPointIndexes, OccupiedAmmoPointIndexes;
	local int i;
	local UTDropped_SentinelAmmo_Random AmmoPickup;

	SortAmmoPoints(AmmoPoints, EmptyAmmoPointIndexes, OccupiedAmmoPointIndexes);

	MaxAmmoQuantity = GetMaxAmmoQuantity();

	while(EmptyAmmoPointIndexes.length > 0)
	{
		if(MaxAmmoQuantity != 0 && (CurrentAmmoQuantity + class'UTAmmo_SentinelDeployer_Small'.default.AmmoAmount) > MaxAmmoQuantity)
		{
			break;
		}

		i = Rand(EmptyAmmoPointIndexes.length - 1);

		AmmoPickup = AmmoPoints[EmptyAmmoPointIndexes[i]].Point.Spawn(class'UTDropped_SentinelAmmo_Random',,, AmmoPoints[EmptyAmmoPointIndexes[i]].Point.Location);

		if(AmmoPickup != none)
		{
			AmmoPickup.InitializeFromAmmoClass(class'UTAmmo_SentinelDeployer_Small');
			AmmoPickup.PickupDestroyed = RandomAmmoDestroyed;
			AmmoPickup.SetPhysics(PHYS_Falling);
			AmmoPoints[EmptyAmmoPointIndexes[i]].Pickup = AmmoPickup;
			CurrentAmmoQuantity += class'UTAmmo_SentinelDeployer_Small'.default.AmmoAmount;
		}
		else
		{
			`warn("Failed to spawn a random ammo pickup at:"@AmmoPoints[EmptyAmmoPointIndexes[i]].Point);
		}

		EmptyAmmoPointIndexes.Remove(i, 1);
	}
}

/**
 * Finds all NavigationPoints that could potentially have ammo crates spawned on them.
 */
function FindRandomAmmoPoints(out array<AmmoPointInfo> SuitableRandomAmmoPoints)
{
	local NavigationPoint N;
	local class<NavigationPoint> NoAmmoPointClass;
	local AmmoPointInfo AmmoPoint;

	foreach WorldInfo.AllNavigationPoints(class'NavigationPoint', N)
	{
		foreach NoAmmoPointClasses(NoAmmoPointClass)
		{
			if(ClassIsChildOf(N.Class, NoAmmoPointClass))
			{
				N = none;
				break;
			}
		}

		if(N != none)
		{
			//Only use if the distance to the ground is no more than 80 units.
			if(!FastTrace(N.Location - vect(0.0, 0.0, 80.0), N.Location))
			{
				AmmoPoint.Point = N;
				SuitableRandomAmmoPoints.AddItem(AmmoPoint);
			}
		}
	}
}

/**
 * Sorts AllAmmoPoints into those with a pickup and those without.
 *
 * @param AllAmmoPoints				ammo points to sort
 * @param EmptyAmmoPointIndexes		indexes into AllAmmoPoints of points without a pickup
 * @param OccupiedAmmoPointIndexes	indexes into AllAmmoPoints of points with a pickup
 */
function SortAmmoPoints(array<AmmoPointInfo> AllAmmoPoints, out array<int> EmptyAmmoPointIndexes, out array<int> OccupiedAmmoPointIndexes)
{
	local int i;

	for(i = 0; i < AllAmmoPoints.length; i++)
	{
		if(AllAmmoPoints[i].Pickup == none || AllAmmoPoints[i].Pickup.bDeleteMe)
		{
			EmptyAmmoPointIndexes[EmptyAmmoPointIndexes.length] = i;
		}
		else
		{
			OccupiedAmmoPointIndexes[OccupiedAmmoPointIndexes.length] = i;
		}
	}
}

function bool CheckReplacement(Actor Other)
{
	local bool bReplace;
	local PickupFactory PickupFactory;
	local UTAmmoPickupFactory AmmoPickupFactory;
	local ReplacementInfo ReplacedPickup;
	local Vector PickupLocation;

	//Don't replace anything if map already has ammo.
	if(bMapHasOwnSettings)
	{
		if(UTAmmo_SentinelDeployer(Other) != none)
		{
			//Pre-placed ammo pickups should be counted towards the total.
			CurrentAmmoQuantity += UTAmmoPickupFactory(Other).AmmoAmount;
		}

		return true;
	}

	bReplace = true;
	PickupFactory = PickupFactory(Other);

	if(PickupFactory != none)
	{
		foreach ReplacedPickups(ReplacedPickup)
		{
			if(Other.Class == ReplacedPickup.PickupClass)
			{
				if(Rand(100) > ReplacedPickup.ReplacementPercent)
				{
					break;
				}

				PickupLocation.Z = ReplacedPickup.ReplacementClass.default.CylinderComponent.CollisionHeight - PickupFactory.CylinderComponent.CollisionHeight;
				PickupLocation = PickupLocation >> PickupFactory.Rotation;
				PickupLocation += PickupFactory.Location;

				AmmoPickupFactory = PickupFactory.Spawn(ReplacedPickup.ReplacementClass,,, PickupLocation, PickupFactory.Rotation);

				if(AmmoPickupFactory != none)
				{
					AmmoPickupFactory.OriginalFactory = PickupFactory;
					PickupFactory.ReplacementFactory = AmmoPickupFactory;
					CurrentAmmoQuantity += ReplacedPickup.ReplacementClass.default.AmmoAmount;

					bReplace = false;
					//`log("Replaced"@Other@"with"@ReplacedPickup.ReplacementClass);
				}
				else
				{
					`warn("Failed to replace"@Other@"with"@ReplacedPickup.ReplacementClass);
				}

				break;
			}
		}
	}

	return bReplace;
}

function bool GetAllowUpgrades()
{
	return bAllowUpgrades;
}

function int GetMaxCannon()
{
	return MaxCannon;
}

function int GetMaxTeamCannon()
{
	return MaxTeamCannon;
}

function int GetMaxTeamAdvantageCannon()
{
	return MaxTeamAdvantageCannon;
}

function float GetProviderPriority()
{
	return bOverrideMapSettings ? 2.0 : 0.0;
}

defaultproperties
{
	NoAmmoPointClasses[0]=class'Objective'
	NoAmmoPointClasses[1]=class'PickupFactory'
	NoAmmoPointClasses[2]=class'PlayerStart'
	NoAmmoPointClasses[3]=class'Teleporter'
	NoAmmoPointClasses[4]=class'UTJumpPad'
	NoAmmoPointClasses[5]=class'UTOnslaughtNodeTeleporter'
	NoAmmoPointClasses[6]=class'UTVehicleFactory'

	GRClass=class'UTGameRules_VehicleKillAmmo'
}