//=============================================================================
// A convenient superclass for any mutator which adds upgrades/weapons to the
// list of those available. Simply add the desired upgrades/weapons to the
// default properties and they will be added without the need to edit
// UTMutator_Sentinel.ini
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class UTMutator_SentinelUpgrades extends UTMutator
	abstract;

/** Upgrades in this array will be added to the existing upgrade list. */
var() array<string> UpgradeClasses;
var() array<string> WeaponClasses;

/** If a class is set here, upgrades it lists will completely replace the original list of upgrades. This is currently the only way to remove allowed upgrades without editing the ini. */
var() class<UpgradeList> ReplacementUpgradesList;
var() class<UpgradeList> ReplacementWeaponsList;

function PostBeginPlay()
{
	super.PostBeginPlay();

	//This should be done in PostBeginPlay to ensure that the lists are updated before the main mutator adds the PreCachingActor via its InitMutator.
	ChangeUpgradeLists();
	AddToUpgradeLists();
}

/**
 * Removes all standard weapons and upgrades, replacing them with only those in the new lists.
 */
function ChangeUpgradeLists()
{
	if(ReplacementUpgradesList != none)
	{
		class'UTUIDataStore_SentinelUpgrades'.static.ChangeUpgradeList(ReplacementUpgradesList, 'Upgrades');
	}

	if(ReplacementWeaponsList != none)
	{
		class'UTUIDataStore_SentinelUpgrades'.static.ChangeUpgradeList(ReplacementWeaponsList, 'Weapons');
	}

}

/**
 * Adds new upgrades to the existing list of upgrades.
 */
function AddToUpgradeLists()
{
	local string UpgradeClass;
	local class<UpgradeList> UpgradeListClass;

	UpgradeListClass = class'UTUIDataStore_SentinelUpgrades'.static.GetUpgradeList('Upgrades');

	foreach UpgradeClasses(UpgradeClass)
	{
		if(UpgradeListClass.default.ClassNames.Find(UpgradeClass) == INDEX_NONE)
		{
			UpgradeListClass.default.ClassNames.AddItem(UpgradeClass);
		}
	}

	UpgradeListClass = class'UTUIDataStore_SentinelUpgrades'.static.GetUpgradeList('Weapons');

	foreach WeaponClasses(UpgradeClass)
	{
		if(UpgradeListClass.default.ClassNames.Find(UpgradeClass) == INDEX_NONE)
		{
			UpgradeListClass.default.ClassNames.AddItem(UpgradeClass);
		}
	}
}

defaultproperties
{
}