//=============================================================================
// Thanks to sinx for working out this method of using UTUIDataStore_StringList
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class UTUIDataStore_SentinelUpgrades extends UTUIDataStore_StringList;

/** Array that mirrors StringData. */
var array< class<UpgradeList> > Upgrades;

function Registered(LocalPlayer PlayerOwner)
{
	`log("UTUIDataStore_SentinelUpgrades: Registered");

	super.Registered(PlayerOwner);
	
	FillList();
}

function bool NotifyGameSessionEnded()
{
	`log("UTUIDataStore_SentinelUpgrades: NotifyGameSessionEnded");

	//Returning true causes this datastore to be unregistered after each game, so modifications to the lists do not persist unless saved.
	return true;
}

/**
 * For each type of upgrade, loads its list of classes and fills the corresponding element of StringData with their FriendlyName.
 */
function FillList()
{
	local class<UpgradeList> UpgradeInfo;
	local string UpgradeClassName;
	local class<SentinelUpgrade> UpgradeClass;

	foreach Upgrades(UpgradeInfo)
	{
		foreach UpgradeInfo.default.ClassNames(UpgradeClassName)
		{
			UpgradeClass = class<SentinelUpgrade>(DynamicLoadObject(UpgradeClassName, class'Class'));

			if(UpgradeClass != none)
				AddStr(UpgradeInfo.default.Tag, UpgradeClass.default.FriendlyName,  true);
		}
	}
}

/**
 * Gets a list of one type of upgrade.
 *
 * @param	FieldName	type of upgrade to find, e.g. 'Weapons' or 'Upgrades'
 * @return	the list of upgrades of the specified type
 */
static function class<UpgradeList> GetUpgradeList(name FieldName)
{
	local class<UpgradeList> UpgradeInfo;

	foreach default.Upgrades(UpgradeInfo)
	{
		if(UpgradeInfo.default.Tag == FieldName)
			break;
		else
			UpgradeInfo = none;
	}

	return UpgradeInfo;
}

/**
 * Determines if the given upgrade class is in any of the UpgradeLists.
 *
 * @param	UpgradeClass	upgrade to look for
 * @return	true if upgrade class found, false otherwise
 */
static function bool IsInList(class<SentinelUpgrade> UpgradeClass)
{
	local class<UpgradeList> UpgradeInfo;
	local string UpgradeClassString, ExistingUpgradeClassString;
	local bool bResult;

	bResult = false;
	UpgradeClassString = PathName(UpgradeClass);

	foreach default.Upgrades(UpgradeInfo)
	{
		foreach UpgradeInfo.default.ClassNames(ExistingUpgradeClassString)
		{
			if(UpgradeClassString == ExistingUpgradeClassString)
			{
				bResult = true;
				break;
			}
		}

		if(bResult)
		{
			break;
		}
	}

	return bResult;
}

/**
 * DLOs all the upgrade classes.
 *
 * @param	FieldName	type of upgrade to load, e.g. 'Weapons' or 'Upgrades'
 * @return	array of all loaded classes
 */
static function array< class<SentinelUpgrade> > LoadUpgradeClasses(name FieldName)
{
	local class<UpgradeList> UpgradeInfo;
	local string UpgradeClassName;
	local class<SentinelUpgrade> UpgradeClass;
	local array< class<SentinelUpgrade> > UpgradeArray;

	UpgradeInfo = GetUpgradeList(FieldName);

	foreach UpgradeInfo.default.ClassNames(UpgradeClassName)
	{
		UpgradeClass = class<SentinelUpgrade>(DynamicLoadObject(UpgradeClassName, class'Class'));

		if(UpgradeClass != none)
			UpgradeArray.AddItem(UpgradeClass);
	}

	return UpgradeArray;
}

/**
 * Replaces the Upgrades class matching the given FieldName with the new class. Use this to swap out the entire set of allowed upgrades or weapons with a new set.
 * Note that this will only work if it is called *before* this data store is instantiated.
 */
static final function ChangeUpgradeList(class<UpgradeList> NewList, name FieldName)
{
	local int i;
	local UTUIDataStore_SentinelUpgrades DataStoreArchetype;

	for(i = 0; i < default.Upgrades.Length; i++)
	{
		if(default.Upgrades[i].default.Tag == FieldName)
		{
			DataStoreArchetype = UTUIDataStore_SentinelUpgrades(FindObject("UTMutator_Sentinel.Default__UTUIDataStore_SentinelUpgrades", Class'UTUIDataStore_SentinelUpgrades'));

			if(DataStoreArchetype != none)
			{
				DataStoreArchetype.Upgrades[i] = NewList;
			}

			break;
		}
	}
}

defaultproperties
{
	Tag=SentinelUpgrades

	Upgrades[0]=class'UpgradeList_Upgrades'
	Upgrades[1]=class'UpgradeList_Weapons'
}