//=============================================================================
// Configuration for the Multi-Mortar weapon.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class UTUIScene_MultiMortarMenu extends UTUIScene_UpgradeConfigurationMenu;

/** Weapon this menu is configuring. */
var transient SentinelWeapon_MultiMortar Mortar;

var transient UTUIDataStore_MortarShells MortarShellsDataStore;

var transient UIList MortarShellsList;
var transient UILabel MortarShellDescription;

var transient UTUISlider ElevationSlider;
var transient UILabel ElevationValueLabel;

var transient UTUIButtonBar ButtonBar;
var transient UTUIButtonBarButton ApplyButton;
var transient UTUIButtonBarButton CancelButton;

function SceneActivated(bool bInitialActivation)
{
	local int ButtonIndex;

	super.SceneActivated(bInitialActivation);

	if(bInitialActivation)
	{
		UILabel(FindChild('UILabel_Title', true)).SetValue(class'SentinelWeapon_MultiMortar'.default.FriendlyName);

		//Shell list
		UILabel(FindChild('UILabel_ShellTypes', true)).SetDataStoreBinding(class'UIStrings'.default.ShellTypes);
		MortarShellsList = UIList(FindChild('UIList_ShellTypes', true));
		MortarShellsList.OnValueChanged = OnShellTypeChanged;

		UILabel(FindChild('UILabel_DescriptionLabel', true)).SetDataStoreBinding(class'UIStrings'.default.Description);
		MortarShellDescription = UILabel(FindChild('UILabel_Description', true));

		//Slider
		UILabel(FindChild('UILabel_Elevation', true)).SetDataStoreBinding(class'UIStrings'.default.Elevation);
		ElevationSlider = UTUISlider(FindChild('UTUISlider_Elevation', true));
		ElevationSlider.OnValueChanged = OnElevationChanged;
		ElevationValueLabel = UILabel(FindChild('UILabel_ElevationValue', true));

		//Button bar
		ButtonBar = UTUIButtonBar(FindChild('UTUIButtonBar_0', true));
		ButtonBar.Clear();

		ButtonIndex = ButtonBar.AppendButton("<Strings:UTGameUI.ButtonCallouts.Apply>", OnButtonBar_Apply);
		ApplyButton = ButtonBar.Buttons[ButtonIndex];

		ButtonIndex = ButtonBar.AppendButton("<Strings:UTGameUI.ButtonCallouts.Cancel>", OnButtonBar_Cancel);
		CancelButton = ButtonBar.Buttons[ButtonIndex];
		CancelButton.SetDockParameters(UIFACE_Left, ButtonBar, UIFACE_Left, 0);
		CancelButton.SetDockParameters(UIFACE_Right, none, UIFACE_MAX, 0);

		InitializeListDataStore();
	}
}

static function SetRangeData(UISlider Slider, float MinValue, float MaxValue, float CurrentValue, float NudgeValue, bool bIntRange)
{
	Slider.SliderValue.MinValue = MinValue;
	Slider.SliderValue.MaxValue = MaxValue;
	Slider.SliderValue.CurrentValue = CurrentValue;
	Slider.SliderValue.NudgeValue = NudgeValue;
	Slider.SliderValue.bIntRange = bIntRange;

	Slider.NotifyValueChanged();
}

function InitializeListDataStore()
{
	local DataStoreClient DSClient;

	DSClient = class'UIInteraction'.static.GetDataStoreClient();

	if(DSClient != none)
	{
		if(MortarShellsDataStore == none)
		{
			MortarShellsDataStore = UTUIDataStore_MortarShells(FindDataStore(class'UTUIDataStore_MortarShells'.default.Tag));

			if(MortarShellsDataStore == none)
			{
				MortarShellsDataStore = DSClient.CreateDataStore(class'UTUIDataStore_MortarShells');

				if(MortarShellsDataStore != none)
				{
					DSClient.RegisterDataStore(MortarShellsDataStore);
				}
			}
		}

		if(MortarShellsDataStore == none)
		{
			`warn("Failed to find or create MortarShellsDataStore");
		}
	}
}

function InitializeFor(SentinelUpgrade NewUpgrade)
{
	local int CurrentShellIndex;

	super.InitializeFor(NewUpgrade);

	Mortar = SentinelWeapon_MultiMortar(NewUpgrade);

	if(Mortar != none)
	{
		if(MortarShellsDataStore != none)
		{
			MortarShellsDataStore.FillListFrom(Mortar.ShellTypes);
			MortarShellsList.RefreshSubscriberValue();

			CurrentShellIndex = MortarShellsDataStore.GetIndexFromFriendlyName(Mortar.ShellTypes, class<SentinelProjectile_Mortar>(Mortar.FireInfo.ProjectileClass).default.FriendlyName);
			MortarShellsList.SetIndex(CurrentShellIndex);
			OnShellTypeChanged(none, 0); //Ensure the description is updated.
		}

		SetRangeData(ElevationSlider, Mortar.MinElevation, Mortar.MaxElevation, Mortar.DesiredElevation, 0.01, false);
	}
}

function OnShellTypeChanged(UIObject Sender, int PlayerIndex)
{
	local string SelectedShellType;
	local string SelectedShellDescription;

	if(MortarShellsList != none && Mortar != none)
	{
		SelectedShellType = MortarShellsList.GetElementValue(MortarShellsList.GetCurrentItem());
		SelectedShellDescription = MortarShellsDataStore.GetDescriptionFromFriendlyName(Mortar.ShellTypes, SelectedShellType);

		MortarShellDescription.SetValue(SelectedShellDescription);
		//Scroll back up to the top when changing description contents.
		UIScrollFrame(MortarShellDescription.GetOwner()).SetClientRegionPosition(UIORIENT_Vertical, UIScrollFrame(MortarShellDescription.GetOwner()).GetPosition(UIFACE_Top, EVALPOS_PixelViewport));
	}
}

function OnElevationChanged(UIObject Sender, int PlayerIndex)
{
	ElevationValueLabel.SetValue(Round(ElevationSlider.GetValue() * RadToDeg)$"");
}

function bool OnButtonBar_Apply(UIScreenObject InButton, int InPlayerIndex)
{
	local int SelectedShellIndex;

	if(Mortar != none)
	{
		SelectedShellIndex = MortarShellsDataStore.GetIndexFromFriendlyName(Mortar.ShellTypes, MortarShellsList.GetElementValue(MortarShellsList.GetCurrentItem()));
		Mortar.SetShellType(SelectedShellIndex);

		Mortar.SetDesiredElevation(ElevationSlider.GetValue());
	}

	CloseScene(self);

	return true;
}

function bool OnButtonBar_Cancel(UIScreenObject InButton, int InPlayerIndex)
{
	CloseScene(self);

	return true;
}

defaultproperties
{
}