//=============================================================================
// Deploys and upgrades Sentinels.
// https://mrevil.asvachin.eu/unreal/
//=============================================================================
class UTWeap_SentinelDeployer extends UTWeapon
	dependson(SUtils)
	dependson(UTLocalMessage_DeployerAction);

var const SymbolStaticMeshComponent SymbolComponent;

/** Class of LinkedReplicationInfo to create. */
var() class<UTLRI_Sentinel> LRIClass;
/** LinkedReplicationInfo to read settings from. */
var UTLRI_Sentinel LRI;
/** Class used for sending Deployer messages to owning player. */
var() class<UTLocalMessage_DeployerAction> LocalMessageClass;

var() SentinelInfo FloorInfo;
var() SentinelInfo CeilingInfo;
/** If player is aiming higher than this pitch, a Ceiling Sentinel may be deployed. Lower, and it will be a Floor Sentinel. */
var() int TypeSwitchAngle;

/** Maximum distance upgrades can be performed from. */
var() float UpgradeRange;
/** Maximum distance from which a Sentinel can be repaired or sabotaged. */
var() float RepairRange;
/** Amount of health to restore to a Sentinel each "hit". */
var() float RepairAmount;
/** Amount of damage to inflict on a Sentinel each "hit". */
var() float SabotageAmount;
/** Proportion of sabotage damage returned as ammo. */
var() float SabotageAmmoGain;
/** Stores fractional ammo gain from sabotaging, to reduce rounding errors. */
var float AccumulatedSabotageAmmoGain;
/** Chance of an upgrade being destroyed when sabotaging. */
var() float DestroyUpgradeChance;
/** Sound to play when repairing. */
var() SoundCue RepairSound;
/** Sound to play when attemptign to deploy. */
var() SoundCue DeploySound;
/** Fraction of Sentinel value returned as ammo when recycling. */
var() float RecycleEfficiency;
/** Deployment will be prevented within this distance from enemy PlayerStarts. */
var() float MinimumDistanceFromEnemySpawn;

/** Menu to open for upgrading. */
var() UTUIScene_SentinelUpgradeMenu UpgradeMenuTemplate;
/** Currently open menu. */
var UTUIScene_SentinelUpgradeMenu UpgradeMenu;
/** Time between menu refreshes. */
var() float MenuUpdatePeriod;
/** Set when upgrade menu opened. */
var Sentinel SelectedSentinel;

/** Holds upgrades that want to render something to the HUD. They should handle adding themselves to this array. */
var array<SentinelUpgrade> UpgradesToRender;

/** Bots won't deploy Sentinels closer together than this. */
var() float BotSpacing;
/** Bots will look for Sentinels to repair within this radius. */
var() float BotRepairSearchRadius;
/** Bots will look for Sentinels to upgrade within this radius. */
var() float BotUpgradeSearchRadius;
/** Bots have this much range for all actions, relative to human players. */
var() float BotCheatFactor;
/** Owning bot. */
var UTBot Bot;
/** Arrays of Sentinels found within range when last checked. */
var array<Sentinel> OwnSentinels;
var array<Sentinel> AlliedSentinels;
var array<Sentinel> EnemySentinels;
/** Last time Sentinel arrays were refreshed. */
var float LastSentinelArrayFillTime;
/** Set to true when the bot is trying to use the Deployer for something, so it knows not to switch away. */
var bool bBotTryingToUseDeployer;

/** Set to true when player is holding fire, ready to deploy. */
var protected bool bDeploying;

replication
{
	if(Role == ROLE_Authority && bNetDirty && bNetOwner)
		LRI;
}

//Modified to only call Pickup.PickedUpBy if !bWeaponStay, ignoring bSuperWeapon.
//bSuperWeapon needs to be true otherwise the Berserk gives full ammo, but the weapon pickup should respect bWeaponStay.
function bool DenyPickupQuery(class<Inventory> ItemClass, Actor Pickup)
{
	local bool bResult;

	bSuperWeapon = false;
	bResult = super.DenyPickupQuery(ItemClass, Pickup);
	bSuperWeapon = true;

	return bResult;
}

function GivenTo(Pawn thisPawn, optional bool bDoNotActivate)
{
	super.GivenTo(thisPawn, bDoNotActivate);

	AddReplicationInfo();

	if(!Instigator.IsHumanControlled())
	{
		Bot = UTBot(Instigator.Controller);
		UpgradeRange *= BotCheatFactor;
		RepairRange *= BotCheatFactor;
		CeilingInfo.DeployRange *= BotCheatFactor;
	}

	//Set this here instead of in default properties so the pickup factory doesn't turn weapon stay off.
	bSuperWeapon = true;
}

/**
 * Adds a LinkedReplicationInfo to the players's PRI to hold global settings.
 */
function AddReplicationInfo()
{
	local UTPlayerReplicationInfo UTPRI;
	local UTLinkedReplicationInfo UTLRI;
	local bool bLRIAdded;

	UTPRI = UTPlayerReplicationInfo(Instigator.PlayerReplicationInfo);

	if(UTPRI != none)
	{
		UTLRI = UTPRI.CustomReplicationInfo;

		//Check to see if an LRI has already been added.
		while(UTLRI != none)
		{
			if(UTLRI.IsA('UTLRI_Sentinel'))
			{
				if(UTLRI.Owner != none) //This test catches a bug where the LRI is still present after the player reconnects, but it has no owner because the player's previous PlayerController no longer exists.
				{
					LRI = UTLRI_Sentinel(UTLRI);
					bLRIAdded = true;
				}

				break;
			}
			else
			{
				UTLRI = UTLRI.NextReplicationInfo;
			}
		}

		//Insert new LRI at head of list.
		if(!bLRIAdded)
		{
			LRI = UTPRI.Spawn(LRIClass, Instigator.Controller);
			LRI.NextReplicationInfo = UTPRI.CustomReplicationInfo;
			UTPRI.CustomReplicationInfo = LRI;
		}
	}
	else
	{
		`warn("Tried to add a LRI, but no PRI found!");
	}
}

/**
 * Allowed to deploy only if not too many Sentinels already deployed and enough ammo. If false is returned, a message is sent to the holding player to say what caused the failure.
 *
 * @param	ClassToDeploy		what is being considered for deployment. If none, then both types of Sentinel are considered
 * @return	true if nothing is preventing deployment, false otherwise
 */
function bool CanDeploy(class<Sentinel> ClassToDeploy)
{
	local bool bCanDeploy;
	local Sentinel S;
	local int CannonCount;
	local array<int> TeamCannonCount;
	local int i;

	//Don't let bots waste a Sentinel in the water.
	if(Bot != none && Instigator.PhysicsVolume.bWaterVolume)
	{
		return false;
	}

	TeamCannonCount.Length = WorldInfo.GRI.Teams.Length;

	foreach WorldInfo.AllPawns(class'Sentinel', S)
	{
		if(S.bCountsTowardLimit && S.Health > 0)
		{
			if(S.InstigatorController == Instigator.Controller)
			{
				CannonCount++;
			}

			if(S.PlayerReplicationInfo.Team != none)
			{
				TeamCannonCount[S.PlayerReplicationInfo.Team.TeamIndex]++;
			}
		}
	}

	bCanDeploy = true;

	if(LRI.MaxCannon > 0 && CannonCount >= LRI.MaxCannon)
	{
		Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_TooMany, Instigator.PlayerReplicationInfo);
		bCanDeploy = false;
	}
	else if(ClassToDeploy != none && !HasAmmo(0, ClassToDeploy.default.AmmoCost))
	{
		Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_NoAmmo, Instigator.PlayerReplicationInfo);
		bCanDeploy = false;
	}
	else if(!HasAmmo(0, FloorInfo.SentinelClass.default.AmmoCost) && !HasAmmo(0, CeilingInfo.SentinelClass.default.AmmoCost))
	{
		bCanDeploy = false;
	}
	else if(EnemyPlayerStartNearby())
	{
		Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_TooCloseToEnemy, Instigator.PlayerReplicationInfo);
		bCanDeploy = false;
	}
	else if(TeamCannonCount.Length > 0)
	{
		for(i = 0; i < TeamCannonCount.Length; i++)
		{
			if((LRI.MaxTeamCannon > 0) && (i == Instigator.PlayerReplicationInfo.Team.TeamIndex) && (TeamCannonCount[i] >= LRI.MaxTeamCannon))
			{
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_TooManyTeam, Instigator.PlayerReplicationInfo);
				bCanDeploy = false;
				break;
			}
			else if((LRI.MaxTeamAdvantageCannon > 0) && (TeamCannonCount[Instigator.PlayerReplicationInfo.Team.TeamIndex] - TeamCannonCount[i]) >= LRI.MaxTeamAdvantageCannon)
			{
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_TooManyTeamAdvantage, Instigator.PlayerReplicationInfo);
				bCanDeploy = false;
				break;
			}
		}
	}

	return bCanDeploy;
}

/**
 * Returns true if there is a UTTeamPlayerStart nearby that doesn't belong to the Instigator's team.
 */
function bool EnemyPlayerStartNearby()
{
	local UTGameObjective O;
	local PlayerStart P;
	local UTTeamPlayerStart UTP;
	local bool bFoundEnemyPlayerStart;

	//Find PlayerStarts associated with enemy objectives.
	foreach WorldInfo.AllNavigationPoints(class'UTGameObjective', O)
	{
		if(!O.IsNeutral() && O.GetTeamNum() != Instigator.GetTeamNum())
		{
			foreach O.PlayerStarts(P)
			{
				if(VSize(P.Location - Instigator.Location) < MinimumDistanceFromEnemySpawn)
				{
					bFoundEnemyPlayerStart = true;
					break;
				}
			}
		}
	}

	//Check for PlayerStarts directly set as enemy controlled.
	if(!bFoundEnemyPlayerStart)
	{
		foreach WorldInfo.AllNavigationPoints(class'UTTeamPlayerStart', UTP)
		{
			if(UTP.TeamNumber != 255 && UTP.TeamNumber != Instigator.GetTeamNum())
			{
				if(VSize(UTP.Location - Instigator.Location) < MinimumDistanceFromEnemySpawn)
				{
					bFoundEnemyPlayerStart = true;
					break;
				}
			}
		}
	}

	return bFoundEnemyPlayerStart;
}

/**
 * Decides which type of Sentinel to spawn and tries to spawn one.
 */
function Sentinel Deploy()
{
	local Rotator AimRotation;
	local Vector Aim;
	local Sentinel S;

	AimRotation = Instigator.GetBaseAimRotation();
	Aim = Vector(AimRotation);

	if(AimRotation.Pitch > TypeSwitchAngle)
	{
		if(CanDeploy(CeilingInfo.SentinelClass))
		{
			S = class'SUtils'.static.DeployCeiling(CeilingInfo, GetTraceOwner(), InstantFireStartTrace(), Aim);

			if(S == none)
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_Failed, Instigator.PlayerReplicationInfo);
		}
	}
	else
	{
		if(CanDeploy(FloorInfo.SentinelClass))
		{
			S = class'SUtils'.static.DeployFloor(FloorInfo, GetTraceOwner(), InstantFireStartTrace(), Aim);

			if(S == none)
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_Failed, Instigator.PlayerReplicationInfo);
		}
	}

	if(S != none)
	{
		S.InitializeFor(Instigator);
		Instigator.PlaySound(DeploySound);
		AddAmmo(-S.AmmoCost);
	}

	return S;
}

/**
 * A little help for bots to decide which Sentinel to spawn.
 */
function Sentinel BotDeploy()
{
	local Vector TraceStart, Aim;
	local Sentinel S;
	local int i;

	TraceStart = InstantFireStartTrace();
	Aim = Vector(Instigator.GetBaseAimRotation());

	//Try a Ceiling Sentinel first, since they are the least likely to be possible.
	//TODO: Probability should depend on how many Ceiling Sentinels this bot has already deployed.
	if(FRand() < 0.8 && CanDeploy(CeilingInfo.SentinelClass))
	{
		//Bots are unlikely to be looking up, so just pretend they are doing it right.
		S = class'SUtils'.static.DeployCeiling(CeilingInfo, GetTraceOwner(), TraceStart, vect(0.0, 0.0, 1.0));

		//If failed to spawn one directly above the bot, try another angle.
		if(S == none)
		{
			S = class'SUtils'.static.DeployCeiling(CeilingInfo, GetTraceOwner(), TraceStart, Normal(Normal(Aim * vect(1.0, 1.0, 0.0)) + vect(0.0, 0.0, 2.0)));
		}
	}

	//If that failed, try to spawn a Floor Sentinel.
	if(S == none && CanDeploy(FloorInfo.SentinelClass))
	{
		Aim.Z = 0;
		Aim = Normal(Aim);

		//Try in four different directions.
		for(i = 0; i < 4; i++)
		{
			if(!SentinelWouldFall(FloorInfo, TraceStart, Aim, BotSpacing))
			{
				S = class'SUtils'.static.DeployFloor(FloorInfo, GetTraceOwner(), TraceStart, Aim);

				if(S != none)
				{
					break;
				}
			}

			Aim = Aim >> Rot(0, 16384, 0);
		}
	}

	if(S != none)
	{
		S.InitializeFor(Instigator);
		Instigator.PlaySound(DeploySound);
		AddAmmo(-S.AmmoCost);
	}

	return S;
}

/**
 * Helper function for bot deployment.
 *
 * @param	SInfo			type of Sentinel to test with
 * @param	StartLocation	trace start used for Sentinel deployment
 * @parm	Aim				direction Sentinel would be spawned in
 * @param	MaxDistance		maximum fall distance allowed
 * @return	true if deploying here might result in the Sentinel falling further than MaxDistance, or if it might fall into water.
 */
function bool SentinelWouldFall(SentinelInfo SInfo, Vector StartLocation, Vector Aim, float MaxDistance)
{
	local Vector TraceStart, TraceEnd;
	local Vector HitLocation, HitNormal;
	local PhysicsVolume HitVolume;
	local bool bWouldFall;

	TraceStart = StartLocation;
	TraceEnd = TraceStart + (Aim * SInfo.DeployRange);

	if(Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true,,, TRACEFLAG_Bullet) == none)
	{
		HitLocation = TraceEnd;
	}

	TraceStart = HitLocation;
	TraceEnd = TraceStart - (vect(0.0, 0.0, 1.0) * MaxDistance);

	if(Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true,,, TRACEFLAG_Bullet) == none)
	{
		bWouldFall = true; //Sentinel might fall further than MaxDistance, allowing the bot to deploy another in the same spot, which would be undesirable.
	}
	else
	{
		HitVolume = PhysicsVolume(Trace(HitLocation, HitNormal, TraceEnd, TraceStart, true,,, TRACEFLAG_PhysicsVolumes));

		if(HitVolume != none && HitVolume.bWaterVolume)
		{
			bWouldFall = true; //Sentinel might fall directly into water and be destroyed, which would be a waste.
		}
	}

	return bWouldFall;
}

/**
 * Does repairs/sabotage.
 */
simulated function InstantFire()
{
	local Sentinel S;

	IncrementFlashCount();

	S = SentinelInRange(RepairRange, true, true);

	if(S != none)
	{
		DoRepair(S);
	}
	else if(Role == ROLE_Authority)
	{
		S = SentinelInRange(RepairRange, false, false, true);

		if(S != none)
		{
			DoSabotage(S);
		}
		else
		{
			Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_NoSentinel, Instigator.PlayerReplicationInfo);
		}
	}
}

/**
 * Repairs a Sentinel.
 */
simulated function DoRepair(Sentinel S)
{
	if(HasAmmo(CurrentFireMode, ShotCost[CurrentFireMode]))
	{
		if(S.Health >= S.HealthMax)
		{
			if(Instigator.IsLocallyControlled())
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_AlreadyRepaired, Instigator.PlayerReplicationInfo);
		}
		else if(Role < ROLE_Authority)
		{
			//Play sound for owning client (WeaponPlaySound is replicated to everyone else from the server).
			WeaponPlaySound(RepairSound);
		}
		else if(S.HealDamage(RepairAmount * Instigator.GetDamageScaling(), Instigator.Controller, InstantHitDamageTypes[CurrentFireMode]))
		{
			WeaponPlaySound(RepairSound);
			AddAmmo(-ShotCost[CurrentFireMode]);
		}
	}
}

/**
 * Damages a Sentinel, possibly destroying an upgrade at the same time.
 */
function DoSabotage(Sentinel S)
{
	local int OldHealth;

	OldHealth = S.Health;
	S.TakeDamage(SabotageAmount, Instigator.Controller, S.Location + Normal(Instigator.Location - S.Location), vect(0, 0, 0), InstantHitDamageTypes[CurrentFireMode]);
	AccumulatedSabotageAmmoGain += float(OldHealth - S.Health) * SabotageAmmoGain;
	AddAmmo(AccumulatedSabotageAmmoGain);
	AccumulatedSabotageAmmoGain -= int(AccumulatedSabotageAmmoGain);

	//Possibly destroy a random upgrade.
	if(S.Health > 0 && FRand() < DestroyUpgradeChance)
	{
		S.UpgradeManager.DestroyRandomUpgrade();
	}
}

/**
 * Opens the upgrade menu and associates it with the given Sentinel.
 *
 * @param	S		the Sentinel the menu should use
 */
simulated function OpenUpgradeMenuFor(Sentinel S)
{
	if(Instigator.Controller == none || !Instigator.IsLocallyControlled() || !S.IsSameTeam(Instigator))
	{
		return;
	}

	if(!CanUpgrade(S))
	{
		Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_NoUpgrades, Instigator.PlayerReplicationInfo);
		return;
	}

	SelectedSentinel = S;

	if(SelectedSentinel != none)
	{
		UpgradeMenu = UTUIScene_SentinelUpgradeMenu(UTPlayerController(Instigator.Controller).OpenUIScene(UpgradeMenuTemplate));

		if(UpgradeMenu != none)
		{
			UpgradeMenu.InitializeFor(self, SelectedSentinel);
			UpgradeMenu.OnSceneDeactivated = OnMenuDeactivated;
			SetTimer(MenuUpdatePeriod, true, 'MenuTimer');
		}
	}
}

/**
 * Stops the timer when the upgrade menu is closed.
 */
simulated function OnMenuDeactivated(UIScene DeactivatedScene)
{
	ClearTimer('MenuTimer');
}

/*
 * Checks that upgrade menu should still be open, refreshing it if it should, closing it otherwise.
 */
simulated function MenuTimer()
{
	if(UpgradeMenu == none || !UpgradeMenu.IsSceneActive() || SelectedSentinel == none || SelectedSentinel.Health <= 0 || VSize(SelectedSentinel.Location - Instigator.Location) > UpgradeRange || !FastTrace(SelectedSentinel.Location, InstantFireStartTrace(),, true))
	{
		if(UpgradeMenu != none)
		{
			UpgradeMenu.CloseScene(UpgradeMenu);
			Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_ConnectionLost, Instigator.PlayerReplicationInfo);
		}

		ClearTimer('MenuTimer');
	}
	else
	{
		UpgradeMenu.RefreshAll();
	}
}

simulated function bool CanUpgrade(Sentinel S)
{
	local bool bResult;

	bResult = true;

	if(!LRI.bAllowUpgrades || S == none || !S.bUpgradeable)
	{
		bResult = false;
	}
	else if(Instigator.Controller != S.InstigatorController && !S.bTeamUpgradeable)
	{
		bResult = false;
	}

	return bResult;
}

/**
 * Tries to upgrade SelectedSentinel with the given upgrade.
 *
 * @param	NewUpgradeClass		class of upgrade to attempt to appply
 */
reliable server function DoUpgrade(class<SentinelUpgrade> NewUpgradeClass)
{
	local SentinelUpgrade NewUpgrade;

	if(NewUpgradeClass == none || !class'UTUIDataStore_SentinelUpgrades'.static.IsInList(NewUpgradeClass) || !HasAmmo(0, NewUpgradeClass.default.AmmoCost))
		return;

	if(SelectedSentinel != none)
	{
		NewUpgrade = SelectedSentinel.UpgradeManager.CreateUpgrade(NewUpgradeClass);

		if(NewUpgrade != none)
		{
			AddAmmo(-NewUpgrade.AmmoCost);
		}
	}
}

/**
 * For bots, chooses an upgrade and applies it. Note that SelectedSentinel must be set correctly before calling this.
 */
function BotDoUpgrade()
{
	local class<SentinelUpgrade> UpgradeClass, ChosenUpgradeClass;
	local array< class<SentinelUpgrade> > PossibleUpgradeClasses;
	local float UpgradeDesirability, ChosenUpgradeDesirability;

	foreach SelectedSentinel.UpgradeManager.UpgradeClasses(UpgradeClass)
	{
		if(SelectedSentinel.UpgradeManager.AllowUpgrade(UpgradeClass) && HasAmmo(0, UpgradeClass.default.AmmoCost))
		{
			PossibleUpgradeClasses.AddItem(UpgradeClass);
		}
	}

	//Consider a new weapon only if the Sentinel still has the default.
	if(SelectedSentinel.SWeapon.Class == SelectedSentinel.DefaultWeaponClass)
	{
		foreach SelectedSentinel.UpgradeManager.WeaponClasses(UpgradeClass)
		{
			if(SelectedSentinel.UpgradeManager.AllowUpgrade(UpgradeClass) && HasAmmo(0, UpgradeClass.default.AmmoCost))
			{
				PossibleUpgradeClasses.AddItem(UpgradeClass);
			}
		}
	}

	foreach PossibleUpgradeClasses(UpgradeClass)
	{
		UpgradeDesirability = UpgradeClass.static.BotDesirability(Bot, SelectedSentinel);

		if(ChosenUpgradeClass == none || UpgradeDesirability > ChosenUpgradeDesirability)
		{
			ChosenUpgradeClass = UpgradeClass;
			ChosenUpgradeDesirability = UpgradeDesirability;
		}
	}

	if(ChosenUpgradeClass != none && ChosenUpgradeDesirability >= 0.0)
	{
		DoUpgrade(ChosenUpgradeClass);
	}
}

/**
 * Destroys a Sentinel and retrieves some ammo from it.
 *
 * @param	S	the Sentinel to recycle
 */
reliable server function DoRecycle(Sentinel S)
{
	if(S != none && SelectedSentinel == SentinelInRange(UpgradeRange))
	{
		AddAmmo(S.AmmoValue() * RecycleEfficiency);
		S.Suicide();
	}
}

simulated function ToggleLock(Sentinel S)
{
	if(S != none)
	{
		//For clients, change the value right away so the upgrade menu doesn't have to wait for it to be replicated.
		if(Role < ROLE_Authority)
		{
			S.bTeamUpgradeable = !S.bTeamUpgradeable;
		}

		ServerToggleLock(S);
	}
}

reliable server function ServerToggleLock(Sentinel S)
{
	if(S != none)
	{
		S.bTeamUpgradeable = !S.bTeamUpgradeable;
	}
}

/**
 * Returns the closest Sentinel belonging to the player if bOwn, or an allied Sentinel if bIncludeAllied, or an enemy Sentinel if bIncludeEnemy.
 * Only Sentinels directly in front of the player will count.
 *
 * @param	Range			distance to check to
 * @param	bIncludeOwn		if Sentinels owned by holding player may be returned
 * @param	bIncludeAllied	if Sentinels on the holding player's team may be returned
 * @param	bIncludeEnemy	if Sentinels hostile to the holding player may be returned
 * @return	the closest Sentinel in range of holding player that matches the specified criteria
 */
simulated function Sentinel SentinelInRange(float Range, optional bool bIncludeOwn=true, optional bool bIncludeAllied, optional bool bIncludeEnemy)
{
	local Vector HitLocation, HitNormal;
	local Vector X, Y, Z;
	local Vector Start, End;
	local Sentinel S;

	GetViewAxes(X, Y, Z);
	Start = InstantFireStartTrace();
	End = Start + X * Range;

	//Trace has an extent so that it will hit Sentinels' collision cylinders, which don't block zero-extent traces.
	S = Sentinel(GetTraceOwner().Trace(HitLocation, HitNormal, End, Start, true, vect(1.0, 1.0, 1.0),, TRACEFLAG_Bullet));

	if(S == none || S.Health <= 0 || VSize(Instigator.Location - S.Location) > Range)
	{
		S = none;
	}
	else if(S.InstigatorController == Instigator.Controller)
	{
		if(!bIncludeOwn)
		{
			S = none;
		}
	}
	else if(WorldInfo.GRI.OnSameTeam(S, Instigator))
	{
		if(!bIncludeAllied)
		{
			S = none;
		}
	}
	else if(!bIncludeEnemy)
	{
		S = none;
	}

	return S;
}

/**
 * Finds all Sentinels and sorts them by type into OwnSentinels, AlliedSentinels and EnemySentinels, with each array sorted by distance.
 * For bots only.
 */
function FindAllSentinels()
{
	local Sentinel S;

	//Don't bother rechecking too frequently.
	if(WorldInfo.TimeSeconds - LastSentinelArrayFillTime < 0.2)
		return;

	LastSentinelArrayFillTime = WorldInfo.TimeSeconds;

	OwnSentinels.Remove(0, OwnSentinels.Length);
	AlliedSentinels.Remove(0, AlliedSentinels.Length);
	EnemySentinels.Remove(0, EnemySentinels.Length);

	foreach WorldInfo.AllPawns(class'Sentinel', S, Instigator.Location)
	{
		if(S.Health > 0)
		{
			if(S.InstigatorController == Instigator.Controller)
			{
				AddSortedByDistance(OwnSentinels, S);
			}
			else if(WorldInfo.GRI.OnSameTeam(S, Instigator))
			{
				AddSortedByDistance(AlliedSentinels, S);
			}
			else
			{
				AddSortedByDistance(EnemySentinels, S);
			}
		}
	}
}

/**
 * Adds the Sentinel to the array sorted by distance, closest first.
 */
function AddSortedByDistance(out array<Sentinel> SentinelArray, Sentinel NewSentinel)
{
	local int i;
	local float NewDistance, OldDistance;

	if(SentinelArray.Length == 0)
	{
		SentinelArray.AddItem(NewSentinel);
	}
	else
	{
		for(i = 0; i < SentinelArray.Length; i++)
		{
			OldDistance = VSize(SentinelArray[i].Location - Instigator.Location);
			NewDistance = VSize(NewSentinel.Location - Instigator.Location);

			if(NewDistance < OldDistance)
			{
				SentinelArray.InsertItem(i, NewSentinel);
				break;
			}
		}
	}
}

/**
 * Looks to see if any Sentinel in the given array can be repaired.
 * @todo	Should be more likely to return more damaged Sentinels.
 * @todo	Should favour Sentinels that have a lot of kills (they are probably in a good place).
 */
function Sentinel SentinelToRepairFrom(array<Sentinel> Sentinels)
{
	local Sentinel S;

	foreach Sentinels(S)
	{
		if((VSize(S.Location - Instigator.Location) < BotRepairSearchRadius) && CanHeal(S))
			break;

		S = none;
	}

	return S;
}

/**
 * Finds the closest Sentinel that has affordable upgrades left to apply.
 */
function Sentinel SentinelToUpgradeFrom(array<Sentinel> Sentinels)
{
	local Sentinel S;
	local Sentinel BestSentinel;
	local float BestDist;

	if(!HasAmmo(0, 1))
		return none;

	foreach Sentinels(S)
	{
		if(CanUpgrade(S))
		{
			if((VSize(S.Location - Instigator.Location) < BotUpgradeSearchRadius) && !S.UpgradeManager.IsFullyUpgraded(AmmoCount, Bot))
			{
				if(BestSentinel == none || (BestDist > VSize(S.Location - Instigator.Location)))
				{
					BestSentinel = S;
					BestDist = VSize(S.Location - Instigator.Location);
				}
			}
		}
	}

	return BestSentinel;
}

//Don't play firing sound normally.
simulated function PlayFiringSound(){}

//Don't use ammo this way, consume it directly from the fire modes using AddAmmo instead.
function ConsumeAmmo(byte FireModeNum){}

//Always allow this weapon to be used.
simulated function bool HasAmmo(byte FireModeNum, optional int Amount)
{
	return AmmoCount >= Amount || UTInventoryManager(InvManager).bInfiniteAmmo;
}

simulated function bool HasAnyAmmo()
{
	return true;
}

//Always want ammo if not full.
simulated function float DesireAmmo(bool bDetour)
{
	return (AmmoCount < MaxAmmoCount) ? 1.0 : 0.0;
}

//Don't need to do anything when ammo runs out.
simulated function WeaponEmpty(){}

simulated function float MaxRange()
{
	return UpgradeRange;
}

/**
 * Throws ammo.
 */
exec simulated function ThrowAmmo()
{
	ServerThrowAmmo();
}

reliable server function ServerThrowAmmo()
{
	Instigator.ThrowActiveWeapon();
}

/**
 * Drop ammo instead of weapon.
 */
function DropFrom(Vector StartLocation, Vector StartVelocity)
{
	local UTDropped_SentinelAmmo P;
	local class<UTAmmo_SentinelDeployer> AnAmmoClass, DroppedAmmoClass;

	if(CanThrow())
	{
		//Find largest possible class of ammo that can be dropped.
		foreach LRI.default.AmmoClasses(AnAmmoClass)
		{
			if(AmmoCount >= AnAmmoClass.default.AmmoAmount)
			{
				if(DroppedAmmoClass == none || DroppedAmmoClass.default.AmmoAmount < AnAmmoClass.default.AmmoAmount)
				{
					DroppedAmmoClass = AnAmmoClass;
				}
			}
		}

	    if(DroppedAmmoClass != none)
		{
			P = Spawn(class'UTDropped_SentinelAmmo',,, StartLocation);

			if(P != none)
			{
				P.SetPhysics(PHYS_Falling);
				P.Velocity = StartVelocity;
				P.Instigator = Instigator;

				P.InitializeFromAmmoClass(DroppedAmmoClass);
				AddAmmo(-DroppedAmmoClass.default.AmmoAmount);
			}
		}
	}
}

/**
 * Any upgrades which have added themselves to UpgradesToRender will have a chance to do some rendering here.
 */
simulated function ActiveRenderOverlays(HUD H)
{
	local SentinelUpgrade Upgrade;

	super.ActiveRenderOverlays(H);

	UpgradesToRender.RemoveItem(none); //Make sure only existing upgrades are rendered!

	foreach UpgradesToRender(Upgrade)
	{
		if(Upgrade.bRenderOverlays && !Upgrade.bDeleteMe)
		{
			Upgrade.RenderOverlays(H);
		}
	}
}

/**
 * Tell bots to switch to this weapon under the right circumstances.
 * @todo	find appropriate values to return for each case
 */
function float GetAIRating()
{
	local int Result;

	Result = AIRating;

	if(Bot != none)
	{
		FindAllSentinels();

		if(Bot.Enemy == none || !Bot.bEnemyIsVisible)
		{
			//Definitely keep using the Deployer if the bot is doing something useful with it.
			if(bBotTryingToUseDeployer)
			{
				Result = 10.0;
			}
			//Maybe repair.
			else if(SentinelToRepairFrom(OwnSentinels) != none || SentinelToRepairFrom(AlliedSentinels) != none)
			{
				if(HasAmmo(1, 1))
					Result = 2.0;
			}
			//Maybe upgrade.
			else if(SentinelToUpgradeFrom(OwnSentinels) != none || SentinelToUpgradeFrom(AlliedSentinels) != none)
			{
				Result = 3.0;
			}
			//Maybe deploy.
			else if(CanDeploy(none))
			{
				//TODO: Make deployment conditions more specific.
				if(Bot.IsInState('Defending') || Bot.IsInState('StakeOut') || Bot.IsInState('Retreating') || Bot.IsInState('Hunting'))
				{
					Result = 5.0;
				}
			}
		}
		//If the bot thinks it is not likely to win, deploy, so as not to waste any collected ammo.
		else if(CanDeploy(none))
		{
			Result = 2.0 - (2.0 * Bot.RelativeStrength(Bot.Enemy));
		}
		//Maybe sabotage enemy Sentinel.
		else if(Sentinel(Bot.Enemy) != none && VSize(Bot.Enemy.Location - Bot.Pawn.Location) < BotSpacing && Bot.Enemy.Controller != none && Bot.Enemy.Controller.Enemy != Bot.Pawn)
		{
			Result = 2.0;
		}
	}

	return Result;
}

//Can attack enemy Sentinels by sabotaging, and repair damaged friendly Sentinels. May also attack by deploying a Sentinel.
function bool CanAttack(Actor Other)
{
	local Sentinel S;
	local bool bResult;

	bResult = true;
	S = Sentinel(Other);

	if(S != none)
	{
		if(S.IsSameTeam(Instigator))
		{
			if(!CanHeal(S))
			{
				bResult = false;
			}
		}
		else
		{
			bResult = super.CanAttack(Other);
		}
	}
	else
	{
		bResult = CanDeploy(none);
	}

	return bResult;
}

function bool CanHeal(Actor Other)
{
	local Sentinel S;
	local bool bResult;

	S = Sentinel(Other);

	//Can heal damaged, friendly Sentinels only.
	if(S != none && S.Health < S.HealthMax)
	{
		if(S.IsSameTeam(Instigator))
		{
			bResult = true;
		}
	}

	return bResult;
}

/**
 * UTBots call this when 'Pausing' in 'Defending' state, and then fire if true was returned.
 * This should happen when the bots are standing in a good defensive location.
 */
function bool ShouldFireWithoutTarget()
{
	local bool bResult;

	//Bots don't wait long enough for weapon to come up before trying to fire it, so force them to go back and wait again.
	if(Bot != none && IsInState('WeaponEquipping'))
	{
		Bot.GotoState('Defending','Pausing'); //zOMG H4X!
	}

	FindAllSentinels();
	//Fire only if it will deploy something.
	bResult = CanDeploy(none);
	//Don't try to deploy too close together.
	bResult = bResult && (OwnSentinels.Length == 0 || (VSize(OwnSentinels[0].Location - Instigator.Location) > BotSpacing));
	bResult = bResult && (AlliedSentinels.Length == 0 || (VSize(AlliedSentinels[0].Location - Instigator.Location) > BotSpacing));

	return bResult;
}

function byte BestMode()
{
	local byte Mode;

	if(Bot != none)
	{
		//Use secondary fire on any Sentinel.
		if(Sentinel(Bot.Focus) != none || Sentinel(Bot.Enemy) != none)
			Mode = 1;
	}

	return Mode;
}

//TODO: When should this be true, and when false?
/*function bool RecommendLongRangedAttack()
{
	return true;
}*/

simulated function Destroyed()
{
	if(UpgradeMenu != none)
		UpgradeMenu.CloseScene(UpgradeMenu);

	super.Destroyed();
}

//Fix for bots having their own Sentinel as an enemy in DM.
//TODO: Please get rid of this as soon as possible.
simulated function Tick(float DeltaTime)
{
	local int i;

	super.Tick(DeltaTime);

	if(Bot != none)
	{
		for(i = 0; i < ArrayCount(Bot.Squad.Enemies); i++)
		{
			if(Sentinel(Bot.Squad.Enemies[i]) != none && Sentinel(Bot.Squad.Enemies[i]).InstigatorController == Bot)
			{
				Bot.bIgnoreEnemyChange = true;
				Bot.Squad.RemoveEnemy(Bot.Squad.Enemies[i]);
				Bot.bIgnoreEnemyChange = false;
			}
		}
	}
}

//This is all for the bots' benefit. It periodically checks what they are doing, and if there is something
//Sentinel-related that might be appropriate at the time, tells them to do it.
//WARNING: Spaghetti code ahead!
simulated state Active
{
Begin:
	if(Bot != none)
	{
		Goto('Searching');
	}
	else
	{
		Goto('End'); //Do nothing for human players.
	}

Searching:
	bBotTryingToUseDeployer = false;
	Sleep(0.1);

	while((Bot.Enemy != none && Bot.bEnemyIsVisible) || Bot.PriorityObjective())
	{
		Sleep(3.0);
	}

	FindAllSentinels();

	//Look for one of the Bot's Sentinels that can be repaired.
	SelectedSentinel = SentinelToRepairFrom(OwnSentinels);

	if(SelectedSentinel == none)
	{
		//Next priority is repairing friendly Sentinels.
		SelectedSentinel = SentinelToRepairFrom(AlliedSentinels);
	}

	if(SelectedSentinel != none)
		Goto('Repairing');

	//Next priority is upgrading own Sentinel.
	SelectedSentinel = SentinelToUpgradeFrom(OwnSentinels);

	if(SelectedSentinel == none)
	{
		//Last priority is upgrading friendly Sentinels.
		SelectedSentinel = SentinelToUpgradeFrom(AlliedSentinels);
	}

	if(SelectedSentinel != none)
		Goto('Upgrading');

	Sleep(5.0);
	Goto('Searching'); //Loop forever.

Repairing:
	bBotTryingToUseDeployer = true;
	Bot.GoalString = "Going to repair"@SelectedSentinel;

	//The ScriptedMove state causes bots to concentrate wholly on moving towards ScriptedMoveTarget.
	//It's not an ideal way to do it, but it works.
	if(Bot.ActorReachable(SelectedSentinel) && !Bot.IsInState('ScriptedMove'))
	{
		Bot.ScriptedMoveTarget = SelectedSentinel;
		Bot.PushState('ScriptedMove');

		while(Bot.IsInState('ScriptedMove'))
		{
			if(VSize(Bot.ScriptedMoveTarget.Location - Bot.Pawn.Location) < RepairRange)
			{
				Bot.PopState();
				Bot.DoRangedAttackOn(SelectedSentinel);

				//Wait for bot to finish.
				while(Bot.Focus == SelectedSentinel && Bot.IsInState('RangedAttack'))
				{
					Sleep(0.1);
				}

				break;
			}
			else if((Bot.Enemy != none && Bot.bEnemyIsVisible) || Bot.PriorityObjective())
			{
				Bot.PopState();
				break;
			}

			Sleep(0.5);
		}
	}

	Sleep(1.0);
	Goto('Searching'); //Loop forever.

Upgrading:
	bBotTryingToUseDeployer = true;
	Bot.GoalString = "Going to upgrade"@SelectedSentinel;

	if(VSize(SelectedSentinel.Location - Bot.Pawn.Location) < UpgradeRange)
	{
		BotDoUpgrade();
	}
	else if(Bot.FindPathToward(SelectedSentinel) != none && !Bot.IsInState('ScriptedMove'))
	{
		Bot.ScriptedMoveTarget = SelectedSentinel;
		Bot.PushState('ScriptedMove');

		while(Bot.IsInState('ScriptedMove'))
		{
			if(VSize(SelectedSentinel.Location - Bot.Pawn.Location) < UpgradeRange)
			{
				Bot.PopState();
				BotDoUpgrade();
				break;
			}
			else if((Bot.Enemy != none && Bot.bEnemyIsVisible) || Bot.PriorityObjective())
			{
				Bot.PopState();
				break;
			}

			Sleep(0.5);
		}
	}

	Sleep(1.0);
	Goto('Searching'); //Loop forever.

End:
}

/**
 * State entered when primary fire held. Pressing secondary fire in this state does the appropriate action (deploy, open upgrade menu).
 */
simulated state DeployState
{
	simulated function BeginState(Name PreviousStateName)
	{
		bDeploying = false;

		if(Instigator.IsHumanControlled())
		{
			if(Instigator.IsLocallyControlled())
			{
				SymbolComponent.SetHidden(false);
			}
		}
		else if(Role == ROLE_Authority)
		{
			//Bots just deploy straight away. Upgrading is done elsewhere for them.
			bDeploying = true;
			IncrementFlashCount(); //So animations play.
			SetTimer(GetFireInterval(0), true, 'RefireCheckTimer'); //Ensure refiring prevented for the appropriate period.
			BotDeploy();
			Bot.StopFiring();
		}
	}

	simulated function BeginFire(byte FireModeNum)
	{
		global.BeginFire(FireModeNum);
		
		if(FireModeNum == 1 && !bDeploying)
		{
			bDeploying = true;
			IncrementFlashCount(); //So animations play.
			SetTimer(GetFireInterval(0), true, 'RefireCheckTimer'); //Ensure refiring prevented for the appropriate period.

			if(!TryOpenUpgradeMenu())
			{
				Deploy();
			}
		}
		else if(FireModeNum == 0)
		{
			if(Instigator.IsHumanControlled())
			{
				if(Instigator.IsLocallyControlled())
				{
					SymbolComponent.SetHidden(false);
				}
			}
		}
	}

	simulated function EndFire(byte FireModeNum)
	{
		global.EndFire(FireModeNum);

		if(FireModeNum == 0)
		{
			if(!bDeploying)
			{
				RefireCheckTimer();
			}
			else
			{
				SymbolComponent.SetHidden(true);
			}
		}
	}

	simulated function bool TryOpenUpgradeMenu()
	{
		local Sentinel S;
		local bool bResult;

		S = SentinelInRange(UpgradeRange, true, true);

		if(S != none)
		{
			SelectedSentinel = S;
			OpenUpgradeMenuFor(SelectedSentinel);
			bResult = true;
		}

		return bResult;
	}

	simulated function RefireCheckTimer()
	{
		bDeploying = false;

		if(!ShouldRefire())
		{
			GotoState('Active');
		}
	}

	//Go back to playing idle animations when firing animation finishes.
	simulated function OnAnimEnd(optional AnimNodeSequence SeqNode, optional float PlayedTime, optional float ExcessTime)
	{
		local int IdleIndex;

		if ( WorldInfo.NetMode != NM_DedicatedServer && WeaponIdleAnims.Length > 0 )
		{
			IdleIndex = Rand(WeaponIdleAnims.Length);
			PlayWeaponAnimation(WeaponIdleAnims[IdleIndex], 0.0, true);
			if(ArmIdleAnims.Length > IdleIndex && ArmsAnimSet != none)
			{
				PlayArmAnimation(ArmIdleAnims[IdleIndex], 0.0,, true);
			}
		}
	}

	simulated function PlayWeaponAnimation( Name Sequence, float fDesiredDuration, optional bool bLoop, optional SkeletalMeshComponent SkelMesh)
	{
		Global.PlayWeaponAnimation(Sequence,fDesiredDuration,bLoop,SkelMesh);
		ClearTimer('OnAnimEnd');
		if (!bLoop)
		{
			SetTimer(fDesiredDuration,false,'OnAnimEnd');
		}
	}

	/**
	 * Displays the symbol.
	 */
	simulated function Tick(float DeltaTime)
	{
		local Rotator AimRotation;
		local Vector Aim;
		local Vector SentinelLocation, SentinelNormal;
		local Actor SentinelBaseActor;
		local Sentinel S;

		if(!Instigator.IsLocallyControlled())
		{
			return;
		}

		AimRotation = Instigator.GetBaseAimRotation();
		Aim = Vector(AimRotation);

		if(LRI.bAllowUpgrades)
		{
			S = SentinelInRange(UpgradeRange, true, true);

			if(!CanUpgrade(S))
			{
				S = none;
			}
		}

		if(S != none)
		{
			Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_Upgrade, Instigator.PlayerReplicationInfo, S.PlayerReplicationInfo);
			SymbolComponent.DrawUpgradeSymbol(S, Rotator(S.Location - InstantFireStartTrace()));
		}
		else if(AimRotation.Pitch > TypeSwitchAngle)
		{
			if(class'SUtils'.static.FindCeilingSpot(CeilingInfo, GetTraceOwner(), InstantFireStartTrace(), Aim, SentinelLocation, SentinelNormal, SentinelBaseActor))
			{
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_Deploy, Instigator.PlayerReplicationInfo);
				SymbolComponent.DrawArrowSymbol(SentinelLocation, AimRotation, SentinelNormal);
			}
			else
			{
				SymbolComponent.DrawUnableSymbol(SentinelLocation, AimRotation);
			}
		}
		else
		{
			if(class'SUtils'.static.FindFloorSpot(FloorInfo, GetTraceOwner(), InstantFireStartTrace(), Aim, SentinelLocation))
			{
				Instigator.ReceiveLocalizedMessage(LocalMessageClass, DA_Deploy, Instigator.PlayerReplicationInfo);
				SymbolComponent.DrawArrowSymbol(SentinelLocation, AimRotation, vect(0.0, 0.0, 1.0));
			}
			else
			{
				SymbolComponent.DrawUnableSymbol(SentinelLocation, AimRotation);
			}
		}
	}

	simulated event bool IsFiring()
	{
		return true;
	}

	simulated function bool DenyClientWeaponSet()
	{
		return true;
	}

	//Stop bots from becoming stuck in this state.
	simulated function bool ShouldRefire()
	{
		if(!Instigator.IsHumanControlled())
			return false;

		return super.ShouldRefire();
	}

	function bool CanAttack(Actor Other)
	{
		return false;
	}

	simulated function EndState(Name NextStateName)
	{
		SymbolComponent.SetHidden(true);

		ClearFlashCount();
		ClearFlashLocation();
		ClearTimer('RefireCheckTimer');

		if (Instigator != none && AIController(Instigator.Controller) != None)
		{
			AIController(Instigator.Controller).NotifyWeaponFinishedFiring(self,CurrentFireMode);
		}
	}
}

`if(`notdefined(FINAL_RELEASE))
/**
 * For debugging. Causes the Sentinel being pointed at to attack you and ignore everything else.
 */
exec simulated function HitMe()
{
	ServerHitMe();
}

reliable server function ServerHitMe()
{
	local Sentinel S;
	local SentinelControllerComponent_Behaviour_TargetingPawn NewPawnTargetingBehaviour;

	if(WorldInfo.NetMode == NM_Standalone || Instigator.PlayerReplicationInfo.bAdmin)
	{
		S = SentinelInRange(100000.0, true, true, true);

		if(S != none)
		{
			NewPawnTargetingBehaviour = new(S.SController) class'SentinelControllerComponent_Behaviour_DebugTargetingPawn';
			NewPawnTargetingBehaviour.SetTargetInfo(Instigator, true, Instigator.Location, WorldInfo.TimeSeconds);
			S.SController.ChangeBehaviourTo(NewPawnTargetingBehaviour);
		}
	}
}

exec simulated function HitHere()
{
	ServerHitHere();
}

reliable server function ServerHitHere()
{
	local Sentinel S;

	if(WorldInfo.NetMode == NM_Standalone || Instigator.PlayerReplicationInfo.bAdmin)
	{
		S = SentinelInRange(100000.0, true, true, true);

		if(S != none)
		{
			S.SController.LocationTargetingBehaviour.SetTargetLocation(Instigator.Location);
			S.SController.ChangeBehaviourTo(S.SController.LocationTargetingBehaviour);
		}
	}
}

/**
 * For debugging. Reverses the effects of HitMe.
 */
exec simulated function StopHittingMe()
{
	ServerStopHittingMe();
}

reliable server function ServerStopHittingMe()
{
	local Sentinel S;

	if(WorldInfo.NetMode == NM_Standalone || Instigator.PlayerReplicationInfo.bAdmin)
	{
		S = SentinelInRange(100000.0, true, true, true);

		if(S != none && S.Controller.Focus == Instigator)
		{
			S.SController.ChangeBehaviourTo(S.SController.IdleBehaviour);
		}
	}
}
`endif

defaultproperties
{
	ItemName="Sentinel Deployer"
	PickupMessage="Sentinel Deployer"

	LRIClass=class'UTLRI_Sentinel'
	LocalMessageClass=class'UTLocalMessage_DeployerAction'

	FloorInfo=(SentinelClass=class'Sentinel_Floor',DeployRange=100.0)
	CeilingInfo=(SentinelClass=class'Sentinel_Ceiling',DeployRange=300.0,CeilingNormalPitchLimit=16)
	TypeSwitchAngle=1024

	DeploySound=SoundCue'Sentinel_Resources.Sounds.Deployer.BReload2Cue'

	UpgradeRange=320.0
	RepairRange=320.0
	RepairAmount=15.0
    SabotageAmount=30.0
	SabotageAmmoGain=0.05
	DestroyUpgradeChance=0.2
	RecycleEfficiency=0.5
	MinimumDistanceFromEnemySpawn=2300.0

	RepairSound=SoundCue'Sentinel_Resources.Sounds.Deployer.Repair'

	UpgradeMenuTemplate=UTUIScene_SentinelUpgradeMenu'Sentinel_Resources.Menus.SentinelUpgradeMenu'
	MenuUpdatePeriod=0.5

	BotSpacing=512.0
	BotRepairSearchRadius=2048.0
	BotUpgradeSearchRadius=2048.0
	BotCheatFactor=1.3

	WeaponColor=(R=128,G=128,B=128,A=255)

	Begin Object class=AnimNodeSequence Name=MeshSequenceA
	End Object

	Begin Object Name=FirstPersonMesh
		SkeletalMesh=SkeletalMesh'Pickups.Deployables.Mesh.SK_Deployables_Shield'
		AnimSets[0]=AnimSet'Pickups.Deployables.Anims.K_Deployables_Shield_1P'
		Animations=MeshSequenceA
		CollideActors=false
		BlockActors=false
		CastShadow=false
		bForceDirectLightMap=true
		bCastDynamicShadow=false
		BlockRigidBody=false
		FOV=60.0
		bUseAsOccluder=FALSE
	End Object

	ArmsAnimSet=AnimSet'Pickups.Deployables.Anims.K_Deployables_Shield_1P_Arms'

	WeaponFireAnim(0)=WeaponPutDown
	WeaponFireAnim(1)=WeaponIdle
	ArmFireAnim(0)=WeaponPutDown
	ArmFireAnim(1)=WeaponIdle

	AttachmentClass=class'UTAttachment_SentinelDeployer'

	Begin Object Name=PickupMesh
		SkeletalMesh=SkeletalMesh'Pickups.Deployables.Mesh.SK_Deployables_Shield'
		Translation=(Z=-20.0)
	End Object
	PivotTranslation=(X=0.0,Y=0.0,Z=0.0)

	bCanThrow=true //Note that the weapon is not actually throwable, ammo is thrown instead.
	bWarnIfInLocker=true
	bInstantHit=true //Secondary fire is instant-hit.
	bFastRepeater=true //Again, for secondary fire.

	PickupSound=SoundCue'Sentinel_Resources.Sounds.Deployer.ShieldGunPickupCue'
	WeaponPutDownSnd=SoundCue'A_Weapon_FlakCannon.Weapons.A_FlakCannon_LowerCue'
	WeaponEquipSnd=SoundCue'Sentinel_Resources.Sounds.Deployer.shieldgun_changeCue'

	WeaponFireSnd[0]=none
	WeaponFireSnd[1]=none //No sound, so that MakeNoise isn't called.

	FiringStatesArray(0)=DeployState

	WeaponFireTypes(0)=EWFT_Custom
	WeaponFireTypes(1)=EWFT_InstantHit

	InstantHitDamageTypes(1)=class'UTDmgType_SentinelDeployer'

	FireInterval(0)=2.0
	FireInterval(1)=0.3

	ShotCost(0)=0
	ShotCost(1)=1

	ShouldFireOnRelease(0)=0
	ShouldFireOnRelease(1)=0

	MaxDesireability=1.6
	AIRating=-1.0 //Bots should not normally use this weapon.
	CurrentRating=-1.0
	Priority=0.1 //Has to be >0.0 or it is automatically set to InventoryWeight

	EquipTime=0.8

	InventoryGroup=10
	GroupWeight=0.7

	IconX=394
	IconY=38
	IconWidth=60
	IconHeight=38
	
	CrosshairImage=Texture2D'EngineResources.White'
	CrossHairCoordinates=(U=0,V=0,UL=3,VL=3)
	SimpleCrosshairCoordinates=(U=0,V=0,UL=3,VL=3)
	CrosshairColor=(R=64,G=255,B=64)
	IconCoordinates=(U=620,V=587,UL=59,VL=56)

	AmmoCount=0
	LockerAmmoCount=0
	MaxAmmoCount=280
	RespawnTime=40.0
	//bSuperWeapon=true

	Begin Object Class=SymbolStaticMeshComponent Name=SymbolComponent0
	End Object
	SymbolComponent=SymbolComponent0
	Components.Add(SymbolComponent0)
}